/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name:  dio.c
**
**    Description:
**      This module implements a bunch of routines that do I/O through the
**        buffer cache.
**
*******************************************************************************
******************************************************************************/

#if defined(__linux__)
#include <linux/fs.h>
#include <linux/locks.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#endif
#if defined(__FreeBSD__)
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/resourcevar.h>
#endif

#include "gfs.h"
#include "dio.h"
#include "glock.h"


/******************************************************************************
*******************************************************************************
**
** gfs_remove_bh()
**
** Description:
**   Remove a buffer from the list maintained for invalidation.
**
** Input(s):
**   bh - the buffer head to remove.
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void gfs_remove_bh(buffer_head_t *bh)
{
  gfs_bhdata_t *bd = bh2pd(bh);

  ASSERT(bd != NULL);

  bd->bd_bh = NULL;
  bd->bd_glp->gl_touched = TRUE;
#if 0
  bh->b_rm_fn = NULL;
#endif
  bh->b_driver1 = NULL;
}


/******************************************************************************
*******************************************************************************
**
** gfs_inval_bh()
**
** Description:
**   Invalidate all buffers associated with a Glock.
**
** Input(s):
**   lock - the lock whose buffers need to be invalidated 
**
** Returns:
**   
**
*******************************************************************************
******************************************************************************/

void gfs_inval_bh(gfs_glock_t *gl)
{
  gfs_list_head_t *next, *head;
  gfs_bhdata_t *bd;

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_inval_bh()\n"););


  GFS_SEMA_DOWN(&gl->gl_buflock);

  head = &gl->gl_bufs;

  next = head->next;
  for (;;) {
    gfs_list_head_t * tmp = next;

    next = next->next;
    if (tmp == head)
      break;

    bd = gfs_list_entry(tmp, gfs_bhdata_t, bd_list);

    if (bd->bd_bh)
	  printf("ERROR shold be calling mark_buffer_uptodate\n");
	  /*   mark_buffer_uptodate(bd->bd_bh, FALSE); */
  }

  GFS_SEMA_UP(&gl->gl_buflock);


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_inval_bh()\n"););

  return;
}


/******************************************************************************
*******************************************************************************
**
** gfs_sync_bh()
**
** Description:
**   Sync all dirty buffers associated with a Glock.
**
** Input(s):
**   lock - the lock whose buffers need to be written to disk
**
** Returns:
**   
**
*******************************************************************************
******************************************************************************/

void gfs_sync_bh(gfs_glock_t *gl)
{
  gfs_list_head_t *next, *head;
  gfs_bhdata_t *bd;

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_sync_bh()\n"););
#if 0

  GFS_SEMA_DOWN(&gl->gl_buflock);

  /*  Go through the list once and start the I/O  */

  head = &gl->gl_bufs;

  next = head->next;
  for (;;) {
    gfs_list_head_t * tmp = next;

    next = next->next;
    if (tmp == head)
      break;

    bd = gfs_list_entry(tmp, gfs_bhdata_t, bd_list);

    if (bd->bd_bh && (buffer_dirty(bd->bd_bh) || buffer_locked(bd->bd_bh)))
    {
      bd->bd_io = TRUE;
      bd->bd_bh->b_count++;

      if (!buffer_locked(bd->bd_bh))
      {
	bd->bd_bh->b_flushtime = 0;
	ll_rw_block(WRITE, 1, &(bd->bd_bh));
      }
    }
  }


  /*  Go through the list again and wait for the I/O to complete  */

  next = head->next;
  for (;;) {
    gfs_list_head_t * tmp = next;

    next = next->next;
    if (tmp == head)
      break;

    bd = gfs_list_entry(tmp, gfs_bhdata_t, bd_list);

    if (bd->bd_io)
    {
      if (bd->bd_bh)
      {
	gfs_wait_on_buffer(bd->bd_bh);

	while (buffer_dirty(bd->bd_bh))
	{
	  if (buffer_locked(bd->bd_bh))
	    gfs_wait_on_buffer(bd->bd_bh);
	  else
	    ll_rw_block(WRITE, 1, &(bd->bd_bh));
	}

	bd->bd_bh->b_count--;
      }

      bd->bd_io = FALSE;
    }
  }

  GFS_SEMA_UP(&gl->gl_buflock);

#endif
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_sync_bh()\n"););

  return;
}


/******************************************************************************
*******************************************************************************
**
** gfs_dgetblk()
**
** Description:
**   Get an empty buffer.
**
** Input(s):
**   sdp - contains the device the block lives on
**   blkno - the file system block number
**   glock - the glock associated with this block
**
** Returns:
**   The buffer_head_t on success, NULL on failure
**
*******************************************************************************
******************************************************************************/

buffer_head_t *gfs_dgetblk(gfs_sbd_t *sdp, uint64 blkno, uint64 glock)
{
  buffer_head_t *bh;
  gfs_bhdata_t *bd;
  gfs_glock_t *gl;
  struct ucred *cred;

  int error;
    
  IFDEBUG(DEBUG_DIO1,
          printk(DEBUG_HOW "Entering gfs_dgetblk()\n"););
  
  IFDEBUG(DEBUG_DIO2,
          printk(DEBUG_HOW "  bbno = %u, glock = %u\n",
                 (uint)blkno,(uint)glock););
  
 
  cred = NOCRED;

#if 0
  /*  linux has set_block_size we have to do it my hand sdp->sd_sb.sb_bsize  */
  /* note add add bshift filed to speed to block calulations */
  printf("gfs_dgetblk: START bread FSBTOBB(bbno): %u size: %u\n", 
		 (uint)FSBTOBB(blkno), sdp->sd_sb.sb_bsize);
  if ((error = bread(sdp->sd_dev, FSBTOBB(blkno), sdp->sd_sb.sb_bsize, cred, &bh)) != 0)
	  goto enderr1;
  printf("gfs_dgetblk: DONE bread FSBTOBB(bbno): %u size: %u bh: %p\n",
		 (uint)FSBTOBB(blkno), sdp->sd_sb.sb_bsize,bh);
#endif
  printf("gfs_dgetblk: START bread FSBTOBB(bbno): %u size: %u\n", 
		 (uint)FSBTOBB(blkno), sdp->sd_sb.sb_bsize);
  bh = getblk(sdp->sd_dev, FSBTOBB(blkno), sdp->sd_sb.sb_bsize, 0, 0);
  printf("gfs_dgetblk: DONE bread FSBTOBB(bbno): %u size: %u bh: %p\n",
		 (uint)FSBTOBB(blkno), sdp->sd_sb.sb_bsize,bh);

#if 0
  print_buf(bh);
#endif

  if (bh->b_usecount > 20)
    printk("GFS: over active buffer %"FMT_64"u, %d\n", blkno, bh->b_usecount);

#if 0
  if (buffer_uptodate(bh) && bh->b_rm_fn != gfs_remove_bh)
  {
    /*printk(KERN_WARNING "GFS:  buffer prematurely up to date!\n");*/
    mark_buffer_uptodate(bh, FALSE);
  }
#endif

  /* buffer is from a deleted file */
#if 0
  if ((bh->b_rm_fn == gfs_remove_bh) && (glock != bh2pd(bh)->bd_glock))
  {
    gfs_remove_bh(bh);
    mark_buffer_uptodate(bh, FALSE);
  }
#endif

  
  /* insert new bh into this lock's bufs list */
  /*   if (bh->b_rm_fn != gfs_remove_bh) { */

  if (1) {  /* assume not at this point RMC */
    bd = (gfs_bhdata_t *)gmalloc(sizeof(gfs_bhdata_t));
    if (!bd)     
      goto enderr2;

    gl = gfs_get_glstruct(sdp, glock, FALSE);
    ASSERT(gl);

    bd->bd_glp = gl;
    bd->bd_bh = bh;
    bd->bd_glock = glock;
    bd->bd_io = FALSE;
    bh->b_driver1 = bd; /* NOTE abstrace this out */
	/*    bh->b_rm_fn = gfs_remove_bh; What it this for? RMC */

    GFS_SEMA_DOWN(&gl->gl_buflock);
    gfs_list_add(&bd->bd_list, &gl->gl_bufs);
    GFS_SEMA_UP(&gl->gl_buflock);

#if 0
    mark_buffer_uptodate(bh, FALSE);
#endif
#if 0
	/* else BSD */
	printf("Marking bh %p as BUSY\n",bh);
	bh->b_flags |= B_BUSY;
#endif

    gfs_put_glstruct(sdp, gl);
  }


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dgetblk() <OK>\n"););

  return(bh);

 enderr2:
  brelse(bh);

 enderr1:
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dgetblk() <ERROR>\n"););
  return(NULL);
}


/******************************************************************************
*******************************************************************************
**
** gfs_mark_uptodate()
**
** Description:
**
** Input(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void gfs_mark_uptodate(buffer_head_t *bh, int state)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_mark_uptodate()\n"););
 
  printf("Yo What does this do!!!!!\n");
#if 0
  printf("Trying UNBUSY it\n");
  bh->b_flags &= ~B_BUSY;
#endif

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_mark_uptodate()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_buffer_uptodate()
**
** Description:
**
** Input(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int gfs_buffer_uptodate(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_buffer_uptodate()\n"););


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_buffer_uptodate()\n"););
  /* 
   * hmm if buffer in cache it'v valid? need to rethink this..
   * do when and were would we invalidate it 
   */
  return ((bh->b_flags & B_DONE) | (bh->b_flags & B_CACHE));
}


/******************************************************************************
*******************************************************************************
**
** gfs_wait_on_buffer()
**
** Description:
**
** Input(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void gfs_wait_on_buffer(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_wait_on_buffer()\n"););


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_wait_on_buffer(NOACTION)\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_dread()
**
** Description:
**   Read a block from disk
**
** Input(s):
**   sdp - contains the device the block lives on
**   blkno - the file system block number
**   glock - the glock number for this buffer
**
** Returns:
**   The read block on success, NULL on failure
**
*******************************************************************************
******************************************************************************/

buffer_head_t *gfs_dread(gfs_sbd_t *sdp, uint64 blkno, uint64 glock)
{
  buffer_head_t *bh;


  IFDEBUG(DEBUG_DIO1,
		  printk(DEBUG_HOW "Entering gfs_dread() glock %u/%u\n",(u_int)(glock >> 32),(uint)glock););
  
  bh = gfs_dgetblk(sdp, blkno, glock);
  if (!bh)
  {
    IFDEBUG(DEBUG_DIO1,
			printk(DEBUG_HOW "Exiting gfs_dread() <First OK>\n"););
    return(bh);
  }

  gfs_dreread(sdp, 1, &bh, TRUE);
  /*  print_buf(bh); */
  

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dread()\n"););

  return(bh);
}


/******************************************************************************
*******************************************************************************
**
** gfs_dreread()
**
** Description:
**   Reread a block
**
** Input(s):
**   bh - array of blocks to reread
**   len - the length of the array
**   wait - TRUE if the function should wait for the read to complete
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_dreread(gfs_sbd_t *sdp, int len, buffer_head_t **bh, int wait)
{

  int x, retries;
  struct buf *bp  = *bh;
  int error;
  struct ucred *cred;
  struct vnode *vp = sdp->sd_dev;

  cred = NOCRED;


  IFDEBUG(DEBUG_DIO1,
		  printk(DEBUG_HOW "Entering gfs_dreread()\n"););
  
  if (!bh){
	IFDEBUG(DEBUG_DIO1,
			printk(DEBUG_HOW "Exiting gfs_dreread() <ERROR>\n"););
	return (-EINVAL);
  }


  /* if not found in cache, do some I/O */
  if ((bp->b_flags & B_CACHE) == 0) {
	if (curproc != NULL)
	  curproc->p_stats->p_ru.ru_inblock++;
	KASSERT(!(bp->b_flags & B_ASYNC), ("bread: illegal async bp %p", bp));
	bp->b_flags |= B_READ;
	bp->b_flags &= ~(B_ERROR | B_INVAL);
	if (bp->b_rcred == NOCRED) {
	  if (cred != NOCRED)
		crhold(cred);
	  bp->b_rcred = cred;
	}
	vfs_busy_pages(bp, 0);
#   if 0
	/* not sure what this does but mckusik changed the lockmgr */
	/* and now things are crashing all over the place */
	BUF_KERNPROC(bp);
#   endif
	VOP_STRATEGY(vp, bp);
	error = biowait(bp);
	if(error){
	  printf("gfs_dreread got an error cp0 %u\n",error);
	}
  }
  
  if (!wait){
    IFDEBUG(DEBUG_DIO1,
			printk(DEBUG_HOW "Exiting gfs_dreread()\n"););
    return(0);
  }
  
  
  for (x = 0; x < len; x++){
	bp = bh[x];
    if (bh[x]){
#if 0
      gfs_wait_on_buffer(bh[x]); /* this is doing nothing at this point RMC */
#endif
      retries = 0;
      while (!gfs_buffer_uptodate(bh[x])){
		/* check the flags for B_DONE this should never fail */
		if (retries++ > GFS_MAX_RETRIES)
		  ASSERT(gfs_buffer_uptodate(bh[x]));
		{
		  /* if not found in cache, do some I/O */
		  if ((bp->b_flags & B_CACHE) == 0) {
			if (curproc != NULL)
			  curproc->p_stats->p_ru.ru_inblock++;
			KASSERT(!(bp->b_flags & B_ASYNC), ("bread: illegal async bp %p", bp));
			bp->b_flags |= B_READ;
			bp->b_flags &= ~(B_ERROR | B_INVAL);
			if (bp->b_rcred == NOCRED) {
			  if (cred != NOCRED)
				crhold(cred);
			  bp->b_rcred = cred;
			}
			vfs_busy_pages(bp, 0);
#           if 0
			/* see above comment */
			BUF_KERNPROC(bp);
#           endif
			VOP_STRATEGY(vp, bp);
			error = biowait(bp);
			if(error){
			  printf("gfs_dreread biowait got an ERROR %u\n",error);
			}	
#if 0
			printf ("gfs_dreread bp->bflag & B_DONE %u\n", ((bp->b_flags & B_DONE)?1:0));
			printf ("gfs_dreread bp->bflag & B_BUSY %u\n", ((bp->b_flags & B_BUSY)?1:0));
			printf ("gfs_dreread bp->bflag & B_CACHE %u\n", ((bp->b_flags & B_CACHE)?1:0));
#endif
		  }
		}
      }

      if (retries)
	printk(KERN_ERR "GFS:  retry successful in gfs_dreread()  (not an error)\n");
    }
  }

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dreread()\n"););
  
  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_dwrite()
**
** Description:
**   Write a block
**
** Input(s):
**   bh - the block to write
**   flags - Flags on what to do with the write
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_dwrite(buffer_head_t *bh, int flags)
{
  IFDEBUG(DEBUG_DIO1,
		  printk(DEBUG_HOW "Entering gfs_dwrite()");
		  printf("\t bh %p\n",bh);
		  );

  if (bh == NULL)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_dwrite()\n"););
    return(-EINVAL);
  }

  /*  Tell the Glock layer to increment the version number for this lock  */

  if (bh2pd(bh))
    bh2pd(bh)->bd_glp->gl_incr = TRUE;
  else
    printk(KERN_ERR "GFS:  Whoah!  Writing an unregistered buffer\n");

  if (bh->b_qindex != QUEUE_NONE) {
	bremfree(bh);
  }

  /* Note this will chage be removed once the new lockmgr stuff settles down */
  bh->b_flags |= B_BUSY; /* seems to be the same a dirty??? */
#if 0
  bdwrite(bh);
#else
  bwrite(bh);
#endif
  /*  bh->b_flags &= ~(B_BUSY|B_CACHE); */
  printf ("gfs_dreread bp->bflag & B_DONE %u\n", ((bh->b_flags & B_DONE)?1:0));
  printf ("gfs_dreread bp->bflag & B_BUSY %u\n", ((bh->b_flags & B_BUSY)?1:0));
  printf ("gfs_dreread bp->bflag & B_CACHE %u\n", ((bh->b_flags & B_CACHE)?1:0));
#if 0
  bh->b_flags &= ~(B_BUSY);
  bh->b_flags |= B_INVAL;
#endif
  /* just mark the thing dirty for now... don't want it to do any caching */
  
  if (flags & DW_NOTMETA)  {
	/*    mark_buffer_dirty(bh, 1); */
	/* Delayed Write... I'm not sure this what we want to do here 
	 * we may have to specifically to a bwrite... or even worse do our 
	 * own cache coherency
	 */
  } else {
	/*    printk("Me lose brain?  Uh-oh!\n"); */
	/* mark_buffer_dirty(bh, 1); */
  }



#if 0
  /* this is done by bdwrite */
  if (flags & DW_NOKEEP)
    gfs_dforget(bh);
#endif

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dwrite()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_dforget()
**
** Description:
**   Release a block, making sure it isn't kept in the cache
**
** Input(s):
**   bh - the block to release
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_dforget(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_dforget()\n"););

  if (bh == NULL)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_dforget()\n"););
    return(-EINVAL);
  }

  brelse(bh);

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dforget()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** free_bufs()
**
** Description:
**   Free gfs_bhdata_t structures associated with this Glock, but not with
**   a buffer head.  If the Glock structure is not in use, free that too.
**
** Input(s):
**   gl - A pointer to the Glock structure holding the buffers
**
*******************************************************************************
******************************************************************************/

static void free_bufs(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_list_head_t *head, *next;
  gfs_bhdata_t *bd;
  int del = FALSE;


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering freebufs()\n"););


  GFS_SEMA_DOWN(&gl->gl_buflock);
    
  head = &gl->gl_bufs;
    
  next = head->next;
  while (1)
  {
    gfs_list_head_t *tmp = next;
      
    next = next->next;
    if (tmp == head)
      break;
    bd = gfs_list_entry(tmp, gfs_bhdata_t, bd_list);
      
    if (!bd->bd_bh)
    {
      gfs_list_del(&bd->bd_list);
      gfree(bd, sizeof(gfs_bhdata_t));
    }
  }

  if (gfs_list_empty(&gl->gl_bufs))
    del = TRUE;
      
  GFS_SEMA_UP(&gl->gl_buflock);

  if (del && gl->gl_count == 0 && !gl->gl_held)
  {
    if (gl->gl_lockdata)
      sdp->sd_lockstruct->release(sdp->sd_lockdata, &gl->gl_lockdata);
    gfs_list_del(&gl->gl_next);
    gfree(gl, sizeof(gfs_glock_t));
  }


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting freebufs()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_bufd()
**
** Description:
**   A kernel thread that wakes up every so often and gets rid of all 
**   gfs_bhdata_t structures laying around.
**
** Input(s):
**   data - a pointer to the gfs_sbd_t structure for the filesystem
**
** Returns:
**   an integer that I don't understand
**
*******************************************************************************
******************************************************************************/

int 
gfs_bufd(void *data){
#if 0
  gfs_sbd_t *sdp = (gfs_sbd_t *)data;

  int x, touched;
  gfs_list_head_t *head, *next;
  gfs_glock_t *gl;
  unsigned long flags;
#endif


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_bufd()\n"););

#if 0
  ASSERT(sdp);


  /*  Set up the thread  */

  lock_kernel();

  exit_mm(current);

  current->session = 1;
  current->pgrp = 1;

  spin_lock_irqsave(&current->sigmask_lock, flags);
  sigfillset(&current->blocked);
  recalc_sigpending(current);
  spin_unlock_irqrestore(&current->sigmask_lock, flags);

  current->fs->umask = 0;
  
  sprintf(current->comm, "gfs_bufd");

  unlock_kernel();


  /*  Wake up the process who started us  */

  up(&sdp->sd_bufd_lock);


  while (1)
  {
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(5 * HZ);

    IFDEBUG(DEBUG_DIO3,
	    printk(DEBUG_HOW "\n\n\n  bufd wakes up\n"););

    if (sdp->sd_bufd_stop)
    {
      IFDEBUG(DEBUG_DIO2,
	      printk(DEBUG_HOW "  bufd caught signal\n"););
      break;
    }

    IFDEBUG(DEBUG_INODE3,
	    printk(DEBUG_HOW "%d\n", sdp->sd_id_count););

    for (x = 0; x < GFS_GHASH_SIZE; x++)
    {
      GFS_SEMA_DOWN(&sdp->sd_gla_lock[x]);

      head = &sdp->sd_gla[x];

      next = head->next;
      while (1)
      {
	gfs_list_head_t *tmp = next;

	next = next->next;
	if (tmp == head)
	  break;
	gl = gfs_list_entry(tmp, gfs_glock_t, gl_next);

	IFDEBUG(DEBUG_DIO3,
		printk(DEBUG_HOW "  Glock %"FMT_64"u, held = %d, buffers = %d\n",
		       gl->gl_glock, gl->gl_held, !gfs_list_empty(&gl->gl_bufs)););

	touched = xchg(&gl->gl_touched, FALSE);

	if (touched)
	  free_bufs(sdp, gl);
      }

      GFS_SEMA_UP(&sdp->sd_gla_lock[x]);
    }
  }


  up(&sdp->sd_bufd_lock);
#endif

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_bufd()\n"););

  return(0);
}


/*
 *
 *
 *
 *
 *  Synchronous routines used an mount/unmount time
 *
 *
 *
 *
 */


/******************************************************************************
*******************************************************************************
**
** gfs_sread()
**
** Description:
**   Read a block from disk.  Don't try to link it into the Glock array.
**
**   THIS WILL HOSE EXISTING BUFFERS IN THE CACHE, SO BE CAREFUL!!!
**
** Input(s):
**   sdp - contains the device the block lives on
**   bbno - the block number
**   bytes - number of bytes in the block
**
** Returns:
**   The read block on success, NULL on failure
**
*******************************************************************************
******************************************************************************/


buffer_head_t *
gfs_sread(gfs_sbd_t *sdp, uint64 bbno, uint32 bytes)
{
  uint32 size;
  buffer_head_t *bh;


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_sread() bbno %d bytes %d\n",(uint)bbno,bytes););

  if (bytes > PAGE_SIZE)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_sread()\n"););
    return(NULL);
  }

  for (size = 512; size < bytes; size <<= 1) ;

  /*  bh = getblk(sdp->sd_dev, bbno, size, 0, 0); */ /* bread does this */
  { /* block */
    struct ucred *cred = NOCRED;
	int error;
	if ((error = bread(sdp->sd_dev, FSBTOBB(bbno), size, cred, &bh)) != 0){
	  printf("GFS: gfs_sread: bread error %d\n",error);
	  gfs_sforget(bh);
	  bh = NULL;
	}
  } /* end block */


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_sread()\n"););

  return(bh);
}


/******************************************************************************
*******************************************************************************
**
** gfs_swrite()
**
** Description:
**   Write a block.  Do it synchrnonously.  Don't worry about the Glock array.
**
** Input(s):
**   bh - the block to write
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_swrite(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_swrite()\n"););

#if 0
  mark_buffer_dirty(bh, 1);
  ll_rw_block(WRITE, 1, &bh);
#endif

  gfs_wait_on_buffer(bh);
#if 0
  ASSERT(!buffer_dirty(bh));
#endif

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_swrite()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_sforget()
**
** Description:
**   Release a block, making sure it isn't kept in the cache
**
** Input(s):
**   bh - the block to release
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_sforget(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_sforget()\n"););


  if (!bh)
    return(-EINVAL);

  brelse(bh);


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_sforget()\n"););

  return(0);
}

int
print_buf(buffer_head_t *bp) {
  printf("hash        %p\n", bp->b_hash);
  printf("vnbufs      %p\n", bp->b_vnbufs);
  printf("bfreelist   %p\n", bp->b_freelist);
  printf("b_act       %p\n", bp->b_act);
  /*  printf("b_proc    0x%x\n", bp->b_proc); */
  printf("Flags:	  %p\n", bp->b_flags);
  printf("qindex:	  %d\n", bp->b_qindex);
  printf("usecount:   %d\n", bp->b_usecount);
  printf("xFlags:	  %p\n", bp->b_xflags);
  printf("Error:	  %d\n", bp->b_error);
  printf("bufsise     %u\n", bp->b_bufsize);
  printf("Xfer Cnt    %lu\n", bp->b_bcount);
  printf("Residual    %lu\n", bp->b_resid);
  printf("Device:	  %d (%d,%d)\n",bp->b_dev, major(bp->b_dev), minor(bp->b_dev));
  printf("data      %p\n",bp->b_data);
  printf("kvabase   %p\n",bp->b_kvabase);
  printf("kvasize   %p\n",bp->b_kvasize);
  printf("lblkno      %d\n",bp->b_lblkno);
  printf("blkno       %d\n",bp->b_blkno);
  printf("offset      %d\n",bp->b_offset);
  printf("iodone    %p\n",bp->b_iodone);
  printf("iodone_ch %p\n",bp->b_iodone_chain);
  printf("vp        %p\n",bp->b_vp);
  printf("dirtyoff    %d\n",bp->b_dirtyoff);
  printf("dirtyend    %d\n",bp->b_dirtyend);
  printf("b_rcred   %p\n",bp->b_rcred);
  printf("b_wcred   %p\n",bp->b_wcred);
/*   printf("validoff    %d\n",bp->b_validoff); */
/*   printf("validend    %d\n",bp->b_validend); */
  printf("pblkno    %p\n", bp->b_pblkno);
  printf("saveaddr  %p\n", bp->b_saveaddr);
/*   printf("savekva   %p\n", bp->b_savekva); */
  printf("driver1   %p\n", bp->b_driver1);
  printf("driver2   %p\n", bp->b_driver2);
  { 
	int i,j;
    int start = 0 ,end = 16;
	u_char *buffer = bp->b_data;
	for (i = start; i < end; i++){
	  for (j = 0; j < 16; j++) {
		printf("%02x ", (unsigned char)buffer[16 * i + j]);
	  }
	  printf("    ");
	  for (j = 0; j < 16; j++){
		/*		if (isprint(buffer[16 * i + j])) */
		if(1)
		  printf("%c", buffer[16 * i + j]);
		else
		  printf(" ");
	  }
	  printf("\n");
	}
	printf("\n");
  }
  
  return (0);
}
