/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name:  dio.c
**
**    Description:
**      This module implements a bunch of routines that do I/O through the
**        buffer cache.
**
*******************************************************************************
******************************************************************************/

#include <linux/locks.h>
#include <linux/smp_lock.h>

#include "gfs.h"
#include "dio.h"
#include "glock.h"


/******************************************************************************
*******************************************************************************
**
** gfs_remove_bh()
**
** Description:
**   Remove a buffer from the list maintained for invalidation.
**
** Input(s):
**   bh - the buffer head to remove.
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void gfs_remove_bh(buffer_head_t *bh)
{
  gfs_bhdata_t *bd = bh2pd(bh);

  ASSERT(bd != NULL);

  bd->bd_bh = NULL;
  bd->bd_glp->gl_touched = TRUE;

  bh->b_rm_fn = NULL;
  bh->b_pdata = NULL;
}


/******************************************************************************
*******************************************************************************
**
** gfs_inval_bh()
**
** Description:
**   Invalidate all buffers associated with a Glock.
**
** Input(s):
**   lock - the lock whose buffers need to be invalidated 
**
** Returns:
**   
**
*******************************************************************************
******************************************************************************/

void gfs_inval_bh(gfs_glock_t *gl)
{
  gfs_list_head_t *next, *head;
  gfs_bhdata_t *bd;

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_inval_bh()\n"););


  GFS_SEMA_DOWN(&gl->gl_buflock);

  head = &gl->gl_bufs;

  next = head->next;
  for (;;) {
    gfs_list_head_t * tmp = next;

    next = next->next;
    if (tmp == head)
      break;

    bd = gfs_list_entry(tmp, gfs_bhdata_t, bd_list);

    if (bd->bd_bh)
      mark_buffer_uptodate(bd->bd_bh, FALSE);
  }

  GFS_SEMA_UP(&gl->gl_buflock);


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_inval_bh()\n"););

  return;
}


/******************************************************************************
*******************************************************************************
**
** gfs_sync_bh()
**
** Description:
**   Sync all dirty buffers associated with a Glock.
**
** Input(s):
**   lock - the lock whose buffers need to be written to disk
**
** Returns:
**   
**
*******************************************************************************
******************************************************************************/

void gfs_sync_bh(gfs_glock_t *gl)
{
  gfs_list_head_t *next, *head;
  gfs_bhdata_t *bd;

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_sync_bh()\n"););


  GFS_SEMA_DOWN(&gl->gl_buflock);

  /*  Go through the list once and start the I/O  */

  head = &gl->gl_bufs;

  next = head->next;
  for (;;) {
    gfs_list_head_t * tmp = next;

    next = next->next;
    if (tmp == head)
      break;

    bd = gfs_list_entry(tmp, gfs_bhdata_t, bd_list);

    if (bd->bd_bh && (buffer_dirty(bd->bd_bh) || buffer_locked(bd->bd_bh)))
    {
      bd->bd_io = TRUE;
      bd->bd_bh->b_count++;

      if (!buffer_locked(bd->bd_bh))
      {
	bd->bd_bh->b_flushtime = 0;
	ll_rw_block(WRITE, 1, &(bd->bd_bh));
      }
    }
  }


  /*  Go through the list again and wait for the I/O to complete  */

  next = head->next;
  for (;;) {
    gfs_list_head_t * tmp = next;

    next = next->next;
    if (tmp == head)
      break;

    bd = gfs_list_entry(tmp, gfs_bhdata_t, bd_list);

    if (bd->bd_io)
    {
      if (bd->bd_bh)
      {
	gfs_wait_on_buffer(bd->bd_bh);

	while (buffer_dirty(bd->bd_bh))
	{
	  if (buffer_locked(bd->bd_bh))
	    gfs_wait_on_buffer(bd->bd_bh);
	  else
	    ll_rw_block(WRITE, 1, &(bd->bd_bh));
	}

	bd->bd_bh->b_count--;
      }

      bd->bd_io = FALSE;
    }
  }

  GFS_SEMA_UP(&gl->gl_buflock);


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_sync_bh()\n"););

  return;
}


/******************************************************************************
*******************************************************************************
**
** gfs_dgetblk()
**
** Description:
**   Get an empty buffer.
**
** Input(s):
**   sdp - contains the device the block lives on
**   blkno - the file system block number
**   glock - the glock associated with this block
**
** Returns:
**   The buffer_head_t on success, NULL on failure
**
*******************************************************************************
******************************************************************************/

buffer_head_t *gfs_dgetblk(gfs_sbd_t *sdp, uint64 blkno, uint64 glock)
{
  buffer_head_t *bh;
  gfs_bhdata_t *bd;
  gfs_glock_t *gl;


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_dgetblk()\n"););

  IFDEBUG(DEBUG_DIO2,
	  printk(DEBUG_HOW "  blkno = %"FMT_64"u\n", blkno););


  /*  This is Linux, so sdp->sd_fsb2bb_shift will always be zero.
      If this was Irix or FreeBSD, the block number would be:
      blkno << sdp->sd_fsb2bb_shift.  */

  bh = getblk(sdp->sd_vsb->s_dev, blkno, sdp->sd_sb.sb_bsize);

  if (!bh)
    goto enderr1;


  if (bh->b_count > 20)
    printk("GFS: over active buffer %"FMT_64"u, %d\n", blkno, bh->b_count);


  if (buffer_uptodate(bh) && bh->b_rm_fn != gfs_remove_bh)
  {
    /*printk(KERN_WARNING "GFS:  buffer prematurely up to date!\n");*/
    mark_buffer_uptodate(bh, FALSE);
  }


  /* buffer is from a deleted file */

  if ((bh->b_rm_fn == gfs_remove_bh) && (glock != bh2pd(bh)->bd_glock))
  {
    gfs_remove_bh(bh);
    mark_buffer_uptodate(bh, FALSE);
  }

  
  /* insert new bh into this lock's bufs list */

  if (bh->b_rm_fn != gfs_remove_bh) 
  {
    bd = (gfs_bhdata_t *)gmalloc(sizeof(gfs_bhdata_t));
    if (!bd)     
      goto enderr2;

    gl = gfs_get_glstruct(sdp, glock, FALSE);
    ASSERT(gl);

    bd->bd_glp = gl;
    bd->bd_bh = bh;
    bd->bd_glock = glock;
    bd->bd_io = FALSE;
    bh->b_pdata = bd;
    bh->b_rm_fn = gfs_remove_bh;

    GFS_SEMA_DOWN(&gl->gl_buflock);
    gfs_list_add(&bd->bd_list, &gl->gl_bufs);
    GFS_SEMA_UP(&gl->gl_buflock);

    mark_buffer_uptodate(bh, FALSE);

    gfs_put_glstruct(sdp, gl);
  }


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dgetblk()\n"););

  return(bh);

 enderr2:
  brelse(bh);

 enderr1:
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dgetblk()\n"););
  return(NULL);
}


/******************************************************************************
*******************************************************************************
**
** gfs_mark_uptodate()
**
** Description:
**
** Input(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void gfs_mark_uptodate(buffer_head_t *bh, int state)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_mark_uptodate()\n"););

  mark_buffer_uptodate(bh, state);

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_mark_uptodate()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_buffer_uptodate()
**
** Description:
**
** Input(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int gfs_buffer_uptodate(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_buffer_uptodate()\n"););

  return(buffer_uptodate(bh));
}


/******************************************************************************
*******************************************************************************
**
** gfs_wait_on_buffer()
**
** Description:
**
** Input(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void gfs_wait_on_buffer(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_wait_on_buffer()\n"););

  wait_on_buffer(bh);

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_wait_on_buffer()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_dread()
**
** Description:
**   Read a block from disk
**
** Input(s):
**   sdp - contains the device the block lives on
**   blkno - the file system block number
**   glock - the glock number for this buffer
**
** Returns:
**   The read block on success, NULL on failure
**
*******************************************************************************
******************************************************************************/

buffer_head_t *gfs_dread(gfs_sbd_t *sdp, uint64 blkno, uint64 glock)
{
  buffer_head_t *bh;


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_dread()\n"););


  bh = gfs_dgetblk(sdp, blkno, glock);
  if (!bh)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_dread()\n"););
    return(bh);
  }

  gfs_dreread(sdp, 1, &bh, TRUE);
  

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dread()\n"););

  return(bh);
}


/******************************************************************************
*******************************************************************************
**
** gfs_dreread()
**
** Description:
**   Reread a block
**
** Input(s):
**   bh - array of blocks to reread
**   len - the length of the array
**   wait - TRUE if the function should wait for the read to complete
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_dreread(gfs_sbd_t *sdp, int len, buffer_head_t **bh, int wait)
{
  int x, retries;

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_dreread()\n"););

  if (!bh)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_dreread()\n"););
    return(-EINVAL);
  }


  ll_rw_block(READ, len, bh);


  if (!wait)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_dreread()\n"););
    return(0);
  }

  for (x = 0; x < len; x++)
    if (bh[x])
    {
      gfs_wait_on_buffer(bh[x]);

      retries = 0;

      while (!gfs_buffer_uptodate(bh[x]))
      {
	if (retries++ > GFS_MAX_RETRIES)
	{
	  if (!gfs_buffer_uptodate(bh[x]))
	  {
	    printk(KERN_EMERG "b_blocknr = %lu, b_size = %lu\n", BH_BLKNO(bh[x]), BH_SIZE(bh[x]));
	    ASSERT(gfs_buffer_uptodate(bh[x]));
	  }
	}

	ll_rw_block(READ, 1, &bh[x]);
	gfs_wait_on_buffer(bh[x]);
      }

      if (retries)
	printk(KERN_ERR "GFS:  retry successful in gfs_dreread()  (not an error)\n");
    }
  

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dreread()\n"););
  
  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_dwrite()
**
** Description:
**   Write a block
**
** Input(s):
**   bh - the block to write
**   flags - Flags on what to do with the write
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_dwrite(buffer_head_t *bh, int flags)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_dwrite()\n"););

  if (bh == NULL)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_dwrite()\n"););
    return(-EINVAL);
  }

  /*  Tell the Glock layer to increment the version number for this lock  */

  if (bh2pd(bh))
    bh2pd(bh)->bd_glp->gl_incr = TRUE;
  else
    printk(KERN_ERR "GFS:  Whoah!  Writing an unregistered buffer\n");

  if (flags & DW_NOTMETA)
  {
    mark_buffer_dirty(bh, 1);
  }
  else
  {
    /*printk("Me lose brain?  Uh-oh!\n");*/mark_buffer_dirty(bh, 1);
  }

  if (flags & DW_NOKEEP)
    gfs_dforget(bh);

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dwrite()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_dforget()
**
** Description:
**   Release a block, making sure it isn't kept in the cache
**
** Input(s):
**   bh - the block to release
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_dforget(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_dforget()\n"););

  if (bh == NULL)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_dforget()\n"););
    return(-EINVAL);
  }

  brelse(bh);

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_dforget()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** free_bufs()
**
** Description:
**   Free gfs_bhdata_t structures associated with this Glock, but not with
**   a buffer head.  If the Glock structure is not in use, free that too.
**
** Input(s):
**   gl - A pointer to the Glock structure holding the buffers
**
*******************************************************************************
******************************************************************************/

static void free_bufs(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  gfs_list_head_t *head, *next;
  gfs_bhdata_t *bd;
  int del = FALSE;


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering freebufs()\n"););


  GFS_SEMA_DOWN(&gl->gl_buflock);
    
  head = &gl->gl_bufs;
    
  next = head->next;
  while (1)
  {
    gfs_list_head_t *tmp = next;
      
    next = next->next;
    if (tmp == head)
      break;
    bd = gfs_list_entry(tmp, gfs_bhdata_t, bd_list);
      
    if (!bd->bd_bh)
    {
      gfs_list_del(&bd->bd_list);
      gfree(bd, sizeof(gfs_bhdata_t));
    }
  }

  if (gfs_list_empty(&gl->gl_bufs))
    del = TRUE;
      
  GFS_SEMA_UP(&gl->gl_buflock);

  if (del && gl->gl_count == 0 && !gl->gl_held)
  {
    if (gl->gl_lockdata)
      sdp->sd_lockstruct->release(sdp->sd_lockdata, &gl->gl_lockdata);
    gfs_list_del(&gl->gl_next);
    gfree(gl, sizeof(gfs_glock_t));
  }


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting freebufs()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_bufd()
**
** Description:
**   A kernel thread that wakes up every so often and gets rid of all 
**   gfs_bhdata_t structures laying around.
**
** Input(s):
**   data - a pointer to the gfs_sbd_t structure for the filesystem
**
** Returns:
**   an integer that I don't understand
**
*******************************************************************************
******************************************************************************/

int gfs_bufd(void *data)
{
  gfs_sbd_t *sdp = (gfs_sbd_t *)data;

  int x, touched;
  gfs_list_head_t *head, *next;
  gfs_glock_t *gl;
  unsigned long flags;


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_bufd()\n"););

  ASSERT(sdp);


  /*  Set up the thread  */

  lock_kernel();

  exit_mm(current);

  current->session = 1;
  current->pgrp = 1;

  spin_lock_irqsave(&current->sigmask_lock, flags);
  sigfillset(&current->blocked);
  recalc_sigpending(current);
  spin_unlock_irqrestore(&current->sigmask_lock, flags);

  current->fs->umask = 0;
  
  sprintf(current->comm, "gfs_bufd");

  unlock_kernel();


  /*  Wake up the process who started us  */

  up(&sdp->sd_bufd_lock);


  while (1)
  {
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(5 * HZ);

    IFDEBUG(DEBUG_DIO3,
	    printk(DEBUG_HOW "\n\n\n  bufd wakes up\n"););

    if (sdp->sd_bufd_stop)
    {
      IFDEBUG(DEBUG_DIO2,
	      printk(DEBUG_HOW "  bufd caught signal\n"););
      break;
    }

    IFDEBUG(DEBUG_DIO3,
	    printk(DEBUG_HOW "inoded count = %d\n", sdp->sd_id_count);
	    printk(DEBUG_HOW "gl held count = %d\n", sdp->sd_gl_heldcount););

    for (x = 0; x < GFS_GHASH_SIZE; x++)
    {
      GFS_SEMA_DOWN(&sdp->sd_gla_lock[x]);

      head = &sdp->sd_gla[x];

      next = head->next;
      while (1)
      {
	gfs_list_head_t *tmp = next;

	next = next->next;
	if (tmp == head)
	  break;
	gl = gfs_list_entry(tmp, gfs_glock_t, gl_next);

	IFDEBUG(DEBUG_DIO2,
		printk(DEBUG_HOW "  Glock %"FMT_64"u, held = %d, buffers = %d\n",
		       gl->gl_glock, gl->gl_held, !gfs_list_empty(&gl->gl_bufs)););

	touched = xchg(&gl->gl_touched, FALSE);

	if (touched)
	  free_bufs(sdp, gl);
      }

      GFS_SEMA_UP(&sdp->sd_gla_lock[x]);
    }
  }


  up(&sdp->sd_bufd_lock);


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_bufd()\n"););

  return(0);
}


/*
 *
 *
 *
 *
 *  Synchronous routines used an mount/unmount time
 *
 *
 *
 *
 */


/******************************************************************************
*******************************************************************************
**
** gfs_sread()
**
** Description:
**   Read a block from disk.  Don't try to link it into the Glock array.
**
**   THIS WILL HOSE EXISTING BUFFERS IN THE CACHE, SO BE CAREFUL!!!
**
** Input(s):
**   sdp - contains the device the block lives on
**   bbno - the block number
**   bytes - number of bytes in the block
**
** Returns:
**   The read block on success, NULL on failure
**
*******************************************************************************
******************************************************************************/

buffer_head_t *gfs_sread(gfs_sbd_t *sdp, uint64 bbno, uint32 bytes)
{
  uint32 size;
  buffer_head_t *bh;


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_sread()\n"););


  if (bytes > PAGE_SIZE)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_sread()\n"););
    return(NULL);
  }

  for (size = 512; size < bytes; size <<= 1) ;

  set_blocksize(sdp->sd_vsb->s_dev, size);
  bh = getblk(sdp->sd_vsb->s_dev, bbno, size);
  if (!bh)
  {
    IFDEBUG(DEBUG_DIO1,
	    printk(DEBUG_HOW "Exiting gfs_sread()\n"););
    return(bh);
  }

  mark_buffer_uptodate(bh, FALSE);

  ll_rw_block(READ, 1, &bh);
  gfs_wait_on_buffer(bh);

  if (!buffer_uptodate(bh))
  {
    gfs_sforget(bh);
    bh = NULL;
  }


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_sread()\n"););

  return(bh);
}


/******************************************************************************
*******************************************************************************
**
** gfs_swrite()
**
** Description:
**   Write a block.  Do it synchrnonously.  Don't worry about the Glock array.
**
** Input(s):
**   bh - the block to write
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_swrite(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_swrite()\n"););


  mark_buffer_dirty(bh, 1);

  ll_rw_block(WRITE, 1, &bh);
  gfs_wait_on_buffer(bh);

  ASSERT(!buffer_dirty(bh));


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_swrite()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_sforget()
**
** Description:
**   Release a block, making sure it isn't kept in the cache
**
** Input(s):
**   bh - the block to release
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_sforget(buffer_head_t *bh)
{
  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_sforget()\n"););


  if (!bh)
    return(-EINVAL);


  mark_buffer_uptodate(bh, FALSE);
  brelse(bh);


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_sforget()\n"););

  return(0);
}
