/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: file.c
**
*******************************************************************************
******************************************************************************/

#include <linux/locks.h>

#include "gfs.h"
#include "gfs_ioctl.h"
#include "blklist.h"
#include "bmap.h"
#include "dio.h"
#include "dir.h"
#include "glock.h"
#include "file.h"
#include "file_linux.h"
#include "inode.h"
#include "gfs_profile.h"
#include "fs.h"
#include "super.h"


/******************************************************************************
*******************************************************************************
**
** gfs_read()
**
** Description:
**   Read bytes from a file
**
**   Called by the VFS layer.  Glock the dinode and call gfs_readi()
**
** Input(s):
**   filp - the file to read from
**   buf - the buffer to load the data into
**   size - the length of data to read
**   offset - read from this offset in the file
**
** Output(s):
**   offset - updated by the number of bytes read
**
** Returns:
**   The number of bytes read, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

ssize_t gfs_read(struct file *filp, char *buf, size_t size, loff_t *offset)
{
  gfs_inode_t *ip = vi2ino(filp->f_dentry->d_inode);
  gfs_sbd_t *sdp = ip->i_sbd;

  int error;
  int64 count;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Entering gfs_read()\n"););


  if ((error = gfs_glock(sdp, ip->i_di.di_glock, ip, GL_SHARED)))
  {
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_read()\n"););
    return(error);
  }

  if (!access_ok(VERIFY_WRITE, buf, size))
  {
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_read()\n"););
    return(-EFAULT);
  }

  count = gfs_readi(ip, buf, *offset, size, FALSE);

  gfs_gunlock(sdp, ip->i_di.di_glock, 0);

  if (count > 0)
    *offset += count;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Exiting gfs_read()\n"););

  return(count);
}


/******************************************************************************
*******************************************************************************
**
** gfs_write()
**
** Description:
**   Write bytes to a file
**
**   Called by the VFS layer.  Glock the dinode and call gfs_writei()
**
** Input(s):
**   filp - the file to write to
**   buf - the buffer to write the data from
**   size - the length of data to written
**   offset - write to this offset in the file
**
** Output(s):
**   offset - updated by the number of bytes written
**
** Returns:
**   The number of bytes written, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

ssize_t gfs_write(struct file *filp, const char *buf, size_t size, loff_t *offset)
{
  gfs_inode_t *ip = vi2ino(filp->f_dentry->d_inode);
  gfs_sbd_t *sdp = ip->i_sbd;

  int error;
  int64 count;
  uint64 pos;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Entering gfs_write()\n"););


  if ((error = gfs_glock(sdp, ip->i_di.di_glock, ip, 0)))
  {
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_write()\n"););
    return(error);
  }


  if (filp->f_flags & O_APPEND)
    pos = ip->i_di.di_bsize;
  else
    pos = *offset;

  if (!access_ok(VERIFY_READ, buf, size)) {
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_writei()\n"););
    return(-EFAULT);
  }


  count = gfs_writei(ip, buf, pos, size, FALSE);

  gfs_gunlock(sdp, ip->i_di.di_glock, 0);


  if (count > 0)
    *offset = pos + count;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Exiting gfs_write()\n"););

  return(count);
}


/******************************************************************************
*******************************************************************************
**
** gfs_readdir()
**
** Description:
**   Read directory entries from a directory
**
** Input(s):
**   filp - the directory to read from
**   dirent - copy the dents to here
**   filldir - use this function to copy the directories
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
  gfs_inode_t *dip = vi2ino(filp->f_dentry->d_inode);
  gfs_sbd_t *sdp = dip->i_sbd;

  int error;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Entering gfs_readdir()\n"););


  if ((error = gfs_glock(sdp, dip->i_di.di_glock, dip, GL_SHARED)))
  {
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_readdir()\n"););
    return(error);
  }

  error = gfs_dir_trans(dip, (int64 *)&(filp->f_pos), dirent, filldir);

  gfs_gunlock(sdp, dip->i_di.di_glock, 0);


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Exiting gfs_readdir()\n"););

  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_show_subpools()
**
** Description:
**   Perform I/O control operations on a filesystem and report subpool status
**   i.e. the Number of Blocks free, Total Number of Blocks available,
**   and the Number of Dinodes on the subpool.
**
** Input(s):
**   inode - the inode.
**   ip    - gfs inode.
**   arg   - the argument.
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/
int gfs_show_subpools(struct inode *inode, gfs_inode_t *ip, unsigned long arg)
{
  gfs_sbd_t 		*sdp ;
  gfs_rgrpd_t 		*rgd;
  struct super_block 	*sb;
  int 			error = 0, i;
  gfs_spstatlist_t 	spstatlistk;
  gfs_spstat_t * 	sp_listk;
  uint32 		sp, num_sp;
  

  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Entering ioctl call - GFS_SHOW_SUBPOOLS \n"););
  
  num_sp = 0;
  sb = inode->i_sb;
  
  if (!sb || !(sdp = ip->i_sbd))
    return(-EIO);

  
  
  if(copy_from_user(&spstatlistk,(gfs_spstatlist_t *)arg,sizeof(gfs_spstatlist_t)))
    return(-EFAULT);
  
  sp_listk = (gfs_spstat_t *)gmalloc(spstatlistk.sp_cnt * sizeof(gfs_spstat_t));
  


  if(sp_listk)
    memset(sp_listk, 0, spstatlistk.sp_cnt * sizeof(gfs_spstat_t));
  else
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Exiting gfs_ioctl ()\n"););
      return(-ENOMEM);
    }
     


  for (i = 0; i < sdp->sd_sb.sb_rgrps; i++)
    {
      rgd = &sdp->sd_rgd[i];
      
      if (gfs_glock(sdp, rgd->rd_ri.ri_glock, NULL, GL_SHARED) != 0)
	continue;


      /*  Read/Reread the RG block  */
      if ((error = gfs_rgrp_read(sdp, rgd)))
	{
	  gfs_gunlock(sdp, rgd->rd_ri.ri_glock,0);
	  gfree(sp_listk, spstatlistk.sp_cnt * sizeof(gfs_spstat_t));
	      
	  IFDEBUG(DEBUG_FILE1,
		  printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););
	  return(error);
	}


      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "%.4u:  %"FMT_64"u\t%"FMT_64"u\t%"FMT_64"u\n",
		     i,
		     rgd->rd_rg.rg_dinodes,
		     rgd->rd_ri.ri_data - rgd->rd_rg.rg_free,
		     rgd->rd_ri.ri_data););
	
      sp = rgd->rd_ri.ri_subpool;


      if(sp < spstatlistk.sp_cnt)
	{
	  sp_listk[sp].blocks += rgd->rd_ri.ri_data;
	  sp_listk[sp].dinds += rgd->rd_rg.rg_dinodes;
	  sp_listk[sp].bfree += rgd->rd_rg.rg_free;
	}


      if(sp > num_sp) 
	num_sp = sp;

      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
    }

    
  spstatlistk.sp_cnt = num_sp + 1;



  if(copy_to_user(spstatlistk.sp_list,sp_listk,(spstatlistk.sp_cnt) * sizeof(gfs_spstat_t)))
    {
      gfree(sp_listk, spstatlistk.sp_cnt * sizeof(gfs_spstat_t));
      return(-EFAULT);
    }


  if(copy_to_user((gfs_spstatlist_t *)arg, &spstatlistk,sizeof(gfs_spstatlist_t)))
    {
      gfree(sp_listk, spstatlistk.sp_cnt * sizeof(gfs_spstat_t));
      return(-EFAULT);
    }


  gfree(sp_listk, spstatlistk.sp_cnt * sizeof(gfs_spstat_t));
  return(error);
}  


/******************************************************************************
*******************************************************************************
**
** gfs_which_subpool()
**
** Description:
**   Perform I/O control operations on a fileor directory and report the
**   file block statistics i.e. Number of Blocks in each subpool.
**
** Input(s):
**   inode - the inode.
**   ip    - gfs inode.
**   arg   - the argument.
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/
int gfs_which_subpool(struct inode *inode, gfs_inode_t *ip, unsigned long arg)
{
  gfs_sbd_t 	       *sdp ;
  struct super_block   *sb;
  int 			error = 0;

  uint64 		*blksk, size, rqsize, offset, dr_no;
  gfs_direct_t 	       *direct = NULL;
  gfs_fspstatlist_t 	fspstatlistk; 
  uint32 		rgrp, num_sp;
  gfs_rindex_t 	       *ri = NULL;
  int 			j;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Entering ioctl call - GFS_WHICH_SUBPOOL \n"););


  num_sp = 0;
  sb = inode->i_sb;


  if (!sb || !(sdp = ip->i_sbd))
    return(-EIO);


  if(copy_from_user(&fspstatlistk,(gfs_fspstatlist_t *)arg,sizeof(gfs_fspstatlist_t)))
    return(-EFAULT);
  
  blksk =(uint64 *)gmalloc(fspstatlistk.fsp_cnt * sizeof(uint64));


  if(blksk)
    memset(blksk, 0, fspstatlistk.fsp_cnt * sizeof(uint64));
  else
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););

      return(-ENOMEM);
    }

  if ((error = gfs_glock(sdp, ip->i_di.di_glock, ip, GL_SHARED)))
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););

      gfree(blksk, fspstatlistk.fsp_cnt * sizeof(uint64));

      return(error);
    }



  size = ip->i_di.di_bsize;
  offset = 0;

  while(size > 0)
    {

      /* If the size is greater than 4MB, the request has to be broken */
      if(size > 4*MB)
	rqsize = 4*MB;
      else
	rqsize = size;


      if ((error = gfs_bmapi_read(ip, offset, rqsize, &direct)))
	{
	  gfs_gunlock(sdp, ip->i_di.di_glock, 0);
	  IFDEBUG(DEBUG_FILE1,
		  printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););
	  gfree(blksk, fspstatlistk.fsp_cnt * sizeof(uint64));
	  return(error);
	}


      size -= rqsize;
      offset += rqsize;


      if (!direct)
	{
	  gfs_gunlock(sdp, ip->i_di.di_glock, 0);
	  IFDEBUG(DEBUG_FILE1,
		  printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););
	  gfree(blksk, fspstatlistk.fsp_cnt * sizeof(uint64));
	  return(-1);
	}


      if (direct[0].dr_flags & GFS_BMAP_STUFFED_DINODE)
	ASSERT(direct[0].dr_num == 1);


      for (j = 0; j < direct[0].dr_num; j++)
	{
	  if (direct[j].dr_flags & GFS_BMAP_HOLE)
	    continue;

	  if(direct[j].dr_flags & GFS_BMAP_STUFFED_DINODE)
	    dr_no = ip->i_num;
	  else
	    dr_no = direct[j].dr_no;

	  for (rgrp = 0; rgrp < sdp->sd_sb.sb_rgrps; rgrp++)
	    {
	      ri = &sdp->sd_rgd[rgrp].rd_ri;

	      if(dr_no < ri->ri_data1 + ri->ri_data)
		{ 

		  if(ri->ri_subpool < fspstatlistk.fsp_cnt)
		    blksk[ri->ri_subpool]++;
		  
		  if(ri->ri_subpool > num_sp)
		    num_sp = ri->ri_subpool;
		  
		  break;
		}
	    }
	}

      gfree(direct, direct[0].dr_num * sizeof(gfs_direct_t));
    }



  /* Release the inode lock */
  gfs_gunlock(sdp, ip->i_di.di_glock, 0);



  /*
   * this is sp with the highest number .. so the number of subpools are 
   * num_sp + 1 
   */
  fspstatlistk.fsp_cnt = num_sp + 1; 
    

  if(copy_to_user(fspstatlistk.blks,blksk,(fspstatlistk.fsp_cnt) * sizeof(uint64)))
    {
      gfree(blksk, fspstatlistk.fsp_cnt * sizeof(uint64));
      return(-EFAULT);
    }


  if(copy_to_user((gfs_fspstatlist_t *)arg, &fspstatlistk,sizeof(gfs_fspstatlist_t)))
    {
      gfree(blksk, fspstatlistk.fsp_cnt * sizeof(uint64));
      return(-EFAULT);
    }


  gfree(blksk, fspstatlistk.fsp_cnt * sizeof(uint64));

  return(error);
}


/******************************************************************************
*******************************************************************************
**
** init_new_dinode
**
** Description:
**   Used internally by the GFS_MOVE_TO_SUBPOOL ioctl call. It initialises 
**   the new dinode structure taking the old dinode structure as a parameter.
**   
**  NOTE: Change this function if new fields are added to the dinode 
**	  structure.
**
** Input(s):
**   oldd 	- old dinode.
**   sdp  	- superblock.
**   blkno	- block number.
**   rgrp 	- the resource group where the new dinode lies.
**
** Output:
**   newd - new dinode structure after initialization.
**
** Returns:
**   void
**
*******************************************************************************
******************************************************************************/

void init_new_dinode(gfs_sbd_t *sdp, gfs_dinode_t * newd, gfs_dinode_t * oldd,uint64 blkno, uint32 rgrp)
{

  /* most of the fields are going to remain the same */
  memcpy((char *)newd,(char *)oldd,sizeof(gfs_dinode_t));


  /* overwrite the changes -- check the fields*/
  
  newd->di_addr = blkno;
  newd->di_blocks = 1;

  newd->di_rgrp = rgrp;

  newd->di_glock = gfs_ino2gl(sdp, blkno);
  
  newd->di_atime = CURRENT_TIME; 

  newd->di_regime = GFS_DIR_LINEAR; /* is this correct?? */

  newd->di_height = 0;
  newd->di_bsize = 0;

  newd->di_depth = 0;
  newd->di_entries = 0;

  newd->di_rgrp_meta = newd->di_rgrp_data  = rgrp;
  newd->di_last_meta = newd->di_last_data = blkno - sdp->sd_rgd[rgrp].rd_ri.ri_data1;

  /* sanity check */
  ASSERT(newd->di_nopen == 1);

  return;
}


/******************************************************************************
*******************************************************************************
**
** check_for_holes
**
** Description:
**   Used internally by the GFS_MOVE_TO_SUBPOOL ioctl call. It check the
**   original file for holes.
**   
**  NOTE: Change this function if new fields are added to the dinode 
**	  structure.
**
** Input(s):
**  ip  -  the file.
**
** Returns:
**   1  -  if the file contains a hole.
**   0  -  No hole.
** EXXX -  Error occurred while processing.
**
*******************************************************************************
******************************************************************************/

int check_for_holes(gfs_inode_t *ip)
{

  int 			error = 0, flag = 0;
  uint64 		size, rqsize, offset, j;
  gfs_direct_t 	       *direct = NULL;


  size = ip->i_di.di_bsize;
  offset = 0;

  while((size > 0) && (flag == 0))
    {
      
      /* If the size is greater than 4MB, the request has to be broken */
      if(size > 4*MB)
	rqsize = 4*MB;
      else
	rqsize = size;

      
      if ((error = gfs_bmapi_read(ip, offset, rqsize, &direct)))
	return(error);


      size -= rqsize;
      offset += rqsize;


      if (!direct)
	return(0);

      if (direct[0].dr_flags & GFS_BMAP_STUFFED_DINODE)
	ASSERT(direct[0].dr_num == 1);


      for (j = 0; j < direct[0].dr_num; j++)
	{
	  if (direct[j].dr_flags & GFS_BMAP_HOLE)
	    {
	      flag = 1;
	      break;
	    }
	}
      gfree(direct, direct[0].dr_num * sizeof(gfs_direct_t));

    }

  return(flag);
}


/******************************************************************************
*******************************************************************************
**
** gfs_mv_to_subpool()
**
** Description:
**   Move a file to a given Resource Group.
**
** Input(s):
**   inode - the inode.
**   ip    - gfs inode.
**   arg   - the argument.
**
** Returns:
**   0 on success, -EXXXX on failure
**
** PS: It returns -EPERM if the file contains a GFS_BMAP_HOLE.
**
*******************************************************************************
******************************************************************************/
int gfs_mv_to_subpool(struct inode *inode, gfs_inode_t *ip, unsigned long arg)
{
  gfs_sbd_t 		*sdp ;
  struct super_block 	*sb;
  int 			error, destroy = 1; 
  gfs_lnkinfo_t 	lnkinfok;
  gfs_lnkname_t 	*lnk_names;
  char	 	  	*buf, name[GFS_MAXPATHLEN];
  int 		  	j,k,/*second=0,*/pos,nread,ncopied,ino;
  uint64 		temp;
  struct dentry  	*ldentry;
  gfs_inode_t   	**dip=NULL; /* store the inodes of the directories */
  gfs_inode_t 		n_inode;    /* new temporary inode */
  gfs_blklist_t 	*blk,*fblk;
  gfs_dinode_t  	di;
  buffer_head_t 	*bh;
  struct qstr 		lnkname;
  uint32 		rgrp, rgrpctr=0;
  gfs_lockop_t          *lock;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Entering ioctl call - GFS_MOVE_TO_SUBPOOL \n"););
  

  /* Error if the request is for a directory */
  if(ip->i_di.di_type ==  FILE_DIR)
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Is a Directory::Exiting gfs_ioctl()\n"););
      return(-EISDIR);
    }


  sb = inode->i_sb;
  error = 0; /* Initialise it as success */


  if (!sb || !(sdp = ip->i_sbd))
    return(-EIO);

  
  if (copy_from_user(&lnkinfok,(gfs_lnkinfo_t *)arg,sizeof(gfs_lnkinfo_t)))
    return(-EFAULT);



  lnk_names = (gfs_lnkname_t *)gmalloc(lnkinfok.lnk_cnt *  
				       sizeof(gfs_lnkname_t));
  if(lnk_names)
    memset(lnk_names, 0,lnkinfok.lnk_cnt * sizeof(gfs_lnkname_t));  
  else
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););
      return(-ENOMEM);
    } 



  if(copy_from_user(lnk_names,(gfs_lnkname_t *)lnkinfok.l_name,
		    sizeof(gfs_lnkname_t) * lnkinfok.lnk_cnt))
    {
      error = -EFAULT;
      goto free_lnk_names;
    }


  dip = (gfs_inode_t **) gmalloc(lnkinfok.lnk_cnt *  
				   sizeof(gfs_inode_t *));
  if(!dip)
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););
      error = -ENOMEM;
      goto free_lnk_names;
    }



  lock = (gfs_lockop_t *)gmalloc( (1+lnkinfok.lnk_cnt)*sizeof(gfs_lockop_t)); 
  if(lock)
    memset(lock, 0, (2 + lnkinfok.lnk_cnt) * sizeof(gfs_lockop_t));
  else
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););
      error = -ENOMEM;
      goto free_dip;
    }



  /* all locks are exclusive so flag = 0 */
  lock[0].lock = ip->i_di.di_glock;
  lock[0].ip   = ip;



  /* 
   * Verify each directory and prepare to get locks on them 
   */
    
  for( j = 0; j < lnkinfok.lnk_cnt; j++)
    {
      /* 
       *  For each directory - 
       *  1. Use lookup to get the Linux dentry to the directory.
       *  2. Get lock on the directory
       */
      
      ldentry = lookup_dentry(lnk_names[j].dname, NULL ,0);

      if(IS_ERR(ldentry))
	{
	  error = ENOENT;
	  
	  IFDEBUG(DEBUG_FILE1,
		  printk(DEBUG_HOW "Directory not found :: Exiting gfs_ioctl()\n"););
	  goto free_dip;
	}   

      /* Get the GFS specific inode pointer */
      dip[j] = vi2ino(ldentry->d_inode);
      
      /* Free the dcache entry */
      dput(ldentry);

      /* lock for the directory inode */
      lock[j+1].lock = dip[j]->i_di.di_glock;
      lock[j+1].ip   = dip[j];
    }

  

  /* 
   * Get a new dinode block in the required Subpool. 
   */
  
  for (rgrp = 0; rgrp < sdp->sd_sb.sb_rgrps; rgrp++)
    {
      if(sdp->sd_rgd[rgrp].rd_ri.ri_subpool != lnkinfok.d_sp)
	continue;

      error = gfs_blkdalloc(dip[0], &blk, rgrp, 0xFFFFFFFFFFFFFFFFll, 0);

      if(error != 0)  
	{
	  IFDEBUG(DEBUG_FILE1,
		  printk(DEBUG_HOW "gfs_blkdalloc failed :: Exiting gfs_ioctl()\n"););
	  goto free_dip;
	}
      else
	{

	  /* ***************************************************************
	   * Depending on the way gfs_blkdalloc is written (i.e. whether it
	   * gaurantees the destination resource group (fails if it cannot) 
	   * OR it just initialises it as a starting point for the search 
	   * *************************************************************** */
	  
	  for (rgrpctr = 0; rgrpctr < sdp->sd_sb.sb_rgrps; rgrpctr++) 
	    if (blk->blk_no < sdp->sd_rgd[rgrpctr].rd_ri.ri_data1 + sdp->sd_rgd[rgrpctr].rd_ri.ri_data) 
	      break; /* break out of the inner for loop */
	  
	  if(sdp->sd_rgd[rgrpctr].rd_ri.ri_subpool == lnkinfok.d_sp) 
	    break; /* break out of the outer for loop */ 
	  else /* Oh !! gfs_blkdalloc allocated the blocks in the wrong SP */ 
	    { 
	      gfs_blkfree(sdp, &blk); 
	      if (blk) 
		{ 
		  GARBAGE_BLOCKS(); 
		  gfs_blklist_free(blk); 
		} 
	      rgrp = rgrpctr; /* no point in trying the RGs in between */ 
	    } 
	}

    } 


  if (rgrp >= sdp->sd_sb.sb_rgrps)
    {
      error = -EADDRNOTAVAIL;
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Unable to allocate the dinode block in the required Subpool :: Exiting gfs_ioctl()\n"););
      goto free_dip;
    }


  lock[j+1].lock = gfs_ino2gl(sdp, blk->blk_no);
  lock[j+1].ip   = NULL;



  /* 
   * Get the locks in a sorted order 
   */

  if ( (error = gfs_glock_m(sdp, (2 + lnkinfok.lnk_cnt), lock)) )
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Unable to obtain the locks :: Exiting gfs_ioctl()\n"););

      goto free_new_d_blk;
    }


  /* Check the number of links */
  if(inode->i_nlink != lnkinfok.lnk_cnt )
    {
      error = -EINVAL ;
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Number of Links don't match :: Exiting gfs_ioctl()\n"););

      goto unlock_inodes;
    }



  /* 
   * Does anyone else have this file opened ? (its okay if the directory cache
   * has it open, therefore i_count == lnkinfok.lnk_cnt  is fine)) 
   */
  
  if( (inode->i_count > lnkinfok.lnk_cnt ) || (ip->i_di.di_nopen > 1))
    {	
      error = -EBUSY ;
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "File opened by other hosts/processes :: Exiting gfs_ioctl()\n"););

      goto unlock_inodes;
    }	



  /*
   * Verify the source and the destination subpools
   */
  
  if(lnkinfok.d_sp == ip->i_di.di_rgrp) 
    { 
      IFDEBUG(DEBUG_FILE1, 
	      printk(DEBUG_HOW "Nothing to be done !\n");); 
      
      goto unlock_inodes; 
    } 
    


  /*
   * Check for GFS_BMAP_HOLE's
   */

  if((error = check_for_holes(ip)))
    {
      IFDEBUG(DEBUG_FILE1, 
	      printk(DEBUG_HOW "File contains a GFS_BMAP_HOLE \n");); 

      if(error == 1)
	error = -EPERM;

      goto unlock_inodes; 
    } 



  /* 
   * Verify each link. 
   */
    
  for( j = 0; j < lnkinfok.lnk_cnt; j++)
    {
      /* 
       *  Call gfs_dir_search and compare the returned inode number
       *  to the input inode number .. if error - rollback and 
       *  return error  
       */
      
      memset(name, 0, GFS_MAXPATHLEN);
      memcpy(name, lnk_names[j].fname, lnk_names[j].flen);
      lnkname.len = lnk_names[j].flen;
      lnkname.name = name;
      
      ino = gfs_dir_search(dip[j],&lnkname,&temp);
      
      /* search successful ? */
      if(ino < 0)
	{
	  error = ino;
	  IFDEBUG(DEBUG_FILE1,
		  printk(DEBUG_HOW "gfs_dir_search failed :: Exiting gfs_ioctl()\n"););
	  goto unlock_inodes;
	}


      /* check if the inodes are the same */
      if(temp != ip->i_num)
	{
	  error = -EINVAL;
	  IFDEBUG(DEBUG_FILE1,
		  printk(DEBUG_HOW "Invalid Link :: Exiting gfs_ioctl()\n"););

	  goto unlock_inodes;
	}
    }	




  bh = gfs_dgetblk(sdp, blk->blk_no, gfs_ino2gl(sdp, blk->blk_no));
 
  if (!bh)
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "gfs_dgetblk failed : Exiting gfs_ioctl()\n"););
    
      error = -EIO;

      goto unlock_inodes;
    }


  gfs_mark_uptodate(bh, TRUE);
  memset(BH_DATA(bh), 0, sdp->sd_sb.sb_bsize);

  init_new_dinode(sdp, &di, &ip->i_di, blk->blk_no, rgrpctr);

  gfs_dinode_out(&di, BH_DATA(bh));
  gfs_dwrite(bh, 0); /* Don't destroy the buffer head */


  /* Make a new inode for creating the new copy */
  memcpy((char *)&n_inode,(char *)ip, sizeof(gfs_inode_t));
  n_inode.i_num = blk->blk_no;

  memcpy((char *)&n_inode.i_di, (char *) &di, sizeof(gfs_dinode_t));
  n_inode.i_didata = bh;
    
  GFS_SEMA_INIT(&n_inode.i_sema);



  /* 
   *  Start copying the blocks - read from the old inode, write to the 
   *  new inode 
   */
    
  buf = (char *) gmalloc(MB * sizeof(char));
  if (!buf)
    {
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Unable to allocate memory :: Exiting gfs_ioctl()\n"););
      error = -ENOMEM;
      goto free_new_file;
    }

  for(pos = nread = ncopied = 0; pos < ip->i_di.di_bsize; pos += ncopied)
    {
      /* 
       * If less than a MB of data is left, gfs_readi reads only 
       * till the end of file 
       */

      nread = gfs_readi(ip, buf, pos, MB, TRUE);

      if(nread < 0)
	{
	  /* handle the error */
	  error = nread;
	  IFDEBUG(DEBUG_FILE1,
		  printk(DEBUG_HOW "Read failed :: Exiting gfs_ioctl()\n"););
	  gfree(buf, MB * sizeof(char));

	  goto free_new_file;
	}

      ncopied = gfs_writei(&n_inode, buf, pos, nread, TRUE);

      if(ncopied != nread)
	{
	  /* handle the error */
	  error = ncopied;
	  IFDEBUG(DEBUG_FILE1,
		  printk(DEBUG_HOW "Write Failed :: Exiting gfs_ioctl()\n"););
	  gfree(buf, MB * sizeof(char));

	  goto free_new_file;
	}  
    }

  gfree(buf, MB * sizeof(char));




  /* 
   * Change the links 
   */
  
  for(k = 0; k < lnkinfok.lnk_cnt; k++)
    {

      /* 
       * For each directory containing the link - change the directory entry 
       * (inode number) using gfs_dir_mvino(). And increment the version.
       */

      memset(name,0,GFS_MAXPATHLEN);
      memcpy(name, lnk_names[k].fname, lnk_names[k].flen);
      lnkname.len = lnk_names[k].flen;
      lnkname.name = name;
      
      if( (error = gfs_dir_mvino(dip[k], &lnkname, blk->blk_no))  )
	{
	  printk("Unable to change the parent directory entry :: Exiting gfs_ioctl()\n");
	  goto free_new_file;
	}
    }



  /*
   * At this stage we do not want to remove the new file even if something 
   * goes wrong
   */
  destroy = 0;



  /* 
   * Remove the old copy of the data and change the inode.
   */
  
  fblk = gfs_difree(ip);
  
  /*  Zero out the dinode magic number  */
  ip->i_di.di_magic = 0;
    
  gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));

  gfs_dwrite(ip->i_didata, 0);  


  if (fblk)
    {
      gfs_blkfree(sdp, &fblk);
      if (fblk)
	{
	  GARBAGE_BLOCKS();
	  gfs_blklist_free(fblk);
	}
    }

  /* free the bufferhead of the old file */
  gfs_dforget(ip->i_didata);




  /* 
   * Overwrite the inode with the new info 
   */

  gfs_ihashdel(ip);
  ip->i_num = blk->blk_no;
  ip->i_vn = 0;
  memcpy((char *)&ip->i_di, (char *) &n_inode.i_di, sizeof(gfs_dinode_t));
  ip->i_didata = bh;
  ip->i_vino->i_ino = n_inode.i_num;

  /* Write dinode back out  */
  gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
  gfs_dwrite(ip->i_didata, 0);
  gfs_inode_attr_in(ip);




  /* 
   * We are done -- Now clean up 
   */

  gfs_blklist_free(blk);


 free_new_file:
  if(destroy)
    {
      fblk = gfs_difree(&n_inode);
  
      /*  Zero out the dinode magic number  */
      n_inode.i_di.di_magic = 0;
    
      gfs_dinode_out(&n_inode.i_di, BH_DATA(n_inode.i_didata));

      gfs_dwrite(n_inode.i_didata, 0);  


      if (fblk)
	{
	  gfs_blkfree(sdp, &fblk);
	  if (fblk)
	    {
	      GARBAGE_BLOCKS();
	      gfs_blklist_free(fblk);
	    }
	}

      /* free the bufferhead of the old file */
      gfs_dforget(ip->i_didata);
      destroy = 0;
    }


 unlock_inodes:
  gfs_gunlock_m(sdp, (2 + lnkinfok.lnk_cnt), lock);


 free_new_d_blk:
  if(destroy)
    {
      gfs_blkfree(sdp, &blk);
      if (blk)
	{
	  GARBAGE_BLOCKS();
	  gfs_blklist_free(blk);
	}
    }


 free_dip:
  gfree(dip, lnkinfok.lnk_cnt * sizeof(gfs_inode_t *));

 free_lnk_names:   
  gfree(lnk_names, lnkinfok.lnk_cnt * sizeof(gfs_lnkname_t));

  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_ioctl()
**
** Description:
**   Perform I/O control operations on a file or directory
**   Currently the following ioctls are supported:
**	GFS_DEBUG_ON	    - Add debug flags.
**	GFS_DEBUG_OFF       - Remove debug flags.
**	GFS_SHOW_SUBPOOLS   - Report subpool status i.e. the Number of
**			      Blocks free, Total Number of Blocks available,
**			      and the Number of Dinodes on the subpool.
**	GFS_WHICH_SUBPOOL   - Report file block statistics i.e. Number of
** 			      Blocks in each subpool.
**	GFS_MOVE_TO_SUBPOOL - Move a file to a given Resource Group.
**
** Input(s):
**   inode - the inode
**   filp - the file pointer
**   cmd - the ioctl command
**   arg - the argument
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
#ifdef DEBUG
  uint64 flags;
#endif  /*  DEBUG  */
  
  gfs_inode_t 		*ip = vi2ino(inode);
  int 			error = 0;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Entering gfs_ioctl()\n"););

  switch (cmd)
  {
  case GFS_DEBUG_ON:
#ifdef DEBUG
    if (copy_from_user(&flags, (uint64 *)arg, sizeof(uint64)))
      return(-EFAULT);

    printk(DEBUG_HOW "GFS:  debug on (0x%.16"FMT_64"X + 0x%.16"FMT_64"X = 0x%.16"FMT_64"X)\n",
	   GFS_dflags, flags, GFS_dflags | flags);

    GFS_dflags |= flags;
#else  /*  DEBUG  */
    printk(KERN_ERR "GFS:  not compiled with debugging\n");
#endif  /*  DEBUG  */
    break;

  case GFS_DEBUG_OFF:
#ifdef DEBUG
    if (copy_from_user(&flags, (uint64 *)arg, sizeof(uint64)))
      return(-EFAULT);

    printk(DEBUG_HOW "GFS:  debug off (0x%.16"FMT_64"X - 0x%.16"FMT_64"X = 0x%.16"FMT_64"X)\n",
	   GFS_dflags, flags, GFS_dflags & ~flags);

    GFS_dflags &= ~flags;
#else  /*  DEBUG  */
    printk(KERN_ERR "GFS:  not compiled with debugging\n");
#endif  /*  DEBUG  */
    break;

  case GFS_SHOW_SUBPOOLS: /* returns filesystem subpool information */
    error = gfs_show_subpools(inode, ip, arg);
    break;

  case GFS_WHICH_SUBPOOL:   /* returns file subpool information */
    error = gfs_which_subpool(inode, ip, arg);
    break;

  case GFS_MOVE_TO_SUBPOOL:	
    error = gfs_mv_to_subpool(inode, ip, arg);
    break;

    case GFS_DUMPSTATS:
#ifdef PROFILE
      if (!arg)
	return NR_FUNCTIONS;

      if (copy_to_user((uint64 *)arg,  gfs_prof, NR_FUNCTIONS * sizeof(gfs_profile_t)))
	return (-EFAULT);
#endif
      break;

  default:
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););
    return(-ENOTTY);
    break;
  }

  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Exiting gfs_ioctl()\n"););

  return(error);
}





struct file_operations gfs_file_fops =
{
  NULL,  /*  llseek  */
  gfs_read,
  gfs_write,
  NULL,  /*  readdir  */
  NULL,  /*  poll  */
  gfs_ioctl,
  generic_file_mmap,
  NULL,  /*  open  */
  NULL,  /*  flush  */
  NULL,  /*  release  */
  NULL,  /*  fsync  */
  NULL,  /*  fasync  */
  NULL,  /*  check_media_change  */
  NULL,  /*  frevalidate  */
  NULL   /*  lock  */
};


struct file_operations gfs_dir_fops =
{
  NULL,  /*  llseek  */
  NULL,  /*  read  */
  NULL,  /*  write  */
  gfs_readdir,
  NULL,  /*  poll  */
  gfs_ioctl,
  NULL,  /*  mmap  */
  NULL,  /*  open  */
  NULL,  /*  flush  */
  NULL,  /*  release  */
  NULL,  /*  fsync  */
  NULL,  /*  fasync  */
  NULL,  /*  check_media_change  */
  NULL,  /*  frevalidate  */
  NULL   /*  lock  */
};


