/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: file.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "bmap.h"
#include "dio.h"
#include "file.h"
#include "inode.h"


#define GFS_MAX_TRANSFER (1048576)



/******************************************************************************
*******************************************************************************
**
** gfs_readi()
**
** Description:
**   Read bytes from a file
**
** Input(s):
**   ip - the file to read from
**   buf - the buffer to load the data into
**   size - the length of data to read
**   offset - read from this offset in the file
**   internal - if true, use memcpy()s instead of copy_from_user
**
** Output(s):
**   offset - updated by the number of bytes read
**
** Returns:
**   The number of bytes read, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int64 gfs_readi(gfs_inode_t *ip, void *buf, uint64 offset, uint64 size, int internal)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_direct_t *direct = NULL, *dtmp;
  buffer_head_t **bh;
  
  int i, j;
  int error;
  int64 count = 0;
  uint32 transfer = GFS_MAX_TRANSFER / sdp->sd_sb.sb_bsize;
  int retries;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Entering gfs_readi()\n"););

  IFDEBUG(DEBUG_FILE2,
	  printk(DEBUG_HOW "  off = %ld, size = %lu\n",
		 (unsigned long)offset, (unsigned long)size););


  if ((error = gfs_bmapi_read(ip, offset, size, &direct)))
  {
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_readi()\n"););
    return(error);
  }


  /*  Handle a transfer of zero bytes gracefully  */

  if (!direct)
  {
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_readi()\n"););    
    return(0);
  }


  IFDEBUG(DEBUG_FILE3,
	  for (i = 0; i < direct->dr_num; i++)
	  {
	    printk(DEBUG_HOW "  direct #%d\n", i);
	    printk(DEBUG_HOW "    dr_num = %d\n", direct[i].dr_num);
	    printk(DEBUG_HOW "    dr_flags = %d\n", direct[i].dr_flags);
	    printk(DEBUG_HOW "    dr_no = %"FMT_64"u\n", direct[i].dr_no);
	    printk(DEBUG_HOW "    dr_start = %d\n", direct[i].dr_start);
	    printk(DEBUG_HOW "    dr_len = %d\n", direct[i].dr_len);
	  });


  if (direct[0].dr_flags & GFS_BMAP_STUFFED_DINODE)
  {
    dtmp = &direct[0];

    ASSERT(dtmp->dr_num == 1);

    if (internal)
      memcpy((char *)buf + count, BH_DATA(ip->i_didata) + dtmp->dr_start, dtmp->dr_len);
    else
      GFS_IO_READ(buf, count, BH_DATA(ip->i_didata) + dtmp->dr_start, dtmp->dr_len);    

    count += dtmp->dr_len;
  }


  else
  {
    bh = (buffer_head_t **)gmalloc(transfer * sizeof(buffer_head_t *));
    if (!bh)
    {
      gfree(direct, direct[0].dr_num * sizeof(gfs_direct_t));
      IFDEBUG(DEBUG_FILE1,
	      printk(DEBUG_HOW "Exiting gfs_readi()\n"););
      return(-ENOMEM);
    }

    for (i = 0; i < direct->dr_num; i += j)
    {
      for (j = 0; i + j < direct->dr_num && j < transfer; j++)
      {
	dtmp = &direct[i + j];

	ASSERT(!(dtmp->dr_flags & GFS_BMAP_STUFFED_DINODE));
	
	if (dtmp->dr_flags & GFS_BMAP_HOLE)
	  bh[j] = NULL;
	else
	{
	  bh[j] = gfs_dgetblk(sdp, dtmp->dr_no, ip->i_di.di_glock);
	  if (!bh[j])
	  {
	    do
	    {
	      if (bh[j])
		gfs_dforget(bh[j]);
	    } while (j--);

	    gfree(bh, transfer * sizeof(buffer_head_t *));
	    gfree(direct, direct[0].dr_num * sizeof(gfs_direct_t));

	    IFDEBUG(DEBUG_FILE1,
		    printk(DEBUG_HOW "Exiting gfs_readi()\n"););
	    return(-EIO);	    
	  }
	}
      }

      gfs_dreread(sdp, j, bh, FALSE);

      for (j = 0; i + j < direct->dr_num && j < transfer; j++)
      {
	dtmp = &direct[i + j];

	if (dtmp->dr_flags & GFS_BMAP_HOLE)
	{
	  if (internal)
	    memset((char *)buf + count, 0, dtmp->dr_len);
	  else
	    GFS_IO_READ_HOLE(buf, count, dtmp->dr_len);
	}
	else
	{
	  gfs_wait_on_buffer(bh[j]);

	  retries = 0;

	  while (!gfs_buffer_uptodate(bh[j]))
	  {
	    if (retries++ > GFS_MAX_RETRIES)
	    {
	      if (!gfs_buffer_uptodate(bh[j]))
	      {
		printk(KERN_EMERG "b_blocknr = %lu, b_size = %lu\n",
		       BH_BLKNO(bh[j]), (unsigned long)BH_SIZE(bh[j]));
		ASSERT(gfs_buffer_uptodate(bh[j]));
	      }
	    }

	    gfs_dreread(sdp, 1, &bh[j], TRUE);
	    gfs_wait_on_buffer(bh[j]);
	  }

	  if (internal)
	    memcpy((char *)buf + count, BH_DATA(bh[j]) + dtmp->dr_start, dtmp->dr_len);
	  else
	    GFS_IO_READ(buf, count, BH_DATA(bh[j]) + dtmp->dr_start, dtmp->dr_len);

	  gfs_dforget(bh[j]);
	}
	
	count += dtmp->dr_len;	
      }
    }

    gfree(bh, transfer * sizeof(buffer_head_t *));
  }


  gfree(direct, direct[0].dr_num * sizeof(gfs_direct_t));


  IFDEBUG(DEBUG_FILE2,
	  printk(DEBUG_HOW "  count = %"FMT_64"d\n", count););

  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Exiting gfs_readi()\n"););

  return(count);
}


/******************************************************************************
*******************************************************************************
**
** gfs_writei()
**
** Description:
**   Write bytes to a file
**
** Input(s):
**   ip - the file to write to
**   buf - the buffer to write the data from
**   size - the length of data to written
**   offset - write to this offset in the file
**   internal - if true, use memcpy()s instead of copy_from_user
**
** Output(s):
**   offset - updated by the number of bytes written
**
** Returns:
**   The number of bytes written, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int64 gfs_writei(gfs_inode_t *ip, const void *buf, uint64 offset, uint64 size, int internal)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_direct_t *direct = NULL, *dtmp;
  buffer_head_t *bh;

  int i;
  int error;
  int64 count = 0;


  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Entering gfs_writei()\n"););

  IFDEBUG(DEBUG_FILE2,
	  printk(DEBUG_HOW "  off = %ld, size = %lu\n",
		 (unsigned long)offset, (unsigned long)size););
  
  
  if ((error = gfs_bmapi_write(ip, offset, size, &direct)))
  {
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_writei()\n"););
    return(error);
  }
  
  
  /*  Handle a transfer of zero bytes gracefully  */
  
  if (!direct)
  {
    IFDEBUG(DEBUG_FILE1,
	    printk(DEBUG_HOW "Exiting gfs_writei()\n"););    
    return(0);
  }
  
  
  IFDEBUG(DEBUG_FILE3,
	  for (i = 0; i < direct->dr_num; i++)
	  {
	    printk(DEBUG_HOW "  direct #%d\n", i);
	    printk(DEBUG_HOW "    dr_num = %d\n", direct[i].dr_num);
	    printk(DEBUG_HOW "    dr_flags = %d\n", direct[i].dr_flags);
	    printk(DEBUG_HOW "    dr_no = %"FMT_64"u\n", direct[i].dr_no);
	    printk(DEBUG_HOW "    dr_start = %d\n", direct[i].dr_start);
	    printk(DEBUG_HOW "    dr_len = %d\n", direct[i].dr_len);
	  });



  if (direct[0].dr_flags & GFS_BMAP_STUFFED_DINODE)
  {
    dtmp = &direct[0];

    ASSERT(dtmp->dr_num == 1);

    if (internal)
      memcpy(BH_DATA(ip->i_didata) + dtmp->dr_start, (char *)buf + count, dtmp->dr_len);
    else
      GFS_IO_WRITE(buf, count, BH_DATA(ip->i_didata) + dtmp->dr_start, dtmp->dr_len);

    count += dtmp->dr_len;

    ip->i_di.di_atime = ip->i_di.di_mtime = ip->i_di.di_ctime = CURRENT_TIME;

    gfs_inode_attr_in(ip);
    gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
    gfs_dwrite(ip->i_didata, 0);
  }
  else
  {
    ip->i_di.di_atime = ip->i_di.di_mtime = ip->i_di.di_ctime = CURRENT_TIME;

    gfs_inode_attr_in(ip);
    gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
    gfs_dwrite(ip->i_didata, 0);
    
    for (i = 0; i < direct->dr_num; i++)
    {
      dtmp = &direct[i];

      ASSERT(!(dtmp->dr_flags & GFS_BMAP_STUFFED_DINODE));
	
      if (dtmp->dr_len == sdp->sd_sb.sb_bsize)
      {
	bh = gfs_dgetblk(sdp, dtmp->dr_no, ip->i_di.di_glock);
      }
      else
      {
	if (dtmp->dr_flags & GFS_BMAP_NEW)
	{
	  bh = gfs_dgetblk(sdp, dtmp->dr_no, ip->i_di.di_glock);
	  if (bh)
	    memset(BH_DATA(bh), 0, sdp->sd_sb.sb_bsize);
	}
	else
	  bh = gfs_dread(sdp, dtmp->dr_no, ip->i_di.di_glock);
      }
      
      if (!bh)
      {
	gfree(direct, direct[0].dr_num * sizeof(gfs_direct_t));
	IFDEBUG(DEBUG_FILE1,
		printk(DEBUG_HOW "Exiting gfs_writei()\n"););
	return(-EIO);	    
      }
      gfs_mark_uptodate(bh, TRUE);
      
      if (internal)
      {
	memcpy(BH_DATA(bh) + dtmp->dr_start, (char *)buf + count, dtmp->dr_len);
	gfs_dwrite(bh, DW_NOKEEP);
      }
      else
      {
	GFS_IO_WRITE(buf, count, BH_DATA(bh) + dtmp->dr_start, dtmp->dr_len);
	gfs_dwrite(bh, DW_NOKEEP | DW_NOTMETA);
      }

      count += dtmp->dr_len; 
    }
  }


  gfree(direct, direct[0].dr_num * sizeof(gfs_direct_t));


  IFDEBUG(DEBUG_FILE2,
	  printk(DEBUG_HOW "  count = %"FMT_64"d\n", count););

  IFDEBUG(DEBUG_FILE1,
	  printk(DEBUG_HOW "Exiting gfs_writei()\n"););

  return(count);
}


/******************************************************************************
*******************************************************************************
**
** gfs_rgrp_read()
**
** Description:
**   Read in a resource group.
**
**   We go though a bit of gymnastics here to make sure all the buffers
**   are read in one ll_rw_block() call.  Hopefully they will get 
**   merged into one request to the disk and save some time.
**
** Input(s):
**   sdp - the superblock of the filesystem that contains the resource group
**   rgd - a resource group data structure that describes the resource group
**           to read.
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_rgrp_read(gfs_sbd_t *sdp, gfs_rgrpd_t *rgd)
{
  int x;
  int length = rgd->rd_ri.ri_length;
  uint32 bytes_left, bytes;


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_rgrp_read()\n"););


  /*  Allocate space for the data if it isn't there  */

  if (!rgd->rd_bh)
  {
    rgd->rd_bh = (buffer_head_t **)gmalloc(length * sizeof(buffer_head_t *));
    rgd->rd_bits = (gfs_bitmap_t *)gmalloc(length * sizeof(gfs_bitmap_t));
    if (!rgd->rd_bh || !rgd->rd_bits)
    {
      if (rgd->rd_bh)
      {
	gfree(rgd->rd_bh, length * sizeof(buffer_head_t *));
	rgd->rd_bh = NULL;
      }
      IFDEBUG(DEBUG_DIO1,
	      printk(DEBUG_HOW "Exiting gfs_rgrp_read()\n"););
      return(-ENOMEM);
    }

    bytes_left = rgd->rd_ri.ri_bitbytes;

    for (x = 0; x < length; x++)
    {
      rgd->rd_bh[x] = gfs_dgetblk(sdp, (rgd->rd_ri.ri_addr + x), rgd->rd_ri.ri_glock);
      if (!rgd->rd_bh[x])
	break;

      if (length == 1)
      {
	rgd->rd_bits[x].bi_bitmap = BH_DATA(rgd->rd_bh[x]) + sizeof(gfs_rgrp_t);
	rgd->rd_bits[x].bi_start = 0;
	rgd->rd_bits[x].bi_len = bytes_left;
      }
      else if (x == 0)
      {
	bytes = sdp->sd_sb.sb_bsize - sizeof(gfs_rgrp_t);
	rgd->rd_bits[x].bi_bitmap = BH_DATA(rgd->rd_bh[x]) + sizeof(gfs_rgrp_t);
	rgd->rd_bits[x].bi_start = 0;
	rgd->rd_bits[x].bi_len = bytes;
	bytes_left -= bytes;
      }
      else if (x + 1 == length)
      {
	rgd->rd_bits[x].bi_bitmap = BH_DATA(rgd->rd_bh[x]);
	rgd->rd_bits[x].bi_start = rgd->rd_ri.ri_bitbytes - bytes_left;
	rgd->rd_bits[x].bi_len = bytes_left;
      }
      else
      {
	bytes = sdp->sd_sb.sb_bsize;
	rgd->rd_bits[x].bi_bitmap = BH_DATA(rgd->rd_bh[x]);
	rgd->rd_bits[x].bi_start = rgd->rd_ri.ri_bitbytes - bytes_left;
	rgd->rd_bits[x].bi_len = bytes;
	bytes_left -= bytes;
      }
    }
    
    if (x != length)
    {
      for (x = 0; x < length; x++)
	gfs_dforget(rgd->rd_bh[x]);
      gfree(rgd->rd_bh, length * sizeof(buffer_head_t *));
      gfree(rgd->rd_bits, length * sizeof(gfs_bitmap_t));
      rgd->rd_bh = NULL;
      rgd->rd_bits = NULL;
      IFDEBUG(DEBUG_DIO1,
	      printk(DEBUG_HOW "Exiting gfs_rgrp_read()\n"););
      return(-EIO);
    }
  }


  /*  Read the blocks in one command  */

  gfs_dreread(sdp, length, rgd->rd_bh, TRUE);


  gfs_rgrp_in(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));

  if (rgd->rd_rg.rg_magic != GFS_MAGIC_RG)
  {
    printk(KERN_EMERG "GFS:  rgd->rd_rg.rg_magic (%"FMT_64"u) != GFS_MAGIC_RG\n",
	   rgd->rd_ri.ri_addr);
    ASSERT(0);
  }


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_rgrp_read()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_rgrp_write()
**
** Description:
**   Write out a resource group.
**
**   We go though a bit of gymnastics here to make sure all the buffers
**   are written in one ll_rw_block() call.  Hopefully they will get 
**   merged into one request to the disk and save some time.
**
** Input(s):
**   sdp - the superblock of the filesystem that contains the resource group
**   rgd - a resource group data structure that describes the resource group
**           to write.
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_rgrp_write(gfs_sbd_t *sdp, gfs_rgrpd_t *rgd)
{
  int x;
  int length = rgd->rd_ri.ri_length;
  

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_rgrp_write()\n"););

  ASSERT(rgd->rd_bh != NULL);


  gfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));

  for (x = 0; x < length; x++)
    gfs_dwrite(rgd->rd_bh[x], 0);


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_rgrp_write()\n"););

  return(0);
}


