/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name:  gfs_ondisk.h
**
**    Description:  The GFS on-disk structures and the prototypes for the
**      routines to read and write them.
**
**    NOTE:
**      
**      If you add 8 byte fields to these structures, they must be 8 byte
**      aligned.  4 byte field must be 4 byte aligned, etc...
**
**      All structures must be a multiple of 8 bytes long.
**
*******************************************************************************
******************************************************************************/

#ifndef __GFS_ONDISK_DOT_H__
#define __GFS_ONDISK_DOT_H__

#include "global.h"



/* Magic Numbers */

#define GFS_MAGIC_SB       	(0x01011670)
#define GFS_MAGIC_RG       	(0x02011670)
#define GFS_MAGIC_DI       	(0x03011670)
#define GFS_MAGIC_LF            (0x04011670)



/* Values and Constants */

#define GFS_BASIC_BLOCK         (512)

#define GFS_FNAMESIZE       	(256)

/* Number of blocks per byte in rgrp */
#define GFS_NBBY		(4)
/*  Address of SuperBlock in GFS basic blocks
    Must be zero or a multiple of 16  */
#define GFS_SB_ADDR     	(0)

#define GFS_SB_LOCK     	(0)
#define GFS_RENAME_LOCK         (1)
#define GFS_ROOT_LOCK           (2)
#define GFS_RGRP0_LOCK          (3)

#define FILE_NON                (0)
#define FILE_REG                (1)
#define FILE_DIR                (2)
#define FILE_LNK                (5)
#define FILE_BLK                (7)
#define FILE_CHR                (8)
#define FILE_FIFO               (101)

#define GFS_DIR_LINEAR          (0)
#define GFS_DIR_EXHASH          (1)

#define GFS_DIR_MAX_DEPTH       (21)

/*  Requirement:  GFS_MOUNT_BITS % 8 == 0  */
#define GFS_MOUNT_BITS          (8)

/*  Requirement:  (GFS_LOCKNAME_LEN + 1) % 8 == 0  */
#define GFS_LOCKNAME_LEN        (31)



/*
 *  super-block structure
 *
 *  It's probably good if SIZEOF_SB <= GFS_BASIC_BLOCK
 */

typedef struct gfs_sb_s
{
  /*  Order is important  */
  uint32              sb_magic;         /* super-block magic number */
  uint32              sb_version;       /* Version of GFS */

  /*  Important information  */
  uint32              sb_bsize;         /* fundamental fs block size in bytes */
  uint32              sb_length;        /* length of super-block + resource index
					   in FS blocks*/
  uint64              sb_rootdi;        /* root dinode number */
  uint32              sb_rgrps;         /* number of resource groups */

  char                sb_name[12];      /* file system name */

  uint64              sb_fssize;        /* number of blocks in file system */

  uint64              sb_glocks;        /* num of global locks */
  char                sb_lockproto[GFS_LOCKNAME_LEN + 1];  /* Type of locking this FS uses */
  char                sb_locktable[GFS_LOCKNAME_LEN + 1];  /* Name of lock table for this FS */

  /*  OnDisk structure sizes in bytes  */
  uint16              sb_sbsize;        /* Super Block */
  uint16              sb_risize;        /* Resource Index */
  uint16              sb_rgsize;        /* Resource Group */
  uint16              sb_disize;        /* Dinode */
  uint16              sb_desize;        /* Directory Entry */
  uint16              sb_lfsize;        /* Directory Leaf */

  /*  Usage Stuff  */
  /*  Note:  These are the only field in the SuperBlock and ResourceIndex
      that change.  Everything else is static  */
  uint32              sb_mounts;        /* number of mounts */
  uint8               sb_mountids[GFS_MOUNT_BITS];  /* bitmaps of unique ids */
  int32               sb_time;          /* last super-block update */
  uint32              sb_pad;
} gfs_sb_t;


/*
 *  resource index stucture 
 */

typedef struct gfs_rindex_s
{
  uint64              ri_addr;          /* rgrp block disk address */
  uint32              ri_length;        /* length of rgrp header in fs blocks */
  uint32              ri_subpool;       /* subpool of rgrp */
  uint64              ri_glock;         /* disk lock number */

  uint64              ri_data1;         /* first data location */
  uint64              ri_data;          /* num of data blocks in rgrp */

  uint64              ri_glock1;        /* first glock in rgrp */
  uint64              ri_glocks;        /* glocks allocated to rgrp */

  uint32              ri_bitbytes;      /* number of bytes in data bitmaps */
  uint32              ri_pad;
} gfs_rindex_t;


/*
 *  resource group header structure
 */

typedef struct gfs_rgrp_s
{
  uint32              rg_magic;         /* resource group header magic number */
  int32               rg_time;          /* last rgrp block update */
  uint64              rg_free;          /* number of free data blocks */
  uint64              rg_dinodes;       /* number of dinodes */
} gfs_rgrp_t;


/*
 *  dinode structure
 */

typedef struct gfs_dinode_s
{
  uint32              di_magic;         /* dinode block magic number */
  uint32              di_pad;

  uint64              di_addr;          /* dinode number (disk addr) */

  uint32              di_mode;          /* mode of file */
  uint32              di_uid;	        /* owner's user id */
  uint32              di_gid;           /* owner's group id */
  uint32              di_nlink;         /* number of links to this file */
  uint64              di_bsize;         /* number of bytes in file */
  uint64              di_blocks;        /* number of blocks in file */
  int32               di_atime;         /* time last accessed */
  int32               di_mtime;         /* time last modified */
  int32               di_ctime;         /* time last changed */
  uint32              di_major;         /* device major number */
  uint32              di_minor;         /* device minor number */

  uint32              di_rgrp;          /* rgrp number */
  uint32              di_rgrp_meta;     /* rgrp of last meta data alloc */
  uint32              di_rgrp_data;     /* rgrp of last data blk alloc */
  uint64              di_last_meta;     /* last blk for meta for this dinode */
  uint64              di_last_data;     /* last blk for data for this dinode */
  uint64              di_glock;         /* global lock number */
  uint16              di_type;          /* type of file */
  uint16              di_height;        /* height of metadata */
  uint16              di_stamp;         /* random value at creation */

  uint16              di_nopen;         /* number of hosts who have this file open */ 
  uint8               di_open[GFS_MOUNT_BITS];  /* bitmap of hosts who have this file open */

  /*  These only apply to directories  */
  uint16              di_regime;        /* Linear/ExHash */
  uint16              di_depth;         /* Number of bits in the table */
  uint32              di_entries;       /* The number of entries in the directory */
  uint32              di_same_depth;    /* # of leaves at same depth as directory */
  uint32              di_pad1;
} gfs_dinode_t;


/*
 *  directory structure - many of these per directory file
 */

#define GFS_DIR_PAD (8)
#define GFS_DIR_ROUND (GFS_DIR_PAD - 1)
#define GFS_DIR_REC_LEN(name_len) (((name_len) + 24 + GFS_DIR_ROUND) & \
                                   (~GFS_DIR_ROUND))

typedef struct gfs_dirent_s
{
  uint64              de_ino;           /* dinode number */
  uint32              de_hash;          /* hash of the filename */
  uint32              de_name_len;
  uint32              de_rec_len;
 
  uint32              de_pad;  
  
  char                de_name[GFS_FNAMESIZE];  /* filename */
} gfs_dirent_t;


/*
 *  Header of leaf directory nodes
 */

typedef struct gfs_leaf_s
{
  uint32              lf_magic;         /* Magic Number */

  uint16              lf_entries;        /* Number of dirents in leaf */
  uint16              lf_depth;         /* Depth of leaf */
  uint64              lf_next;          /* Next leaf, if overflow */
  uint32              lf_bytes_used;    /* bytes actually needed by dirents */
  uint32              lf_hash;           /* we might used this to start a linked list of free dirents */
  uint64              lf_leafno;        /* the leaf # */
} gfs_leaf_t;



/*  Translation functions  */

void gfs_sb_in(gfs_sb_t *sb, char *buf);
void gfs_sb_out(gfs_sb_t *sb, char *buf);
void gfs_rindex_in(gfs_rindex_t *rindex, char *buf);
void gfs_rindex_out(gfs_rindex_t *rindex, char *buf);
void gfs_rgrp_in(gfs_rgrp_t *rgrp, char *buf);
void gfs_rgrp_out(gfs_rgrp_t *rgrp, char *buf);
void gfs_dinode_in(gfs_dinode_t *dinode, char *buf);
void gfs_dinode_out(gfs_dinode_t *dinode, char *buf);
void gfs_dirent_in(gfs_dirent_t *dirent, char *buf);
void gfs_dirent_out(gfs_dirent_t *dirent, char *buf);
void gfs_leaf_in(gfs_leaf_t *leaf, char *buf);
void gfs_leaf_out(gfs_leaf_t *leaf, char *buf);



#endif  /*  __GFS_ONDISK_DOT_H__  */


