/******************************************************************************
*******************************************************************************
**
**    This code was shamelessly stolen from the Linux kernel.
**
**    It is Copyright Linus Torvalds and others.  It is licensed under the
**    GNU GPL.
**    
**    Module name: gfs_list.h
**
**    Description:
**      This code implements standard list code across different OSs.
**
*******************************************************************************
******************************************************************************/

#ifndef __GFS_LIST_DOT_H__
#define __GFS_LIST_DOT_H__

/*
 * Simple doubly linked gfs_list implementation.
 *
 * Some of the internal functions ("__xxx") are useful when
 * manipulating whole gfs_lists rather than single entries, as
 * sometimes we already know the next/prev entries and we can
 * generate better code by using them directly rather than
 * using the generic single-entry routines.
 */

struct gfs_list_head {
	struct gfs_list_head *next, *prev;
};
typedef struct gfs_list_head gfs_list_head_t;

#define GFS_LIST_HEAD(name) \
	struct gfs_list_head name = { &name, &name }

#define INIT_GFS_LIST_HEAD(ptr) do { \
	(ptr)->next = (ptr); (ptr)->prev = (ptr); \
} while (0)

/*
 * Insert a new entry between two known consecutive entries. 
 *
 * This is only for internal gfs_list manipulation where we know
 * the prev/next entries already!
 */
static __inline__ void __gfs_list_add(struct gfs_list_head * new,
	struct gfs_list_head * prev,
	struct gfs_list_head * next)
{
	next->prev = new;
	new->next = next;
	new->prev = prev;
	prev->next = new;
}

/*
 * Insert a new entry after the specified head..
 */
static __inline__ void gfs_list_add(struct gfs_list_head *new, struct gfs_list_head *head)
{
	__gfs_list_add(new, head, head->next);
}

/*
 * Delete a gfs_list entry by making the prev/next entries
 * point to each other.
 *
 * This is only for internal gfs_list manipulation where we know
 * the prev/next entries already!
 */
static __inline__ void __gfs_list_del(struct gfs_list_head * prev,
				  struct gfs_list_head * next)
{
	next->prev = prev;
	prev->next = next;
}

static __inline__ void gfs_list_del(struct gfs_list_head *entry)
{
	__gfs_list_del(entry->prev, entry->next);
}

static __inline__ int gfs_list_empty(struct gfs_list_head *head)
{
	return head->next == head;
}

/*
 * Splice in "gfs_list" into "head"
 */
static __inline__ void gfs_list_splice(struct gfs_list_head *gfs_list, struct gfs_list_head *head)
{
	struct gfs_list_head *first = gfs_list->next;

	if (first != gfs_list) {
		struct gfs_list_head *last = gfs_list->prev;
		struct gfs_list_head *at = head->next;

		first->prev = head;
		head->next = first;

		last->next = at;
		at->prev = last;
	}
}

#define gfs_list_entry(ptr, type, member) \
	((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))

#endif  /*  __GFS_LIST_DOT_H__  */
