/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: glock.c
**
**    Description:
**	This module implements the GFS interface to the global locks.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "dio.h"
#include "glock.h"
#include "inode.h"



/*  Macros  */

#define DROPPED_LOCK(x, y) {printk(KERN_ERR "GFS:  dropped lock %"FMT_64"u (%d) in %s on line %d\n", \
				   (x), (y), __FILE__, __LINE__);};



/******************************************************************************
*******************************************************************************
**
** handle_expired()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/
static void handle_expired(gfs_sbd_t *sdp, gfs_glock_t *gl, int flag)
{
  printk(KERN_ERR "GFS:  glock %"FMT_64"u has expired (0x%.8X)\n",
	 gl->gl_glock, flag);
  printk(KERN_ERR "\tEstimated time to panic:  0.2633333 seconds\n");
}


/******************************************************************************
*******************************************************************************
**
** gfs_demote_glock()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int gfs_demote_glock(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  int flags, ret, mod = FALSE;



  if (!(gl->gl_release & (CB_NEED_S | CB_NEED_E)))
    return(0);
  if (!gl->gl_held)
    return(0);
  if (!(gl->gl_release & CB_NEED_E) && gl->gl_shared)
  {
    gl->gl_release &= ~CB_NEED_S;
    return(0);
  }


  ASSERT(gl->gl_locked == 0);


  if (gl->gl_incr)
  {
    ASSERT(!gl->gl_shared);
    gfs_sync_bh(gl);
    gl->gl_incr = FALSE;
    mod = TRUE;
  }


  do
  {
    if (gl->gl_release & CB_NEED_E)
    {
      flags = (mod) ? ULCK_MODIFIED : 0;

      ret = sdp->sd_lockstruct->unlock(sdp->sd_lockdata, gl->gl_glock, flags, &gl->gl_lockdata);

      GFS_SEMA_DOWN(&sdp->sd_gl_lru_lock);
      sdp->sd_gl_heldcount--;
      gfs_list_del(&gl->gl_lru);
      GFS_SEMA_UP(&sdp->sd_gl_lru_lock);

      gl->gl_held = FALSE;

      if (ret)
	DROPPED_LOCK(gl->gl_glock, ret);

      return(ret);
    }
    else 
    {
      flags = LCK_SHARED | ((mod) ? LCK_COMMUTE_MOD : LCK_COMMUTE);

      ret = sdp->sd_lockstruct->lock(sdp->sd_lockdata, gl->gl_glock, LCK_OP_TRY, flags, &gl->gl_lockdata);
      if (ret < 0)
      {
	GFS_SEMA_DOWN(&sdp->sd_gl_lru_lock);
	sdp->sd_gl_heldcount--;
	gfs_list_del(&gl->gl_lru);
	GFS_SEMA_UP(&sdp->sd_gl_lru_lock);

	gl->gl_held = FALSE;

	DROPPED_LOCK(gl->gl_glock, ret);

	return(ret);
      }

      if (!(ret & LCKOUT_CACHEABLE))
      {
	gfs_inval_bh(gl);
	gl->gl_vn++;
      }

      if (ret & LCKOUT_HELD)
      {
	gl->gl_shared = ret & LCKOUT_SHARED;
	gl->gl_release = ret & LCKOUT_NEED_E;
	if (ret & (LCKOUT_EXPIRED_S | LCKOUT_EXPIRED_E))
	  handle_expired(sdp, gl, ret & (LCKOUT_EXPIRED_S | LCKOUT_EXPIRED_E));
      }
      else
      {
	GFS_SEMA_DOWN(&sdp->sd_gl_lru_lock);
	sdp->sd_gl_heldcount--;
	gfs_list_del(&gl->gl_lru);
	GFS_SEMA_UP(&sdp->sd_gl_lru_lock);

	gl->gl_held = FALSE;

	return(0);
      }
    }
  } while (gl->gl_release);


  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_get_glstruct()
**
** Description:
**   Search through the hash table in the superblock and find the Glock 
**   structure for this lock number.  If a structure doesn't exist, create
**   one.
**
** Input(s):
**   sdp - the superblock data for the file system
**   lock - the lock nubmer
**
** Returns:
**   A pointer to the Glock structure, or NULL on error
**
*******************************************************************************
******************************************************************************/

gfs_glock_t *gfs_get_glstruct(gfs_sbd_t *sdp, uint64 lock, int create)
{
  uint32 bucket = lock % GFS_GHASH_SIZE;
  gfs_list_head_t *head, *next;
  gfs_glock_t *gl = NULL, *tmp_gl;


  IFDEBUG(DEBUG_GLOCK1,
	  printk(DEBUG_HOW "Entering gfs_get_glstruct()\n"););


  GFS_SEMA_DOWN(&(sdp->sd_gla_lock[bucket]));

  head = &(sdp->sd_gla[bucket]);

  next = head->next;
  while (1)
  {
    gfs_list_head_t *tmp = next;

    next = next->next;
    if (tmp == head)
      break;
    tmp_gl = gfs_list_entry(tmp, gfs_glock_t, gl_next);

    if (tmp_gl->gl_glock == lock)
    {
      gl = tmp_gl;
      gl->gl_count++;
      break;
    }
  }

  if (!gl && create)
  {
    gl = (gfs_glock_t *)gmalloc(sizeof(gfs_glock_t));
    if (gl)
    {
      memset(gl, 0, sizeof(gfs_glock_t));

      gfs_list_add(&gl->gl_next, &sdp->sd_gla[bucket]);
      gl->gl_glock = lock;
      gl->gl_count = 1;
      INIT_GFS_LIST_HEAD(&gl->gl_lru);

      GFS_SEMA_INIT(&gl->gl_lock);
      GFS_SEMA_INIT_LKD(&gl->gl_wait);

      INIT_GFS_LIST_HEAD(&gl->gl_bufs);
      GFS_SEMA_INIT(&gl->gl_buflock);
    }
  }

  GFS_SEMA_UP(&(sdp->sd_gla_lock[bucket]));
  

  IFDEBUG(DEBUG_GLOCK1,
	  printk(DEBUG_HOW "Exiting gfs_get_glstruct()\n"););

  return(gl);
}


/******************************************************************************
*******************************************************************************
**
** gfs_put_glstruct()
**
** Description:
**   Put back a Glock structure.  If it's unused, free it.
**
** Input(s):
**   sdp - the superblock data for the file system
**   gl - the Glock structure
**
*******************************************************************************
******************************************************************************/

void gfs_put_glstruct(gfs_sbd_t *sdp, gfs_glock_t *gl)
{
  uint32 bucket = gl->gl_glock % GFS_GHASH_SIZE;
  int del = FALSE;


  IFDEBUG(DEBUG_GLOCK1,
	  printk(DEBUG_HOW "Entering gfs_put_glstruct()\n"););


  GFS_SEMA_DOWN(&(sdp->sd_gla_lock[bucket]));

  ASSERT(gl->gl_count != 0);

  gl->gl_count--;

  if (gl->gl_count == 0 && !gl->gl_held)
  {
    GFS_SEMA_DOWN(&gl->gl_buflock);

    if (gfs_list_empty(&gl->gl_bufs))
      del = TRUE;

    GFS_SEMA_UP(&gl->gl_buflock);

    if (del)
    {
      if (gl->gl_lockdata)
	sdp->sd_lockstruct->release(sdp->sd_lockdata, &gl->gl_lockdata);
      gfs_list_del(&gl->gl_next);
      gfree(gl, sizeof(gfs_glock_t));
    }
  }

  GFS_SEMA_UP(&(sdp->sd_gla_lock[bucket]));


  IFDEBUG(DEBUG_GLOCK1,
	  printk(DEBUG_HOW "Exiting gfs_put_glstruct()\n"););
}


/******************************************************************************
*******************************************************************************
**
** int gfs_glock()
**
** Description:
**   This routine locks a glock.  Allows Glocks to be recursively held by the
**   same process.  If a different process tries to obtain a lock that is already
**   held, it sleeps.  Pass in a valid inode as "ip" if the lock is associated
**   with this inode and the dinode will be reread if necessary.  Pass in NULL
**   otherwise.
**
** Input(s):
**   sdp - the filesystem
**   lock - the lock number
**   ip - the inode associated with the lock (Optional)
**   flags - This doesn't seem to do anything at the moment 
**   cachable - if not NULL, set to 1 if cachable, 0 if not cachable
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_glock(gfs_sbd_t *sdp, uint64 lock, gfs_inode_t *ip, int flags)
{
  int done = FALSE, new_hold = FALSE;
  gfs_glock_t *gl;
  int lck_op, lck_flags, lck_ret;


  IFDEBUG(DEBUG_GLOCK1,
	  printk(DEBUG_HOW "Entering gfs_glock()\n"););

  gl = gfs_get_glstruct(sdp, lock, TRUE);
  if (!gl)
  {
    IFDEBUG(DEBUG_GLOCK1,
	    printk(DEBUG_HOW "Exiting gfs_glock()\n"););
    return(-ENOMEM);
  }


  
  do
  {
    GFS_SEMA_DOWN(&gl->gl_lock);

    if (gl->gl_locked == 0)
    {
      if (!gl->gl_held  || (gl->gl_shared && !(flags & GL_SHARED)))
      /*  Lock not held || Lock is being promoted  */ 
      {
	lck_op = LCK_OP_ACQUIRE;
	lck_flags = ((flags & GL_SHARED) ? LCK_SHARED : 0) | ((gl->gl_held) ? LCK_COMMUTE : 0);

	lck_ret = sdp->sd_lockstruct->lock(sdp->sd_lockdata, lock, lck_op, lck_flags, &gl->gl_lockdata);
	if (lck_ret < 0)
	{
	  if (gl->gl_held)
	  {
	    GFS_SEMA_DOWN(&sdp->sd_gl_lru_lock);
	    sdp->sd_gl_heldcount--;
	    gfs_list_del(&gl->gl_lru);
	    GFS_SEMA_UP(&sdp->sd_gl_lru_lock);
	    gl->gl_held = FALSE;

	    DROPPED_LOCK(lock, lck_ret);
	  }

	  GFS_SEMA_UP(&gl->gl_lock);
	  gfs_put_glstruct(sdp, gl);

	  IFDEBUG(DEBUG_GLOCK1,
		  printk(DEBUG_HOW "Exiting gfs_glock()\n"););
	  return(lck_ret);
	}

	ASSERT(lck_ret & LCKOUT_HELD);

	if (!(lck_ret & LCKOUT_CACHEABLE))
	{
	  gfs_inval_bh(gl);
	  gl->gl_vn++;
	}

	/*  If we asked for exclusive, we better have gotten exclusive.  */
	ASSERT((flags & GL_SHARED) || !(lck_ret & LCKOUT_SHARED));

	if (!gl->gl_held)
	  new_hold = TRUE;

	gl->gl_held = TRUE;
	gl->gl_shared = lck_ret & LCKOUT_SHARED;
	gl->gl_release = lck_ret & (LCKOUT_NEED_S | LCKOUT_NEED_E);

	if (lck_ret & (LCKOUT_EXPIRED_S | LCKOUT_EXPIRED_E))
	  handle_expired(sdp, gl, lck_ret & (LCKOUT_EXPIRED_S | LCKOUT_EXPIRED_E));	  
      }


     /*  Update the inode  */

      if (ip && !gfs_buffer_uptodate(ip->i_didata))
      {
	gfs_dreread(sdp, 1, &ip->i_didata, TRUE);
	gfs_dinode_in(&ip->i_di, BH_DATA(ip->i_didata));
	ASSERT(ip->i_di.di_magic == GFS_MAGIC_DI);
	gfs_inode_attr_in(ip);
	ip->i_vn = gl->gl_vn;
      }

      gl->gl_locked = 1;
      gl->gl_pid = kpid();
      
      GFS_SEMA_UP(&gl->gl_lock);

      done = TRUE;
    }


    else if (gl->gl_pid == kpid())
    {
      ASSERT(gl->gl_held);

      /*  Promotion during recursive locking doesn't work.  */
      ASSERT((flags & GL_SHARED) || !gl->gl_shared);

     /*  Update the inode  */

      if (ip && !gfs_buffer_uptodate(ip->i_didata))
      {
	gfs_dreread(sdp, 1, &ip->i_didata, TRUE);
	gfs_dinode_in(&ip->i_di, BH_DATA(ip->i_didata));
	ASSERT(ip->i_di.di_magic == GFS_MAGIC_DI);
	gfs_inode_attr_in(ip);
	ip->i_vn = gl->gl_vn;
      }

      gl->gl_locked++;

      GFS_SEMA_UP(&gl->gl_lock);

      done = TRUE;
    }


    else
    {
      gl->gl_waiting++;
      
      GFS_SEMA_UP(&gl->gl_lock);
      GFS_SEMA_DOWN(&gl->gl_wait);
    }
  } while (!done);
  


  if (gl->gl_locked == 1)
  {
    /*  Make this lock MRU  */

    GFS_SEMA_DOWN(&sdp->sd_gl_lru_lock);

    if (new_hold)
      sdp->sd_gl_heldcount++;
    else
      gfs_list_del(&gl->gl_lru);
    gfs_list_add(&gl->gl_lru, &sdp->sd_gl_lru);

    if (sdp->sd_gl_heldcount > GFS_MAX_HELD_LOCKS)
    {
      gfs_glock_t *gl2 = gfs_list_entry(sdp->sd_gl_lru.prev, gfs_glock_t, gl_lru);

      gl2 = gfs_get_glstruct(sdp, gl2->gl_glock, FALSE);
      ASSERT(gl2);

      GFS_SEMA_UP(&sdp->sd_gl_lru_lock);
      
      if (!GFS_SEMA_DOWN_TRY(&gl2->gl_lock))
      {
	if (!gl2->gl_locked && gl2->gl_held)
	{
	  gl2->gl_release = CB_NEED_E;
	  gfs_demote_glock(sdp, gl2);
	}
      
	GFS_SEMA_UP(&gl2->gl_lock);
      }

      gfs_put_glstruct(sdp, gl2);
    }    
    else
      GFS_SEMA_UP(&sdp->sd_gl_lru_lock);
  }


  
  gfs_put_glstruct(sdp, gl);

  IFDEBUG(DEBUG_GLOCK1,
	  printk(DEBUG_HOW "Exiting gfs_glock()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** int gfs_gunlock()
**
** Description:
**   This routine unlocks a Glock.
**
** Input(s):
**   sdp - the filesystem
**   lock - the lock number
**   ip - the inode associated with this lock (optional)
**   flags - GL_INCR or GL_ZAP
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_gunlock(gfs_sbd_t *sdp, uint64 lock, int flags)
{
  int error = 0;
  gfs_glock_t *gl;

  
  IFDEBUG(DEBUG_GLOCK1,
	  printk(DEBUG_HOW "Entering gfs_gunlock()\n"););
  
  gl = gfs_get_glstruct(sdp, lock, FALSE);
  ASSERT(gl);

  
  
  GFS_SEMA_DOWN(&gl->gl_lock);

  /*  GL_SYNC forces this lock's dirty buffers to be synced  */

  if (flags & GL_SYNC)
    gfs_sync_bh(gl);

  /*  GL_INCR forces the version number to incremented.  */

  if (flags & GL_INCR)
    gl->gl_incr = TRUE;

  ASSERT(gl->gl_locked > 0);

  gl->gl_locked--;

  /*  GL_ZAP causes all copies of the lock to be released.  */

  if (flags & GL_ZAP)
    gl->gl_locked = 0;
  
  if (gl->gl_locked == 0)
  {
    if (gl->gl_release)
      error = gfs_demote_glock(sdp, gl);
    
    if (gl->gl_waiting > 0)
    {
      gl->gl_waiting--;
      
      GFS_SEMA_UP(&gl->gl_lock);
      GFS_SEMA_UP(&gl->gl_wait);
    }
    else
      GFS_SEMA_UP(&gl->gl_lock);
  }
  else
    GFS_SEMA_UP(&gl->gl_lock);



  gfs_put_glstruct(sdp, gl);

  IFDEBUG(DEBUG_GLOCK1,
	  printk(DEBUG_HOW "Exiting gfs_gunlock()\n"););

  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_glock_cb()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void gfs_glock_cb(void *fsdata, uint64 lock, int flags)
{
  gfs_sbd_t *sdp = (gfs_sbd_t *)fsdata;
  gfs_glock_t *gl;


  flags &= (CB_NEED_S | CB_NEED_E | CB_EXPIRED_S | CB_EXPIRED_E);
  if (!flags)
  {
    printk("GFS:  Spontaneous call to gfs_glock_cb()\n");
    return;
  }


  gl = gfs_get_glstruct(sdp, lock, FALSE);
  if (!gl)
    return;

  if (flags & (CB_EXPIRED_S | CB_EXPIRED_E))
  {
    handle_expired(sdp, gl, flags & (CB_EXPIRED_S | CB_EXPIRED_E));
    gfs_put_glstruct(sdp, gl);
    return;
  }

  if (GFS_SEMA_DOWN_TRY(&gl->gl_lock))
  {
    gfs_put_glstruct(sdp, gl);
    return;
  }

  if (gl->gl_held)
  {
    gl->gl_release |= flags;

    if (!gl->gl_locked)
      gfs_demote_glock(sdp, gl);
  }

  GFS_SEMA_UP(&gl->gl_lock);


  gfs_put_glstruct(sdp, gl);
}


/******************************************************************************
*******************************************************************************
**
** gfs_glock_m()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int gfs_glock_m(gfs_sbd_t *sdp, int num_lock, gfs_lockop_t *ops)
{
  gfs_lockop_t *point[num_lock], *tmp;
  int x, y;
  int changed;
  int error = 0;


  for (x = 0; x < num_lock; x++)
    point[x] = &ops[x];


  /*  Sort the locks in ascending numerical order.

      If there are locks with the same lock number, sort them so that any
      exclusive locks come before shared locks  */

  do
  {
    changed = FALSE;

    for (x = 0; x < num_lock - 1; x++)
    {
      if (point[x]->lock > point[x + 1]->lock ||
	  ((point[x]->lock == point[x + 1]->lock) &&
	   ((point[x]->flags & GL_SHARED) && !(point[x + 1]->flags & GL_SHARED))))
      {
	tmp = point[x];
	point[x] = point[x + 1];
	point[x + 1] = tmp;
	changed = TRUE;
      }
    }
  } while (changed);



  for (x = 0; x < num_lock; x++)
  {
    error = gfs_glock(sdp, point[x]->lock, point[x]->ip, point[x]->flags);
    if (error)
      break;
  } 


  if (x < num_lock)
    for (y = 0; y < x; y++)
      gfs_gunlock(sdp, point[y]->lock, point[y]->flags);


  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_gunlock_m()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int gfs_gunlock_m(gfs_sbd_t *sdp, int num_lock, gfs_lockop_t *ops)
{
  int x;
  int error, ret = 0;


  for (x = 0; x < num_lock; x++)
  {
    error = gfs_gunlock(sdp, ops[x].lock, ops[x].flags);
    if (error)
      ret = error;
  }


  return(ret);
}

