/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name:  gfs_incore.h
**
**    Description:  The GFS in-core structures.
**
*******************************************************************************
******************************************************************************/

#ifndef __GFS_INCORE_DOT_H__
#define __GFS_INCORE_DOT_H__

#include "gfs.h"
#include "gfs_locking.h"
#include "glist.h"
#include "gfs_profile.h"


/*  Constants  */

#define GFS_MAX_RETRIES         (10)
#define GFS_SUPER_MAGIC         (0x4B50)
#define GFS_IHASH_SIZE          (256)
#define GFS_GHASH_SIZE          (256)
#define GFS_MAX_HELD_LOCKS      (1000)
#define GFS_GLOCK_ROLLOVER      (3600)


typedef struct gfs_bitmap_s
{
  unsigned char       *bi_bitmap;
  uint32              bi_start;
  uint32              bi_len;
} gfs_bitmap_t;



typedef struct gfs_rgrpd_s
{
  gfs_rindex_t        rd_ri;            /* Resource Index structure */
  gfs_rgrp_t          rd_rg;            /* Resource Group structure */
  buffer_head_t       **rd_bh;
  gfs_bitmap_t        *rd_bits;
} gfs_rgrpd_t;



typedef struct gfs_glock_s
{
  gfs_list_head_t     gl_next;          /* Pointers to other gfs_glock_t */
  uint64              gl_glock;         /* The lock that this structure represents */
  uint                gl_count;         /* Usage Counter */

  int                 gl_held;          /* True if the lock is held */
  int                 gl_shared;        /* True if lock is shared */
  int                 gl_release;       /* Non zero if the lock should be released */
  int                 gl_incr;          /* Set to 1 if lock needs to be incremented */
  gfs_list_head_t     gl_lru;           /* LRU list of held locks */
  void                *gl_lockdata;     /* Private data for the underlying lock module */

  int                 gl_pid;           /* Process ID of the locking process */
  int                 gl_locked;        /* Number of times Glock has been locked */
  int                 gl_waiting;       /* Number of processes waiting for lock */    
  semaphore_t         gl_lock;
  semaphore_t         gl_wait;

  uint32              gl_vn;            /* The latest value of the version number */

  int                 gl_touched;       /* There are gfs_bhdata_t to be freed */
  gfs_list_head_t     gl_bufs;        	/* Buffer list for caching */
  semaphore_t         gl_buflock;
} gfs_glock_t;



typedef struct gfs_bhdata_s
{
  gfs_list_head_t   	bd_list;	/* Keeps list of buffers */
  gfs_glock_t           *bd_glp;	/* Pointer to Glock struct for this bh */
  buffer_head_t  	*bd_bh;		/* bh which this struct belongs to */
  uint64              	bd_glock;	/* Glock number for this buffer */
  char                  bd_io;          /* True if I/O is happening */
} gfs_bhdata_t;



typedef struct gfs_diroff_s
{
  int                 do_pid;           /* PID of process */
  uint64              do_offset;        /* Offset of next dirent */

  uint32              do_index;         /* Index into hashtable */
  uint32              do_leafoff;       /* Offset in leaf */

  int                 do_use;           /* gfs_diroff in use */
  int                 do_time;          /* Last usage time */

  struct gfs_diroff_s *do_next;
} gfs_diroff_t;



struct gfs_sbd_s;

typedef struct gfs_inode_s
{
  uint64              i_num;            /* Inode number */

  uint32              i_vn;             /* Clock associated with the glock */

  semaphore_t         i_sema;           /* Semaphore to protect i_diroff */
  gfs_diroff_t        *i_diroff;        /* Offsets in dirs for readdir */

  gfs_dinode_t        i_di;             /* Dinode Structure */
  buffer_head_t       *i_didata;        /* Dinode buffer */

  gfs_list_head_t     i_list;	        /* Next inode in the hash list */

  struct gfs_sbd_s    *i_sbd;           /* GFS superblock pointer */
  gfs_vino_t          *i_vino;          /* FS's device independent ino */
} gfs_inode_t;



struct gfs_tune;

typedef struct gfs_sbd_s
{
  gfs_sb_t            sd_sb;            /* Super Block */

  gfs_rgrpd_t         *sd_rgd;          /* Resource Group data */

  uint32              sd_id;	        /* unique id */
  uint32              sd_mrg;           /* mount rgroup */

  uint32              sd_fsb2bb_shift;  /* Shift FS Block numbers to the left by
					   this to get Basic Blocks  */
  uint32              sd_diptrs;        /* Number of pointers in a dinode */
  uint32              sd_inptrs;        /* Number of pointers in a indirect block */

  gfs_list_head_t     sd_gla[GFS_GHASH_SIZE];  /* Array of glock lists */
  semaphore_t         sd_gla_lock[GFS_GHASH_SIZE];  /* Locks for each glock list */

  int                 sd_gl_heldcount;
  gfs_list_head_t     sd_gl_lru;
  semaphore_t         sd_gl_lru_lock;

  gfs_list_head_t     sd_ihash[GFS_IHASH_SIZE];  /* inode hash table */
  semaphore_t         sd_ihash_lock[GFS_GHASH_SIZE];  /* inode hash table semaphore */

  gfs_vino_t          *sd_rooti;        /* FS's root inode */

  gfs_lock_proto_t    *sd_lockstruct;   /* The structure defining the locking type */
  void                *sd_lockdata;     /* locking module's private data */

  semaphore_t         sd_bufd_lock;     /* Semaphore used to start/stop bufd */
  int                 sd_bufd_stop;     /* Flag to tell bufd to stop */

  semaphore_t         sd_id_mutex;      /* mutex taken out when sd_inoded_ip is changed */
  gfs_list_head_t     sd_id_list;       /* Place inodes to be thrown away here */
  semaphore_t         sd_id_work;       /* up() on this when you want inoded to work */
  semaphore_t         sd_id_lock;       /* Semaphore used to start/stop inoded */
  int                 sd_id_stop;       /* Flag to tell inoded to stop */
  int                 sd_id_count;

  gfs_vino_t          *sd_dev;          /* The virtual inode for the device the FS is mounted on */
  gfs_vsb_t           *sd_vsb;          /* FS's device independent sb */

  struct gfs_tune     *tune;
} gfs_sbd_t;



typedef struct gfs_dcached_s
{
  uint32              dc_vn;            /* version number last time this dentry was revalidated */
  uint32              dc_time;          /* Time this dentry was last revalidated */
} gfs_dcached_t;



#endif  /*  __GFS_INCORE_DOT_H__  */

