/******************************************************************************
*******************************************************************************
**
**    Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: main.c
**
*******************************************************************************
******************************************************************************/

#define __NO_VERSION__ 
#include <linux/module.h>
#include <linux/version.h>
char kernel_version [] = UTS_RELEASE;

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/string.h>

#include "gfs_locking.h"
#define DLOCK_VERSION 3
#include "pool.h"



#ifdef ASSERT
#undef ASSERT
#endif  /*  ASSERT  */
#define ASSERT(x) {if (!(x)) { \
		   printk(KERN_EMERG "\n\nYou've found a bug in dlock-0.6!\n\n"); \
		   printk(KERN_EMERG "Please copy down the following line and send it to gfs-devel@lcse.umn.edu\n\n"); \
		   panic("dlock060:  Assertion \"%s\" failed on line %d of file %s\n", #x, \
			 __LINE__, __FILE__); } }

#define REPLY_LEN (8)



typedef struct dlock060_s
{
  char table_name[GFS_LOCKNAME_LEN];

  gfs_callback_t *cb;
  void *fsdata;

  struct dentry *dcache;
  struct inode *inode;

  struct dlock060_s *next;
} dlock060_t;

typedef struct lock_s
{
  uint32 lock;
  unsigned char buf[sizeof(pool_dlock_t) + REPLY_LEN];
  uint32 vn;
  int time;
} lock_t;


dlock060_t *list = NULL;



/******************************************************************************
*******************************************************************************
**
** fill_cdb()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void fill_cdb(unsigned char *buf, int action, uint32 ln, uint32 input_vn, int alloc_len)
{
  buf[0] =  0xA0;

  buf[1] =  action & 0x0F;

  buf[2] =  (ln >> 24) & 0xFF;
  buf[3] =  (ln >> 16) & 0xFF;
  buf[4] =  (ln >> 8) & 0xFF;
  buf[5] =  ln & 0xFF;

  buf[6] =  (input_vn >> 24) & 0xFF;
  buf[7] =  (input_vn >> 16) & 0xFF;
  buf[8] =  (input_vn >> 8) & 0xFF;
  buf[9] =  input_vn & 0xFF;

  buf[10] = alloc_len;

  buf[11] = 0x00;
}


/******************************************************************************
*******************************************************************************
**
** unlock()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static int unlock(dlock060_t *md, uint64 lock, lock_t *data, int incr)
{
  pool_dlock_t *dlock = (pool_dlock_t *)data->buf;
  unsigned char *ret_data = &data->buf[sizeof(pool_dlock_t)];

  struct timeval tv;
  uint32 vn;
  int error;


  memset(ret_data, 0, REPLY_LEN);
  fill_cdb(dlock->dl_cdb, (incr) ? 3 : 2, lock, 0, REPLY_LEN);

  dlock->dl_ln_not_device = TRUE;
  dlock->dl_ln = lock;
  dlock->dl_lnoffset = 2;
  dlock->dl_lnsize = 4;
  dlock->dl_alloc_len = REPLY_LEN;

  error = pool_ioctl(md->inode, NULL, POOL_DLOCKOP, (unsigned long)dlock);
  if (error)
    return(error);

  if (!(ret_data[0] & 0x80))
  {
    int x;
    printk(KERN_ERR "dlock060:  failed unlock\n  ");
    for (x = 0; x < 12; x++)
      printk("%.2X ", dlock->dl_cdb[x]);
    printk("\n  ");
    for (x = 0; x < REPLY_LEN; x++)
      printk("%.2X ", ret_data[x]);
    printk("\n");
    return(-EIO);
  }

  vn = ((ret_data[1] << 24) |
	(ret_data[2] << 16) |
	(ret_data[3] << 8)  |
	(ret_data[4]));
  
  data->vn = vn;


  do_gettimeofday(&tv);
  data->time = tv.tv_sec;


  return(0);
}


/******************************************************************************
*******************************************************************************
**
** dlock060_mount()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void *dlock060_mount(char *table_name, char *host_data, gfs_callback_t *cb, void *fsdata)
{
  dlock060_t *moddata;
  char buf[256];


  MOD_INC_USE_COUNT;



  for (moddata = list; moddata; moddata = moddata->next)
    if (!strcmp(moddata->table_name, table_name))
    {
      printk(KERN_ERR "dlock060:  table_exists\n");
      MOD_DEC_USE_COUNT;
      return(NULL);
    }


  moddata = (dlock060_t *)kmalloc(sizeof(dlock060_t), GFP_KERNEL);
  if (!moddata)
  {
    printk(KERN_ERR "dlock060:  out of memory\n");
    MOD_DEC_USE_COUNT;
    return(NULL);
  }
  memset(moddata, 0, sizeof(dlock060_t));


  strcpy(moddata->table_name, table_name);

  moddata->cb = cb;
  moddata->fsdata = fsdata;


  strcpy(buf, "/dev/pool/");
  strcat(buf, moddata->table_name);


  moddata->dcache = lookup_dentry(buf, NULL, 0);

  if (!moddata->dcache || !moddata->dcache->d_inode)
  {
    if (moddata->dcache)
      dput(moddata->dcache);
    kfree(moddata);
    printk(KERN_ERR "dlock060:  can't open pool %s\n", buf);
    MOD_DEC_USE_COUNT;
    return(NULL);
  }

  moddata->inode = moddata->dcache->d_inode;


  moddata->next = list;
  list = moddata;


  return(moddata);
}


/******************************************************************************
*******************************************************************************
**
** dlock060_unmount()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void dlock060_unmount(void *moddata)
{
  dlock060_t *tmp, *last = NULL;
  int done = FALSE;


  if (list == moddata)
  {
    list = list->next;
    done = TRUE;
  }
  else
    for (tmp = list; tmp; last = tmp, tmp = tmp->next)
      if (tmp == moddata)
      {
	last->next = tmp->next;
	done = TRUE;
      }


  if (!done)
    printk(KERN_ERR "dlock060:  given a weird moddata pointer\n");
  else
  {
    dput(((dlock060_t *)moddata)->dcache);
    kfree(moddata);
    MOD_DEC_USE_COUNT;
  }
}


/******************************************************************************
*******************************************************************************
**
** dlock060_lock()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int dlock060_lock(void *moddata, uint64 lock, int op, int flags, void **lockdata)
{
  dlock060_t *md = (dlock060_t *)moddata;

  lock_t *data;
  pool_dlock_t *dlock;
  unsigned char *ret_data;

  int cont = TRUE;
  int error, ret = -1;
  uint32 vn;
  int new;
  int tries = 0;


  if (*lockdata)
  {
    data = (lock_t *)*lockdata;
    new = FALSE;
  }
  else
  {
    data = (lock_t *)kmalloc(sizeof(lock_t), GFP_KERNEL);
    if (!data)
      return(-ENOMEM);
    memset(data, 0, sizeof(lock_t));
    data->lock = lock;
    *lockdata = data;
    new = TRUE;
  }

  ASSERT(data->lock == lock);

  dlock = (pool_dlock_t *)data->buf;
  ret_data = &data->buf[sizeof(pool_dlock_t)];


  ASSERT(op != LCK_OP_TEST);
  

  if (flags & (LCK_COMMUTE | LCK_COMMUTE_MOD))
  {
    ASSERT(!new);

    error = unlock(md, lock, data, (flags & LCK_COMMUTE_MOD) ? TRUE : FALSE);
    if (error < 0)
      return(error);
  }


  do
  {
    memset(ret_data, 0, REPLY_LEN);
    fill_cdb(dlock->dl_cdb, 1, lock, 0, REPLY_LEN);

    dlock->dl_ln_not_device = TRUE;
    dlock->dl_ln = lock;
    dlock->dl_lnoffset = 2;
    dlock->dl_lnsize = 4;
    dlock->dl_alloc_len = REPLY_LEN;

    error = pool_ioctl(md->inode, NULL, POOL_DLOCKOP, (unsigned long)dlock);
    if (error)
      return(error);

    if ((ret_data[0] & 0xC0) == 0xC0)
    {
      ret = LCKOUT_HELD;

      vn = ((ret_data[1] << 24) |
	    (ret_data[2] << 16) |
	    (ret_data[3] << 8)  |
	    (ret_data[4]));

      if (!new && vn == data->vn)
      {
	struct timeval tv;
	do_gettimeofday(&tv);
	if (tv.tv_sec - data->time < 3600)
	  ret |= LCKOUT_CACHEABLE;
      }

      data->vn = vn;

      ret |= LCKOUT_NEED_E;

      cont = FALSE;
    }
    else
    {
      if ((ret_data[0] & 0xC0) == 0x80)
      {
	int x;
	printk(KERN_ERR "dlock060:  lock success, but no lock (%"FMT_64"u)\n  ", lock);
	for (x = 0; x < 12; x++)
	  printk("%.2X ", dlock->dl_cdb[x]);
	printk("\n  ");
	for (x = 0; x < REPLY_LEN; x++)
	  printk("%.2X ", ret_data[x]);
	printk("\n");
      }

      if (op == LCK_OP_ACQUIRE)
      {
	if (++tries % 1000 == 0)
	  printk(KERN_ERR "dlock060:  asking for lock %"FMT_64"u\n", lock);
      }
      else if (op == LCK_OP_TRY)
      {
	ret = LCKOUT_CACHEABLE;
	cont = FALSE;
      }
    }
  } while (cont);


  return(ret);
}


/******************************************************************************
*******************************************************************************
**
** dlock060_unlock()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int dlock060_unlock(void *moddata, uint64 lock, int flags, void **lockdata)
{
  dlock060_t *md = (dlock060_t *)moddata;
  lock_t *data = (lock_t *)*lockdata;

  int error;


  ASSERT(data);
  ASSERT(data->lock == lock);


  error = unlock(md, lock, data, (flags & ULCK_MODIFIED) ? TRUE : FALSE);


  return(error);
}


/******************************************************************************
*******************************************************************************
**
** dlock060_release()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void dlock060_release(void *moddata, void **lockdata)
{
  kfree(*lockdata);
  *lockdata = NULL;
}


gfs_lock_proto_t dlock060_lock_proto =
{
  "dlock-0.6",
  dlock060_mount,
  dlock060_unmount,
  dlock060_lock,
  dlock060_unlock,
  dlock060_release,
  NULL
};


/******************************************************************************
*******************************************************************************
**
** init_module()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int init_module(void)
{
  int error;

  if ((error = register_lock_proto(&dlock060_lock_proto)) < 0)
    return(error);

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** cleanup_module()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void cleanup_module(void)
{
  unregister_lock_proto(&dlock060_lock_proto);
}



/*  Symbols  */
EXPORT_NO_SYMBOLS;


