/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Mike C Tilstra <tilstra@lcse.umn.edu>
* 
*    Module name: dlock_ip.o
*
*  Various chunks of code that deal with the callbacks and like.
*
*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/locks.h>
#include <linux/smp_lock.h>
#include <linux/net.h>
#include <asm/uaccess.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
#include "dlip.h"


typedef struct {
	u32 tbl;
	u32 lock;
	u32 flags;
} askdrop_t;

/* some globals to keep track of things. */
static struct socket *mastersock = NULL;
static struct semaphore senderlock = MUTEX;
static struct semaphore earmux = MUTEX_LOCKED;
static int has_ears = FALSE;

/* get a hold on the list of loaded tables. */
extern dlock_ip_t *dlock_ip_list;

/******************************************************************************
 * 
 * checks the socket to see if a request to drop came.
 *  If one is here, processes it.
 *****************************************************************************/
int dlip_ears(void *data)
{
	struct msghdr msg;
	struct iovec  iov;
	mm_segment_t oldfs;
	askdrop_t drpt;
	askdrop_t be_drpt;
	int n;
	unsigned long flags;
	dlock_ip_t *tmp = NULL;

	lock_kernel();

	exit_mm(current);

	current->session = 1;
	current->pgrp = 1;

	spin_lock_irqsave(&current->sigmask_lock, flags);
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, flags);

	current->fs->umask = 0;

	sprintf(current->comm, "dlip_ears");

	unlock_kernel();
	/*  Wake up the process who started us  */
	up(&earmux);

	while (has_ears) {
		iov.iov_base = &be_drpt;
		iov.iov_len  = sizeof(askdrop_t);
		msg.msg_name = NULL;
		msg.msg_namelen = 0;
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_control = NULL;
		msg.msg_controllen = 0;
		msg.msg_flags = 0;

		oldfs = get_fs();
		set_fs(get_ds());
		n = sock_recvmsg(mastersock, &msg, sizeof(askdrop_t), 0);
		set_fs(oldfs);

		if(! has_ears )
			break;

		if( n == sizeof(askdrop_t) ) {
			printd("Got something of correct size.(%d) "
					 "Blindly assuming its what we want.\n", n);
			drpt.tbl   = be32_to_cpup(&be_drpt.tbl);
			drpt.lock  = be32_to_cpup(&be_drpt.lock);
			drpt.flags = be32_to_cpup(&be_drpt.flags);

			printd("  tbl = %d, lock = %d, flags = %d\n",
					 drpt.tbl, drpt.lock, drpt.flags);

			drpt.flags &= CB_NEED_S | CB_NEED_E;

			for(tmp = dlock_ip_list; tmp; tmp=tmp->next) {
				if( drpt.tbl == tmp->serv_ip )
					tmp->cb(tmp->fsdata, drpt.lock, drpt.flags );
			}

		} else if( n != -EAGAIN )
			printk(KERN_ERR "dlock_ip: strange message (%d)\n", n);
	}

	up(&earmux);
	return 0;
}

/******************************************************************************
 * init_masterlock()
 * starts up the socket on which we will send & receive requests to drop
 * locks.
 *****************************************************************************/
int init_mastersock(u32 me, u16 port)
{
	struct sockaddr_in meaddr;
	int err;

	err = sock_create(AF_INET, SOCK_DGRAM, 0, &mastersock);
	if(err < 0) {
		printk(KERN_ERR "dlock_ip: can't open socket. (%d)\n", err);
		return err;
	}

	meaddr.sin_family = AF_INET;
	meaddr.sin_addr.s_addr = me;
	meaddr.sin_port = port;

	err = mastersock->ops->bind(mastersock, (struct sockaddr*)&meaddr, 
					sizeof(struct sockaddr_in));
	if(err < 0) {
		sock_release(mastersock);
		printk(KERN_ERR "dlock_ip: can't bind address (%d)\n", err);
		return err;
	}

	if( has_ears == FALSE ) {
		has_ears = TRUE;
		err = kernel_thread(dlip_ears, NULL, 0);
		if( err < 0) {
			printk(KERN_WARNING
					"dlock_ip: unable to start ears thread (%d)\n", err);
			return err;
		}
		down(&earmux); /* wait until started before continuing */
	}

	return 0;
}

/******************************************************************************
 * close_mastersock()
 *  closes that connection.
 *****************************************************************************/
void close_mastersock(void)
{
	has_ears = FALSE;
	down(&earmux); /* wait for the thread to kill itsself. */
	sock_release(mastersock);
}

/******************************************************************************
 * send_askdrop()
 * sends a message, asking to drop that lock.
 *****************************************************************************/
void send_askdrop(u32 to, u16 port, u32 tbl, u32 lock, u32 flags)
{
	struct sockaddr_in toaddr;
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;
	sigset_t oldset;
	unsigned long proc_flags;
	askdrop_t be_drpt;
	int error;
	
	be_drpt.tbl   = cpu_to_be32p(&tbl);
	be_drpt.lock  = cpu_to_be32p(&lock);
	be_drpt.flags = cpu_to_be32p(&flags);

	toaddr.sin_family = AF_INET;
	toaddr.sin_addr.s_addr = to;
	toaddr.sin_port = port;

	iov.iov_base = &be_drpt;
	iov.iov_len = sizeof(askdrop_t);

	msg.msg_name = (struct sockaddr*)&toaddr;
	msg.msg_namelen = sizeof(struct sockaddr_in);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	down(&senderlock);

	oldfs = get_fs();
	set_fs(get_ds());

	spin_lock_irqsave(&current->sigmask_lock, proc_flags);
	oldset = current->blocked;
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);

	do{
		error = sock_sendmsg(mastersock, &msg, sizeof(askdrop_t));
	} while(error == -ECONNREFUSED);

	  spin_lock_irqsave(&current->sigmask_lock, proc_flags);
  current->blocked = oldset;
  recalc_sigpending(current);
  spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);

  set_fs(oldfs);

  up(&senderlock);
}

