/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Mike C Tilstra <tilstra@lcse.umn.edu>
* 
*    Module name: dlock_tcp.o
*
* This is an attempt to actually use dlock 9.4
* And so far it looks like it works.
*
* This is trying to use TCP pipes.  And so far it is working very well.
* very well.
*
******************************************************************************/

#define __NO_VERSION__ 
#include <linux/module.h>
#include <linux/version.h>
char kernel_version [] = UTS_RELEASE;

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/smp_lock.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include "gfs_locking.h"
#include "dlip.h"
#include "dlip_misc.h"
#include "dlip_trans.h"
#include "dlip_cmd_wrapper.h"
#include "dlip_eraser.h"
#include "dlip_proc_stats.h"

MODULE_AUTHOR("Mike Tilstra <tilstra@lcse.umn.edu>");
MODULE_DESCRIPTION("This is a gfs locking module that uses a server over "
"tcp instead of talking with a dlock enabled drive. "
"More info at <http://globalfilesystem.org>");


/******************************************************************************
 * Load time parameters
 *****************************************************************************/
static int refresh_time = 10;
MODULE_PARM(refresh_time,"i");

/******************************************************************************
 * Various globals that this module works on.
 *****************************************************************************/
dlock_tcp_t *dlock_tcp_list = NULL;
static int materia_intact = FALSE;

/******************************************************************************
 * dlip_materia()
 * Sorta a life thread of this module.
 * Right now it just does a refresh for each of the existing tables.
 * There may be a expired check here in the future.  Though with the server
 * doing callbacks, perhaps not.
 *****************************************************************************/
int dlip_materia(void *data)
{
	struct semaphore *mux = (struct semaphore *)data;
	unsigned long flags;
	dlock_tcp_t *tmp;
	int err=0;

	MOD_INC_USE_COUNT;

	lock_kernel();

	exit_mm(current);

	current->session = 1;
	current->pgrp = 1;

	spin_lock_irqsave(&current->sigmask_lock, flags);
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, flags);

	current->fs->umask = 0;

	sprintf(current->comm, "dlip_materia");

	unlock_kernel();
	/*  Wake up the process who started us  */
	up(mux);

	while (err == 0)
	{
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(refresh_time * HZ);

		if( dlock_tcp_list == NULL )
			break;

		for(tmp = dlock_tcp_list; tmp; tmp = tmp->next) {
			err = do_refreshAll(tmp);
			if(err) {
				printe("refesh got error %d\n", err);
				break;
			}
		}
	}

	printk("dlock_tcp: materia ceasing to exist.\n");
	materia_intact = FALSE;
	MOD_DEC_USE_COUNT;
	return 0;
}


/******************************************************************************
 * dlock_tcp_mount()
 *  table_name should be the ip of the dlock server.
 *****************************************************************************/

void *dlock_tcp_mount(char *table_name, char *host_data, gfs_callback_t *cb, void *fsdata)
{
	dlock_tcp_t *moddata;
	int err;
	char *c, name[GFS_LOCKNAME_LEN + 1];

	MOD_INC_USE_COUNT;

	strcpy(name, table_name);
	/*break up table name into ip part and port part */
	c = strstr(name, ":");
	if (!c) {
		printe("tablename should be \"serverip:serverport\"\n");
		goto fail;
	}
	*c++ = 0;

	for (moddata = dlock_tcp_list; moddata; moddata = moddata->next)
		if (!strcmp(moddata->table_name, table_name)) {
			printe("table_exists\n");
			goto fail;
		}

	moddata = (dlock_tcp_t *)kmalloc(sizeof(dlock_tcp_t), GFP_KERNEL);
	if (!moddata) {
		printe("out of memory\n");
		goto fail;
	}
	memset(moddata, 0, sizeof(dlock_tcp_t));

	strcpy(moddata->table_name, table_name);

	moddata->cb = cb;
	moddata->fsdata = fsdata;

	moddata->servport = atoi(c);
	if (!moddata->servport) {
		printe("bad port number\n");
		goto memfail;
	}
	cpu_to_be16s(&moddata->servport);

	if( inet_aton( name, &moddata->servip ) <0) {
		printe("Bad IP number for dlock server.\n");
		goto memfail;
	}

	if( trans_start_up( moddata ) <0) {
		printe("socket works failed\n");
		goto memfail;
	}
  
	printdd("Getting modepage\n");
	{
		u32 numlock, lock_timeout;
		u8  maxclnts;
		if( (err = do_modepage(moddata, &maxclnts,
					&numlock, &lock_timeout)) < 0 ) {
			printe("mode page failed! (%d)\n", err);
			goto transfail;
		}
		printk("dlock_tcp: using dlock_tcpd server %s\n"
				 "           with %d locks\n"
				 "           %d cids in each list\n"
				 "           and %d ms timeouts\n",
				 moddata->table_name,
				 numlock,
				 maxclnts,
				 lock_timeout);
	}

	if(materia_intact == FALSE) {
		struct semaphore mux = MUTEX_LOCKED;
		err = kernel_thread(dlip_materia, &mux, 0);
		if(err < 0) {
			printe("unable to start materia thread (%d)\n", err);
			goto transfail;
		}
		materia_intact = TRUE;
		down(&mux);
	}

	moddata->next = dlock_tcp_list;
	dlock_tcp_list = moddata;

	return(moddata);

transfail:
	trans_shutdown(moddata);
memfail:
	kfree( moddata );
fail:
	MOD_DEC_USE_COUNT;
	return NULL;
}


/******************************************************************************
 * dlock_tcp_unmount()
 *****************************************************************************/

void dlock_tcp_unmount(void *moddata)
{
	dlock_tcp_t *tmp, *last = NULL;
	int done = FALSE;

	if (dlock_tcp_list == moddata) {
		dlock_tcp_list = dlock_tcp_list->next;
		done = TRUE;
	}
	else
		for (tmp = dlock_tcp_list; tmp; last = tmp, tmp = tmp->next)
			if (tmp == moddata) {
				last->next = tmp->next;
				done = TRUE;
			}

	if (!done)
		printe("given a weird moddata pointer\n");
	else {
		trans_shutdown(moddata);
		kfree(moddata);
		MOD_DEC_USE_COUNT;
	}


}

/******************************************************************************
 * unlock()
 *****************************************************************************/

static int unlock(dlock_tcp_t *md, uint64 lock, lock_t *data, int incr)
{
	struct timeval tv;
	int error;
	dlock_return_t rtn;

	printdd(" unlocking %d, by %d\n", lock, current->pid);
	gunlock_inc;

	error = do_dlock(md, data, ((incr)?6:5), lock, &rtn);
	if (error)
		return(error);
	if(! rtn.res )
		return(-EIO);

	data->vn = rtn.vn;

	do_gettimeofday(&tv);
	data->time = tv.tv_sec;

	printdd("Exiting from unlock\n");
	return(0);
}

/******************************************************************************
 * dlock_tcp_lock()
 * Translates gfs's glocks into ipdlocks.
 *****************************************************************************/

int dlock_tcp_lock(void *moddata, uint64 lock, int op, int flags,
		void **lockdata)
{
	dlock_tcp_t *md = (dlock_tcp_t *)moddata;
	lock_t *data;
	dlock_return_t rtn;
	int cont = TRUE;
	int error, ret = -1;
	int new;
	int tries = 0;
	unsigned long step = 0;

	printdd(" locking %d, by %d\n", lock, current->pid);
	glock_inc;

	if (*lockdata) {
		data = (lock_t *)*lockdata;
		new = FALSE;
	} else {
		data = (lock_t *)kmalloc(sizeof(lock_t), GFP_KERNEL);
		if (!data)
			return(-ENOMEM);
		memset(data, 0, sizeof(lock_t));
		data->lock = lock;
		*lockdata = data;
		new = TRUE;
	}

	ASSERT(data->lock == lock );

	if( (op != LCK_OP_TEST) && (flags & (LCK_COMMUTE | LCK_COMMUTE_MOD)) ){
		ASSERT(!new);

		error = unlock(md, lock, data, (flags & LCK_COMMUTE_MOD) ? TRUE : FALSE);
		if (error < 0)
			return(error);
	}

	do {
		if(op == LCK_OP_TEST) {
			/* Nop the lock */
			error = do_dlock(md, data, 0, lock, &rtn);
			cont = FALSE;
		} else
			error = do_dlock(md, data, (flags & LCK_SHARED)?1:2, lock, &rtn);
		if (error) return(error);

		if ( rtn.res ) {
			ret = LCKOUT_HELD;
#ifndef USE_CALLBACKS
			ret |= LCKOUT_NEED_E;
#endif

			if (!new && rtn.vn == data->vn) {
				struct timeval tv;
				do_gettimeofday(&tv);
				if (tv.tv_sec - data->time < 3600)
					ret |= LCKOUT_CACHEABLE;
			}

			data->vn = rtn.vn;

			switch(rtn.state) {
				case 0:
					printe("Ack! Unlocked state after acquire\n");
					ASSERT(0);
					break;
				case 1:
					ASSERT(flags & LCK_SHARED);
					ret |= LCKOUT_SHARED;
					break;
				case 2:
					ASSERT(!(flags & LCK_SHARED));
					break;
				default:
					printe("weird state value\n");
					break;
			}

			switch(rtn.exp) {
				case 0:
					break;
				case 1:
					ret |= LCKOUT_EXPIRED_S;
					break;
				case 2:
					ret |= LCKOUT_EXPIRED_E;
					break;
				default:
					printe("weird expired value\n");
					break;
			}

			cont = FALSE;
		} else {
			if (op == LCK_OP_ACQUIRE) {
				if (tries % 1000 == 0)
				  printe(" asking for lock %"FMT_64"u\n", lock);
				tries++;
				/* Need to code a backoff routine */
				current->state = TASK_INTERRUPTIBLE;
				step |= 1 << tries;
				schedule_timeout( MIN(step,(2*HZ)) );
			}
			else if (op == LCK_OP_TRY) {
				ret = LCKOUT_CACHEABLE;
				cont = FALSE;
			}
		}
	} while (cont);
	
	printdd("Exiting dlock_tcp_lock\n");
	return(ret);
}


/******************************************************************************
 * dlock_tcp_unlock()
 * Unlocks a lock.
 *****************************************************************************/

int dlock_tcp_unlock(void *moddata, uint64 lock, int flags, void **lockdata)
{
	dlock_tcp_t *md = (dlock_tcp_t *)moddata;
	lock_t *data = (lock_t *)*lockdata;
	int error;

	ASSERT(data);
	ASSERT(data->lock == lock);

	error = unlock(md, lock, data, (flags & ULCK_MODIFIED) ? TRUE : FALSE);

	printdd("Exiting dlock_tcp_unlock\n");
	return(error);
}


/******************************************************************************
 * dlock_tcp_release()
 * cleans up the lockdata when called.
 *****************************************************************************/
void dlock_tcp_release(void *moddata, void **lockdata)
{
#if 0
	lock_t *data = (lock_t*)*lockdata;
#endif
	kfree(*lockdata);
	*lockdata = NULL;
}

/******************************************************************************
 *****************************************************************************/
gfs_lock_proto_t dlock_tcp_lock_proto = {
	"dlock_tcp",
	dlock_tcp_mount,
	dlock_tcp_unmount,
	dlock_tcp_lock,
	dlock_tcp_unlock,
	dlock_tcp_release,
	NULL
};

/******************************************************************************
 * The beginning and the end of this module.
 *****************************************************************************/
int init_module(void)
{
	int error;

	/* load the gfs lock module */
	if ((error = register_lock_proto(&dlock_tcp_lock_proto)) < 0)
		return(error);

#ifdef USE_CALLBACKS
	if((error = create_eraser()) <0)
		return error;
#endif

	dlip_proc_init;
	printk("dlock_tcp: module loaded\n");
	return(0);
}

void cleanup_module(void)
{
	dlip_proc_uninit;
#ifdef USE_CALLBACKS
	destroy_eraser();
#endif
	unregister_lock_proto(&dlock_tcp_lock_proto);
	printk("dlock_tcp: module unloaded\n");
}


/*  Symbols  */
#ifndef DEBUG
EXPORT_NO_SYMBOLS;
#endif

