/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Mike C Tilstra <tilstra@lcse.umn.edu>
*
* This was originally intended to be clear of dlip stuff, but the handling
* of callbacks has infested itself into here.  I shall try to keep it
* clean.
*/ 

/*#include <linux/sched.h>*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include <linux/net.h>
#include <linux/in.h>
#include <linux/socket.h>
#include <net/sock.h>

#include "dlip.h"
#include "dlip_trans.h"
#include "dlip_data_structures.h"
#include "dlip_eraser.h"
#include "dlip_proc_stats.h"

/******************************************************************************
 * prototype the data_xmit funtion here so I can leave it at the bottom of
 * this file.
 */
int data_xmit(int send, struct socket *sock, char *buf, int size);
#define DXwrite 0
#define DXread  1
#define DXpeek  2

/******************************************************************************
 * trans_io()
 */
int trans_io(dlock_tcp_t *md,
		char *ibuf, int icnt,
		char *obuf, int ocnt)
{
	int err = 0;

	/* can't tranfer data without transiod running. */
	if(md->recvd_run == FALSE) {
		printe("transiod is down, communications are broken.\n");
		return -EIO;
	}

	down(&md->sender);
	printdd("    md->sender down\n");

	dltcp_inc;

	/*setup reply space */
	md->recv_buf = obuf;
	md->recv_len = ocnt;

	md->recv_err = 0;
	
	/*send buffer*/
	data_xmit(DXwrite, md->sock, ibuf, icnt);

	/* wait for reply */
	down(&md->recver);
	if(md->recv_err != 0)
		err = md->recv_err;

	/* we'll have to assume that things are correct from here */

	up(&md->sender);
	printdd("    md->sender up\n");
	return err;
}


/******************************************************************************
*******************************************************************************
 * This below is the mechanisism for which callbacks are executed.  This
 * mess of queuing the callbacks is required so that the transiod thread is
 * free to receive the data transmissions that will be the unlock replay to
 * the unlock command that this call back started.  If this callback didn't
 * detach itself from the transiod thread in some maner, the thread would
 * never be able to receive the unlock reply from the server, and gets
 * stuck.
 *
 * this stucture is to pass the needed info into the queued function.
 */
#ifdef USE_CALLBACKS
typedef struct {
	gfs_callback_t *cb;
	void *fs;
	u32 lock;
	u32 flags;
} qcb_t;

/******************************************************************************
 * queued_callback()
 *
 * the super simple function that actually does the callback into gfs.
 */
void queued_callback(void *args)
{
	qcb_t *q = (qcb_t *)args;

	ASSERT( q != NULL );
	ASSERT( q->cb != NULL );

	printe("Telling gfs about %lu\n", q->lock);
	q->cb(q->fs, q->lock, q->flags );
	printe("Told gfs about %lu\n", q->lock);

}
/******************************************************************************
 * do_callback()
 *
 * Need to add suport somehow for the CB_FREED tags.
 */
void do_callback(dlock_tcp_t *md, u32 lock, u32 flags)
{
	static qcb_t qcb;
	static struct tq_struct cb;
	/* should do this some other way than statics, but I can't just alloc
	 * mem since there isn't a way  to free it.  
	 * Have to think on this, but statics should work in the mean time.
	 */
	
	printd("   Lockdrop request is %d, flags are 0x%x\n",
		lock, flags);

	flags &= (CB_NEED_S | CB_NEED_E);
	if( flags ) { /* don't tell gfs if its not the above flags */

		qcb.cb    = md->cb;
		qcb.fs    = md->fsdata;
		qcb.lock  = lock;
		qcb.flags = flags;

		cb.next    = NULL;
		cb.sync    = 0;
		cb.routine = queued_callback;
		cb.data    = (void*)&qcb;

		printdd("Queueing to eraser queue\n");
		eraser_queue(&cb);
	
	}
}
#endif /*USE_CALLBACKS*/
/******************************************************************************
******************************************************************************/


/******************************************************************************
 * transiod()
 * 
 * communications works like this:
 * trans_io is called,
 * sender_mux is locked.
 * data is xmitted
 * recv_mux is waited on
 * trans_listener thread gets data,
 * if it is a server code
 *   do the cmd.
 * else
 *   must be the reply to trans_io
 *   stuff recieved data into a store
 *   up recv_mux
 * loop
 *
 * TODO: Better error handling need to be done in this thread.
 */
int transiod(void *data)
{
	dlock_tcp_t *md = (dlock_tcp_t *) data;
	char cbuf[24]; /* should be bigger than anything we receive */
	int err;

	printdd(">>transiod started\n");
	/*  Set up the thread  */
	lock_kernel();

	exit_mm(current);

#if 0
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGKILL));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
#endif

	current->session = 1;
	current->pgrp = 1;
	current->fs->umask = 0;

	sprintf(current->comm, "dlip_transiod");

	unlock_kernel();

	up( &md->tiodmux );

	/* now wait for data */
	while( md->recvd_run) {
		printdd(">>top of transiod loop\n");
		if(md->sock == NULL) break;
		err = data_xmit(DXpeek, md->sock, cbuf, 1);
		if(err==0) {
			printe("Peeked a size zero!\n");
			md->recvd_run = FALSE;
			break;
		}
		if(err<0) {
			md->recv_err = err;
			printe("Error peeking data (%d)\n",err);
		}
		if(! md->recvd_run)
			break;

		if( cbuf[0] == CodeLess ) {
			printe("Got a codeless field! Barfing.\n");
			md->recvd_run = FALSE;
			break;
		} else if( cbuf[0] == CallBack ) {
			drop_rdf_t bdrp;
			printdd(">>Handling CallBack\n");
			if(md->sock == NULL) break;
			err = data_xmit(DXread, md->sock, (char*)&bdrp, sizeof(drop_rdf_t) );
			if(err==0) {
				printe("Read a size zero!\n");
				md->recvd_run = FALSE;
				break;
			}
			if(err<0) {
				md->recv_err = err;
				printe("Error reading callback (%d)\n",err);
			}
				/* this call needs to be broken off of this thead, if it isn't
				 * we'll never be able to receive the replys needed to finish
				 * the callback.
				 */
#ifdef USE_CALLBACKS
				do_callback(md, be32_to_cpu( bdrp.lock), be32_to_cpu( bdrp.flags) );
#else
				printd("Callback ignored\n");
#endif

		} else {
			printdd(">>Handling RDF\n");
			if(md->sock == NULL) break;
			err = data_xmit(DXread, md->sock, cbuf, md->recv_len);
			if(err==0) {
				printe("Read a size zero!\n");
				md->recvd_run = FALSE;
				break;
			}
			if(err<0) {
				md->recv_err = err;
				printe("Error reading reply (%d)\n",err);
			}
			memcpy(md->recv_buf, cbuf, md->recv_len);
			up(&md->recver);
		}
	} /*while(md->recvd_run)*/

	printk("dlock_tcp: Exiting transiod thread\n");
	up( &md->tiodmux );
	return 0;
}

/******************************************************************************
 * trans_start_up()
 * starts up the trans_recvd thread
 *
 * Should redo this to be like the rpciod_up function?
 */
int trans_start_up(dlock_tcp_t *md)
{
	int err;
	struct sockaddr_in addr;

	err = sock_create(AF_INET, SOCK_STREAM, 0, &md->sock);
	if(err) {
		printe("Failed to create socket (%d)\n", err);
		return err;
	}

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = md->servip;
	addr.sin_port = md->servport;

	err = md->sock->ops->connect(md->sock, (struct sockaddr*)&addr,
			sizeof(addr), 0);
	if(err) {
		sock_release(md->sock);
		printe("Unable to connect to server (%d)\n", err);
		return err;
	}

	/* find out what our IP is and stuff it into the cid 
	 * This is quite linux specific.
	 */
	{
		struct sock *tsk;
		tsk = md->sock->sk;
		md->mycid = cpu_to_be32(tsk->rcv_saddr);

		printd("My ip should be (in big endian) %x\n", md->mycid);
	}

	/* setup some other things */
	md->sender = MUTEX;
	md->recver = MUTEX_LOCKED;
	md->tiodmux = MUTEX_LOCKED;
	md->recvd_run  = FALSE;
	md->recv_buf   = NULL;
	md->recv_len   = 0;
	md->recv_err   = 0;

	/* start up reciever thread */
	if(md->recvd_run == FALSE) {
		md->recvd_run = TRUE;
		err = kernel_thread(transiod, md, 0);
		if(err < 0) {
			sock_release(md->sock);
			printe("Couldn't start transiod thread (%d)\n", err);
			return err;
		}

		/* wait for thread to get itself set up before continuing. */
		down( &md->tiodmux ); 
	}

	return 0;
}

/******************************************************************************
 * trans_shutdown()
 *
 * In order to get the transiod out of it's sleep on the socket, we send a
 * dlock nop to the server, which will reply with a rather empty cdb.  This
 * wakes the thread from its slumber, where it sees that the run flag is
 * now turned off, and exits.  Upping the mux we're waiting on, where we
 * then relase the socket and now the comm layer is shutdown.
 *
 * Perhaps create an actual shutdown command that gets passed to the
 * server?
 */
void trans_shutdown(dlock_tcp_t *md)
{
	dlock_cdb_t nop;

	if(md->recvd_run) { /* don't do this if it already stopped */
		md->recvd_run = FALSE;

		memset(&nop, 0, sizeof(dlock_cdb_t));
		nop.code = CDBdlock;
		
		/* This is all very hacker-ish,
		 * kids, do not try this at home.
		 */
		data_xmit(DXwrite, md->sock, (char*)&nop, sizeof(dlock_cdb_t));
		
		down( &md->tiodmux );
	}
	sock_release(md->sock);
	md->sock = NULL;
}

/******************************************************************************
 * data_xmit()
 * 
 * This function sends or recvs the specified amount of data.
 * This was based on the ndb xmit function.
 *****************************************************************************/
int data_xmit(int send, struct socket *sock, char *buf, int size)
{
	mm_segment_t oldfs;
	int result = -EIO;
	struct msghdr msg;
	struct iovec iov;
	unsigned long flags;
	sigset_t oldset;

	oldfs = get_fs();
	set_fs(get_ds());

	spin_lock_irqsave(&current->sigmask_lock, flags);
	oldset = current->blocked;
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, flags);


	do {
		iov.iov_base = buf;
		iov.iov_len = size;
		msg.msg_name = NULL;
		msg.msg_namelen = 0;
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_control = NULL;
		msg.msg_controllen = 0;
		msg.msg_namelen = 0;
		msg.msg_flags = 0;

		if(send == DXwrite)
			result = sock_sendmsg(sock, &msg, size);
		else if(send == DXread)
			result = sock_recvmsg(sock, &msg, size, 0);
		else if(send == DXpeek)
			result = sock_recvmsg(sock, &msg, size, MSG_PEEK);

		if (result <= 0) {
			printk(KERN_ERR "%s: %s - sock=%ld at buf=%ld, size=%d returned %d.\n",
					"tcp_dlock_ip", (send==DXwrite)?"send":"receive",
					(long) sock, (long) buf, size, result);
			break;
		}
		size -= result;
		buf += result;
	} while (size > 0);

	spin_lock_irqsave(&current->sigmask_lock, flags);
	current->blocked = oldset;
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, flags);

	set_fs(oldfs);
	return result;
}

