/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Manish Agarwal <agarwal@lcse.umn.edu>
** 
**    Module name: gfs_balancer.c
**   The balancer uses ioctl commands to gather stats about the subpools, 
**   and move files around in the subpools to balance the  amount of data 
**   stored on each subpool.
**
*******************************************************************************
******************************************************************************/

#ifdef linux
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <sys/stat.h>
#include <sys/dir.h>
#endif

#include <ctype.h>
#include "global.h"

#define WANT_DFLAG_T
#include "pool_debug.h"

#include "pool.h"

#include "gfs_debug.h"
#include "gfs_ioctl.h"
#include "gfs_balancer.h"


/* Preprocessor Defines */

/* command line options for gfs_tool - balance */
#define OPTION_STRING   "vhl:s:f:"

/* Function Macros */


/* Type Definitions */


/* Global Variables */

static char 	fullpath[GFS_MAXPATHLEN];
static char 	*sp_state;
static float	gfs_load_margin;
static uint64 	gfs_min_filesize;
static uint64	gfs_max_files;
static char    	*optarg;
static int     	optind, opterr, optopt;
static int 	Verbose;

typedef int Func(char*,struct stat *, gfs_fdat_t **, uint32);
static Func gfs_findfiles;


const char *use[]=
{
  "File System Balancer:\n",
  "  gfs_balancer <GFS Directory> [-l <load margin>] [-s <min filesize>] [-f <max files>] [-v --verbose] [-h --help]\n",
  "\n",
  ""
};


const char *version[]=
{
  "GFS File System Balancer\n",
  "Version 1.0\n\n",
  ""
};




/* Function Prototypes */
static int dopath(Func * func, gfs_fdat_t **flists, uint32 num_sps);

extern void fill_params(gfs_lnkname_t * lnk, char * istr);
extern int  gfs_show_subpools_i(gfs_spstatlist_t *spstatlistp, int fd);
extern void gfs_mv_file(gfs_lnkinfo_t *filep, int verbose);




/******************************************************************************
*******************************************************************************
**
** int gfs_print_graph()
**
** Description:
**   This function prints the status of the subpools in a graphical manner.
**
** Input(s):
**   spstatlistp - the subpool information
**
** Returns : -
**
*******************************************************************************
******************************************************************************/

void gfs_print_graph(gfs_spstatlist_t *spstatlistp)
{
  uint32 i;
  int  	 j;
  float  percent;

  for(i = 0; i < spstatlistp->sp_cnt; i++)
    {
      percent = 100.0*(1.0 - (float)(spstatlistp->sp_list[i].bfree*1.0/spstatlistp->sp_list[i].blocks));
      
      fprintf(stdout, "Subpool %2"FMT_32"d :",i);
      for(j = 0; j < (int)percent; j++)
	fprintf(stdout, "x");
      fprintf(stdout, "\n");
      fprintf(stdout,"percent = %.2f \n\n", percent);
    }

  return;
}



/******************************************************************************
*******************************************************************************
**
** int gfs_ins_splist()
**
** Description:
**   This function allocates a node and inserts it into the sp list 
**   (in descending order).
**
** Input(s):
**   lstpp - list pointer (in which the node is to be inserted.
**   spnum - the subpool number to be inserted.
**   nblks - no. of data blks to be transfered from/to the subpool. 
*******************************************************************************
******************************************************************************/

void gfs_ins_splist(gfs_spdat_t ** lstpp, uint32 spnum, uint64 nblks)
{
  gfs_spdat_t *cptr,*lptr,*nodeptr;


  nodeptr = ( gfs_spdat_t *)malloc(sizeof( gfs_spdat_t));
  nodeptr->spnum = spnum;
  nodeptr->nblks = nblks;


  lptr = cptr = *lstpp;


  if((cptr == NULL) || (nblks >= cptr->nblks) )
    {
      *lstpp = nodeptr;
      nodeptr->next = cptr;
      return;
    }


  while((cptr != NULL) && (nblks < cptr->nblks))
    {
      lptr = cptr;
      cptr = cptr->next;
    }

  lptr->next = nodeptr;
  nodeptr->next = cptr;

  return;
}



/******************************************************************************
*******************************************************************************
**
** int gfs_freesplist()
**
** Description:
**   This function frees all the nodes in the list.
**
** Input(s):
**   lstpp - list pointer.
**
*******************************************************************************
******************************************************************************/

int gfs_freesplist(gfs_spdat_t **lstp)
{
  gfs_spdat_t  *lptr;

  lptr = *lstp;

  while((*lstp) != NULL)
    {
      *lstp = (*lstp)->next;
      free(lptr);
      lptr = *lstp;
    }

  return 0;
}



/******************************************************************************
*******************************************************************************
**
** int gfs_find_sps()
**
** Description:
**   This function searches the subpools and groups them into two categories 
**   1. Subpools from which the data should be moved out.
**   2. Subpools to which the data should be moved in.
**
** Input(s):
**   spstatlistp - the subpool information
**  
** Outputs :
**   abovep - List containing the details for subpool from category 1.
**   belowp - List of subpools belonging to category 2.
**
** Returns : -
**   0  - Success.
**   1  - Cannot do anything.
*******************************************************************************
******************************************************************************/

int gfs_find_sps(gfs_spstatlist_t *spstatlistp, gfs_spdat_t **abovepp, gfs_spdat_t **belowpp)
{
  uint64 avail = 0, full = 0;
  uint32 j;
  float  f, avrg;


  for(j = 0; j < spstatlistp->sp_cnt; j++)
    {
      avail += spstatlistp->sp_list[j].blocks;
      full += (spstatlistp->sp_list[j].blocks - spstatlistp->sp_list[j].bfree);
    }


  avrg = (float)(1.0*full)/(1.0*avail);


  for(j = 0; j < spstatlistp->sp_cnt; j++)
    {
      f = (float)(1.0*spstatlistp->sp_list[j].blocks - 1.0*spstatlistp->sp_list[j].bfree)/(1.0*spstatlistp->sp_list[j].blocks);



      /* gfs balancer accuracy */
      if( GFS_ABS(f - avrg) < (float)gfs_load_margin) 
	{
	  sp_state[j] = GFS_AVG;
	  continue;
	}


      if(f > avrg)
	{
	  sp_state[j] = GFS_ABOVE;
	  gfs_ins_splist(abovepp, j, (uint64)((f-avrg)*spstatlistp->sp_list[j].blocks));
	}
      else
	{
	  sp_state[j] = GFS_BELOW;
	  gfs_ins_splist(belowpp, j, (uint64)((avrg-f)*spstatlistp->sp_list[j].blocks));
	}

    }

  if( (*abovepp == NULL) || (*belowpp == NULL) ) 
    return(1); /* Terminating Condition */


  return(0);
}



/******************************************************************************
*******************************************************************************
**
** int init_sp_state()
**
** Description:
**   Similar to gfs_find_sps but only initialises the sp_state global variable.
**
*******************************************************************************
******************************************************************************/

int init_sp_state(char * pathname)
{
  gfs_spstatlist_t spstatlist;
  uint64           avail = 0, full = 0;
  uint32           j;
  float            f, avg;
  int              fd, flag1=0, flag2=0;


  if ((fd = open(pathname, O_RDONLY)) < 0)
    return(1);

  
  if(gfs_show_subpools_i(&spstatlist,fd))
    {
      close(fd);
      return(1);
    }


  for(j = 0; j < spstatlist.sp_cnt; j++)
    {
      avail += spstatlist.sp_list[j].blocks;
      full += (spstatlist.sp_list[j].blocks - spstatlist.sp_list[j].bfree);
    }


  avg = (float)(1.0*full)/(1.0*avail);


  for(j = 0; j < spstatlist.sp_cnt; j++)
    {
      f = (float)(1.0*spstatlist.sp_list[j].blocks - 1.0*spstatlist.sp_list[j].bfree)/(1.0*spstatlist.sp_list[j].blocks);


      /* gfs balancer accuracy */
      if( GFS_ABS(f - avg) < (float)gfs_load_margin) 
	{
	  sp_state[j] = GFS_AVG;
	  continue;
	}


      if(f > avg)
	{
	  flag1++;
	  sp_state[j] = GFS_ABOVE;
	}
      else
	{
	  flag2++;
	  sp_state[j] = GFS_BELOW;
	}
    }


  if( (flag1 == 0) || (flag2 == 0))
    return(1);


  return(0);
}



/******************************************************************************
*******************************************************************************
**
** int print_splist()
**
** Description:
**   Prints all the nodes in the sp list - used for Verbose mode of operation.
**
*******************************************************************************
******************************************************************************/

void print_splist(gfs_spdat_t *lstpp)
{
  gfs_spdat_t *cptr = lstpp;

  while(cptr != NULL)
    {
      printf("%"FMT_32"d %"FMT_64"d\n", cptr->spnum, cptr->nblks);
      cptr = cptr->next;
    }
  return;
}



/******************************************************************************
*******************************************************************************
**
** General Utility functions.
**
** Description:
**  These functions are used to manipulate the flist.
**
** Input(s):
**   pathname  	- the directory to start from.
**   func	- function to be called for each file in the tree.
**
*******************************************************************************
******************************************************************************/

/******************************************************************************
** gfs_ins_flist - inserts a node into the flist.
******************************************************************************/
int gfs_ins_flist(gfs_fdat_t **flist, char *pathname, uint64 blks)
{
  gfs_fdat_t 	*cptr,*lptr,*nodeptr;
  static uint64	maxfiles = 0;


  nodeptr = ( gfs_fdat_t * )malloc(sizeof( gfs_fdat_t ));
  if(nodeptr == NULL)
    {
      printf("Out of memory\n");
      return(-1);
    }


  nodeptr->fname = malloc(strlen(pathname) + 1);
  if(nodeptr->fname == NULL)
    {
      free(nodeptr);
      printf("Out of memory - 2 \n");
      return(-1);
    }


  if(++maxfiles > gfs_max_files)
    {
      maxfiles = 0;
      return(GFS_DONE);
    }


  strcpy(nodeptr->fname, pathname);
  nodeptr->blks = blks;


  lptr = cptr = *flist;


  if((cptr == NULL) || (blks >= cptr->blks) )
    {
      *flist = nodeptr;
      nodeptr->next = cptr;
      return 0 ;
    }


  while((cptr != NULL) && (blks < cptr->blks))
    {
      lptr = cptr;
      cptr = cptr->next;
    }


  lptr->next = nodeptr;
  nodeptr->next = cptr;


  return 0;
}

/******************************************************************************
** printflist - Prints all the nodes in the flist - used for debugging.
******************************************************************************/
uint64 printflist(gfs_fdat_t **flists, uint32 num)
{
  uint64 ctr = 0;
  uint32 i;
  gfs_fdat_t *cptr;


  for(i = 0; i < num; i++)
    {
      printf("Subpool %"FMT_32"d \n\n",i); 
      cptr = flists[i];

      while(cptr != NULL)
	{
	  printf("%"FMT_64"d  %s\n", cptr->blks, cptr->fname);
	  cptr = cptr->next;
	  ctr ++;
	}
      printf("\n\n");
    }

  printf(" Done - %"FMT_64"d!!\n",ctr);
  return(ctr);
}

/******************************************************************************
** freeflist - frees all the nodes from the flist.
******************************************************************************/
void freeflist(gfs_fdat_t **flists, uint32 num)
{
  uint32 i;
  gfs_fdat_t *cptr;

  for(i = 0; i < num; i++)
    {
      cptr = flists[i];
      
      while(flists[i] != NULL)
	{
	  cptr = (flists[i])->next;
	  free((flists[i])->fname);
	  free(flists[i]);
	  flists[i] = cptr;
	}
    }
  
  if(Verbose)
    printf(" Finished freeing stuff \n");
  return;
}



/******************************************************************************
*******************************************************************************
**
** void gfs_files_to_mv()
**
** Description:
**   This routine chooses the appropriate files and moves them.
**   
** Input(s):
**  spstatlistp - the status of the structures.
**  flists 	- list of files.
**  abovep 	- list of subpools to move data from.
**  belowp 	- list of subpools to move data to.
**
** Returns :
**    0     - failed.
**   else   - successful.
**
*******************************************************************************
******************************************************************************/

uint64 gfs_files_to_mv(gfs_spstatlist_t * spstatlistp, gfs_fdat_t **flists, gfs_spdat_t **abovepp, gfs_spdat_t **belowpp)
{
  uint64 	fsize;
  uint32        sp_from, sp_to;
  gfs_spdat_t 	*cabove, *cbelow, *temp;
  gfs_fdat_t 	*fptr;
  gfs_lnkinfo_t file; /* flist contains files with link count 1*/


  /* 
   * 1. Pick the first node from abovep & belowp 
   * 2. Decide the size of the movement - delete the nodes. Re-insert
   *    which has a higher transfer size (with new no. of blocks).
   * 3. Choose the files and reduce both the counters assuming that the
   *    move would be successful.
   * 4. If the counters reduce below the MIN file size, then delete the node
   * 5. Repeat till one of the lists becomes empty.
   */


  file.lnk_cnt = 1;
  file.l_name =(gfs_lnkname_t *) malloc(sizeof(gfs_lnkname_t));
	  

  if(file.l_name == NULL)
    {
      fprintf(stderr, "%s: unable to allocate memory on line %d.\n", __FILE__ , __LINE__);
      return(0);
    }


  while( (*abovepp != NULL) && (*belowpp != NULL) )
    {
      cabove = *abovepp;
      cbelow = *belowpp;

      sp_from = cabove->spnum;
      sp_to   = cbelow->spnum;


      /* No files found for this subpool */
      if(flists[sp_from] == NULL) 
	{
	  *abovepp = (*abovepp)->next;
	  free(cabove);
	  cabove = *abovepp;
	  continue;
	}



      /* 
       *  decide the transfer size and manipulate the lists accordingly 
       */
      if( cabove->nblks > cbelow->nblks)
	{
	  fsize = cbelow->nblks;
	  *belowpp = (*belowpp)->next;
	  free(cbelow);
	  cbelow = *belowpp;


	  /* Update the abovep node and re-sort it i.e. del & re-ins  */
	  *abovepp = (*abovepp)->next;
	  temp = cabove;
	  temp->nblks -= fsize;
	  cabove = *abovepp;
	  gfs_ins_splist(abovepp, temp->spnum, temp->nblks);
	  free(temp);
	}
      else
	{
	  fsize = cabove->nblks;
	  *abovepp = (*abovepp)->next;
	  free(cabove);
	  cabove = *abovepp;


	  /* Update the belowp node and re-sort it i.e. del & re-ins  */
	  *belowpp = (*belowpp)->next;
	  temp = cbelow;
	  temp->nblks -= fsize;
	  cbelow = *belowpp;
	  gfs_ins_splist(belowpp, temp->spnum, temp->nblks);
	  free(temp);
	}



      fptr = flists[sp_from];
      while(fsize > 0)
	{

	  /* 
	   * Remove the files with size greater than the transfer size
	   */
	  while( (fptr != NULL ) && (fptr->blks > fsize) )
	    {
	      flists[sp_from] = fptr->next;
	      free(fptr->fname);
	      free(fptr);
	      fptr = flists[sp_from];
	    }
	      
	  /* Nothing left ? */
	  if(fptr == NULL)
	    break;


	  fsize -= fptr->blks;
	  

	  /* Move the file */
	  file.d_sp = sp_to;
	  fill_params(file.l_name, fptr->fname);
	  gfs_mv_file(&file, Verbose);


	  /* remove the file from flist */
	  flists[sp_from] = fptr->next;
	  free(fptr->fname);
	  free(fptr);
	  fptr = flists[sp_from];

	} /* End of inner while loop */

    } /* End of outer while loop */
  

  free(file.l_name);
  return(1);
}



/******************************************************************************
*******************************************************************************
**
** void gfs_ftw()
**
** Description:
**   This is an implementation of the ftw (file tree walk) library function,
**   it was required because ftw does not distinguish betweens links and 
**   actual directories.
**   i.e. say if the initial directory (/) had a link /lib to  /usr/lib then
**   the files in /usr/lib would be accounted for twice.
**
**   PS: This is remedied by nftw implementation found on many systems, 
**   unfortunately I couldn't find one on Linux. Incidently this also gave me 
**   a chance to implement a better algorithm for the balancer. We call the 
**   function to move the files once gfs_max_files  have been accumulated in 
**   the flists and then restart the file tree parse from the previous point 
**   (such a restart would not be possible if the ftw were used as a black box)
**
** Input(s):
**   pathname  	- the directory to start from.
**   func	- function to be called for each file in the tree.
**
*******************************************************************************
******************************************************************************/

static int gfs_ftw(char *pathname, Func *func, gfs_fdat_t **flists, uint32 num_sps)
{
  strcpy((char *)fullpath, pathname);
  return(dopath(func, flists, num_sps));
}


static int dopath(Func * func, gfs_fdat_t **flists, uint32 num_sps)
{
  struct stat   statbuf;
  struct dirent *dirp;
  DIR 		*dp;
  int 		ret=0;
  char 		*ptr;


  if(lstat(fullpath, &statbuf) < 0)
    return(func(fullpath, &statbuf, flists, num_sps));

  if(S_ISDIR(statbuf.st_mode) == 0)
    return(func(fullpath, &statbuf, flists, num_sps));

  ptr = (char *)(fullpath + strlen(fullpath));

  *ptr++ = '/';
  *ptr = 0;

  if( (dp = opendir(fullpath)) == NULL)
    return(0);


  while ( (dirp = readdir(dp)) != NULL) {
    if(strcmp(dirp->d_name, ".") == 0 || 
       strcmp(dirp->d_name, "..") == 0)
      continue;
    
    strcpy(ptr, dirp->d_name);

    if( (ret = dopath(func, flists, num_sps)) != 0)
         break;

  }

  ptr[-1] = 0;

  if(closedir(dp) < 0)
    {
      printf("can't close dir %s\n", fullpath);
      exit(0);
    }

  return(ret);
}


static int gfs_findfiles(char * pathname, struct stat *sb, gfs_fdat_t **flists, uint32 num_sps)
{
  int 			fd,flag,retval, error;
  uint32                i, ctr;
  gfs_fspstatlist_t 	fspstatlist;
  gfs_spstatlist_t 	spstatlist;
  char			buff[256];
  gfs_spdat_t		*abovep=NULL, *belowp=NULL;



  if( !( S_ISREG(sb->st_mode) && (sb->st_nlink == 1) && 
	 (sb->st_size > gfs_min_filesize) ) )
    return(0); /* Continue */


  if ((fd = open(pathname, O_RDWR)) < 0)
    {
      sprintf(buff, "Can't open --  %s", pathname);
      if(Verbose)
	perror(buff); 
      return(0); /* Continue */
    }



  /* 
   * ioctl call - GFS_WHICH_SUBPOOL 
   */

  i = fspstatlist.fsp_cnt = num_sps;

  if((fspstatlist.blks = (uint64 *)malloc(sizeof(uint64) * i)) != NULL)
    memset((char *)fspstatlist.blks, 0, (sizeof(uint64) * i));
  else
    {
      fprintf(stderr, "%s: unable to allocate memory on line %d.\n", __FILE__, __LINE__); 
      close(fd);
      return(-1); /* Stop the file tree parsing */
    }

  if(ioctl(fd,GFS_WHICH_SUBPOOL,&fspstatlist) )
    {
      strcpy(buff, "gfs_tool: Error doing ioctl");
      perror(buff);
      close(fd);
      free(fspstatlist.blks);
      return(-1);  /* Stop the file tree parsing */
    }

  close(fd);

  if (fspstatlist.fsp_cnt > i)
    {
      free(fspstatlist.blks);
      return(-1);  /* Stop the file tree parsing */
    }

  for(flag = 0,ctr = 0, i=0; i < fspstatlist.fsp_cnt; i++)
    {
      if(fspstatlist.blks[i] != 0)
	flag+=1;
      if(flag > 0)
	ctr = i; 
    }

  /* 
   * Check if it is worth while to even consider the file i.e. 
   * Is it only in one subpool ?  Is that subpool above average ? 
   */
  if((flag != 1) || (sp_state[ctr] != GFS_ABOVE))  
    {
      free(fspstatlist.blks);
      return(0);  /* Continue */
    }


  retval = gfs_ins_flist(&flists[ctr], pathname, fspstatlist.blks[ctr]);
  free(fspstatlist.blks);

  if(retval < 0)
      return(1);  /* Some error occured - Stop parsing the file tree */

  if(retval == 0)
    return(0);


  
  /* 
   * retval == GFS_DONE => gfs_max_files accumulated .. Start processing 
   */

  error = 0;

  if ((fd = open(pathname, O_RDONLY)) < 0)
    {
      sprintf(buff, "Can't open %s", pathname);
      perror(buff);
      return(1);
    }

  if(gfs_show_subpools_i(&spstatlist,fd))
    {
      close(fd);
      return(1);
    }
  
  close(fd);

  if(gfs_find_sps(&spstatlist, &abovep, &belowp))
    {
      error = 1; /* break the recursive call */
      goto freeflist; /* We're Done */
    }
  
  if(Verbose) 
    {
      fprintf(stdout, "Subpools in the above list are \n");
      print_splist(abovep);
  
      fprintf(stdout, "Subpools in the below list are \n");
      print_splist(belowp);
    }

  if(! (gfs_files_to_mv(&spstatlist, flists, &abovep, &belowp) ))
    {
      /*      if(error != -EPERM) */
      error = 1;
      goto freeflist; /* No files further files to move */
    }

  if(!Verbose)
    fprintf(stdout, "*"); 



  /* 
   * Re-init for the next set of files 
   */
  if((init_sp_state(pathname)))
    error = 1; /* break the recursive call */


freeflist:
  freeflist(flists, spstatlist.sp_cnt);

  gfs_freesplist(&abovep);
  gfs_freesplist(&belowp);
  free(spstatlist.sp_list);

  return (error);
}



/******************************************************************************
*******************************************************************************
**
** void decodeOptions()
**
** Description:
**   This routine steps through the command-line arguments, parsing out
**   recognzied options.
**
** NOTE: This routine alters a number of global variables.
**
** Input(s):
**    argc - Number of arguments on the command line.
**  **argv - Pointer to an array of strings containing command-line arguments.
**
*******************************************************************************
******************************************************************************/

void decodeOptions(int argc, char *const argv[])
{
  int	errFlag = 0;
  int	optchar;
  int 	x;

  /* Initialize parameters with default values */
  gfs_load_margin = GFS_LOAD_MARGIN;
  gfs_min_filesize= GFS_MINFILESIZE;
  gfs_max_files   = GFS_MAX_FILES;
  Verbose         = FALSE;

  if (argc < 3)
    {
      for (x = 0; use[x][0]; x++)
	fputs(use[x], stderr);
      
      exit(EXIT_FAILURE);
    }
  
  while((optchar = getopt(argc, argv, OPTION_STRING)) != EOF) 
    {
      switch (optchar) 
	{
	case 'v':
	  Verbose = TRUE;
	  break;
	case 'l':
	  sscanf(optarg, "%f",&gfs_load_margin);
	  if(gfs_load_margin < 0.001)
	    {
	      fprintf(stderr, "Invalid Load Margin ( <=0.001 )\n");
	      exit(EXIT_FAILURE);
	    }
	  break;
	case 's':
	  sscanf(optarg, "%"FMT_64"d",&gfs_min_filesize);
	  if(gfs_min_filesize < 1)
	    {
	      fprintf(stderr, "Invalid Filesize\n");
	      exit(EXIT_FAILURE);
	    }
	  break;
	case 'f':
	  sscanf(optarg, "%"FMT_64"d",&gfs_max_files);
	  if(gfs_max_files < 1)
	    {
	      fprintf(stderr, "Invalid Filesize\n");
	      exit(EXIT_FAILURE);
	    }
	  break;
	case 'h':
	case '?':
	default:
	  errFlag++;
	}
    }

  if (errFlag) 
    {
      fprintf(stdout,"\n\n");
      for (x = 0; use[x][0]; x++)
	fputs(use[x], stderr);
      
      exit(EXIT_FAILURE);
    }

  fprintf(stdout,"Trying to balance the file system with the following parameters:\n");
  fprintf(stdout,"Load Margin   = %.3f \n", gfs_load_margin);
  fprintf(stdout,"Min File size = %"FMT_64"d \n", gfs_min_filesize);
  fprintf(stdout,"Max Files     = %"FMT_64"d \n", gfs_max_files);
  fprintf(stdout,"Verbose       = %d \n", Verbose);

  return;
}



/******************************************************************************
*******************************************************************************
**
** main.
**
*******************************************************************************
******************************************************************************/

int main(int argc,char *argv[])
{
  int 			fd, x;
  struct stat 		statBuf;
  char 			buff[GFS_MAXPATHLEN], exec[GFS_MAXPATHLEN], dname[GFS_MAXPATHLEN];
  gfs_spstatlist_t 	spstatlist;
  gfs_fdat_t 		**flists = NULL;
  gfs_spdat_t		*abovep=NULL, *belowp=NULL;
  uint32 		i;

  fprintf(stdout,"\n\n");
  for (x = 0; version[x][0]; x++)
    fputs(version[x], stdout);

  strcpy(exec, argv[0]);

  if(argc > 1)
    strcpy(dname, argv[1]);

  decodeOptions(argc, argv);


  if (lstat(dname, &statBuf) < 0)
    {
      perror("**lstat");
      exit(EXIT_FAILURE);
    }

  /* Search has to start from a directory */
  if(!(S_ISDIR(statBuf.st_mode)))
    {
      fprintf(stderr, "%s  : %s is not a directory", exec, dname);
      exit(EXIT_FAILURE);
    }


  /* 
   *  Print the initial state 
   */

  if ((fd = open(dname, O_RDONLY)) < 0)
    {
      sprintf(buff, "%s: Can't open %s", exec, dname);
      perror(buff);      
      exit(EXIT_FAILURE);
    }

  if(gfs_show_subpools_i(&spstatlist,fd))
    {
      close(fd);
      return 0;
    }

  if(spstatlist.sp_cnt == 1)
    {
      printf("Single Subpool found::No SP level balancing can be done !\n");
      free(spstatlist.sp_list);
      return 0;
    }

  /* Print the initial state of the subpools */
  fprintf(stdout, "\n\nCurrent Subpool Status\n");
  fprintf(stdout, "======================\n");
  gfs_print_graph(&spstatlist);
  close(fd);

  free(spstatlist.sp_list);



  /* 
   * Initialization. 
   */

  sp_state = malloc ( spstatlist.sp_cnt);
  if(sp_state == NULL)
    {
      fprintf(stderr, "%s :Out of Memory on Line %d\n", __FILE__, __LINE__);
      return 0;
    }

  /* Initialize sp_state table */
  if((init_sp_state(dname)))
    {
      fprintf(stdout,"No Subpools to Balance\n");
      free(sp_state);
      return 0;
    }

  flists = (gfs_fdat_t **) malloc( sizeof(gfs_fdat_t *) * spstatlist.sp_cnt );
  if(flists == NULL)
    {
      fprintf(stderr, "%s :Out of Memory on Line %d\n", __FILE__, __LINE__);
      free(sp_state);
      return 0;
    }

  for(i = 0; i < spstatlist.sp_cnt; i++)
      flists[i] = NULL;




  /* 
   * Try to balance 
   */

  gfs_ftw(dname, gfs_findfiles, flists, spstatlist.sp_cnt);

  /* there may be files left in the flists ? */
  if ((fd = open(dname, O_RDONLY)) < 0)
    {
      sprintf(buff, "Can't open %s", dname);
      perror(buff);
      return 0;
    }

  if(gfs_show_subpools_i(&spstatlist,fd))
    {
      close(fd);
      return 0;
    }
  
  close(fd);
  
  if(gfs_find_sps(&spstatlist, &abovep, &belowp))
    {
      fprintf(stdout,"\nNo more Subpools to Balance\n");
      goto freeflist;
    }
  
  if(Verbose)
    {
      fprintf(stdout, "Subpools in the above list are \n");
      print_splist(abovep);
      
      fprintf(stdout, "Subpools in the below list are \n");
      print_splist(belowp);
    }

  /* Re-init the sp_state list variable */
  if((init_sp_state(dname)))
    {
      fprintf(stdout,"\nNo more Subpools to Balance\n");
      goto freeflist;
    }

  if(! (gfs_files_to_mv(&spstatlist, flists, &abovep, &belowp) ))
    {
      fprintf(stdout,"\nCannot find files to move\n");
      goto freeflist; /* No files further files to move */
    }



  /*
   * Cleanup
   */

 freeflist:
  freeflist(flists, spstatlist.sp_cnt);

  gfs_freesplist(&abovep);
  gfs_freesplist(&belowp);
  free(spstatlist.sp_list);


  /* 
   * Print out final statistics.
   */
  
  if ((fd = open(dname, O_RDONLY)) < 0)
    {
      sprintf(buff, "%s: Can't open %s", exec, dname);
      perror(buff);
      
      exit(EXIT_FAILURE);
    }

  if(!gfs_show_subpools_i(&spstatlist,fd))
    {
      fprintf(stdout, "\n\nNew Subpool Status\n");
      fprintf(stdout, "==================\n");
      gfs_print_graph(&spstatlist);
      free(spstatlist.sp_list);
    }

  close(fd);

  free(sp_state);

  return 0;
}

