/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Andrew Barry <barry@lcse.umn.edu>
** 
**    Module name: partition.c
**
**    Description: data structures and state change functions for those
**    data structures. Specifically the lock structure and the partition
**    structure. 
**
*******************************************************************************
******************************************************************************/


#ifndef __PARTITION_C__
#define __PARTITION_C__

#include "partition.h"


server_dlock *new_server_dlock(){
  server_dlock *sdl;
  sdl=(server_dlock *)malloc(sizeof(server_dlock));
  pthread_mutex_init(&(sdl->pmut), NULL);
  sdl->version_number=0;
  sdl->state=STATE_UNLOCKED;
  sdl->act=ACTIVITY_OFF;
  sdl->num_holders=0;
  sdl->expired=EXPIRED_NOT;
  sdl->exclusive_pending=EXCLUSIVE_NOT_PENDING;
  sdl->expire_time=0;
  sdl->head=NULL;
  sdl->qsize=0;
  sdl->qhead=NULL;
  return(sdl);
}

void delete_server_dlock(server_dlock *sdl){
  sdl_clear_cid(sdl);
  free(sdl);
}

cid_entry *get_cid_list(int i, server_dlock *sdl){
  cid_entry *cid;
  int j;

  if(i>sdl->num_holders){
    return(NULL);}
  else{
    cid=sdl->head;
    for(j=0;j< i; j++){
      cid=cid->next;}
    return(cid);
  }
}

cid_entry *get_q_list(int i, server_dlock *sdl){
  cid_entry *cid;
  int j;

  if(i>sdl->qsize){
    return(NULL);}
  else{
    cid=sdl->qhead;
    for(j=0;j< i; j++){
      cid=cid->next;}
    return(cid);
  }
}


int get_cid_index(uint32 lookup, server_dlock *sdl){
  cid_entry *cid;
  int i;

  cid=sdl->head;
  if(cid==NULL){
    return(-1);}
  else{
    for(i=0;i<sdl->num_holders; i++){
      if(cid==NULL){
	return(-1);}
      if(cid->cid==lookup){
	return i;
      }
      cid=cid->next;
    }
    return(-1);
  }
}

int get_q_index(uint32 lookup, server_dlock *sdl){
  cid_entry *cid;
  int i;

  cid=sdl->qhead;
  if(cid==NULL){
    return(-1);}
  else{
    for(i=0;i<sdl->qsize; i++){
      if(cid==NULL){
	return(-1);}
      if(cid->cid==lookup){
	return i;
      }
      cid=cid->next;
    }
    return(-1);
  }
}


void sdl_reset       (server_dlock *sdl){
  time_t *tttp, ttt;
  tttp=NULL;
  ttt=time(tttp);
  sdl->expire_time=(EXPIRE_TIME + (int)ttt);
}

void sdl_clear_cid   (server_dlock *sdl){
  int i;
  cid_entry *cidn;
  cidn=sdl->head;
  for(i=0; i<sdl->num_holders; i++){
    cidn=sdl->head;
    if(cidn==NULL)
      return;
    sdl->head=sdl->head->next;
    free(cidn);
  }
  sdl->num_holders=0;
  return;
}

void sdl_clear_q   (server_dlock *sdl){
  int i;
  cid_entry *cidn;
  cidn=sdl->qhead;
  for(i=0; i<sdl->qsize; i++){
    cidn=sdl->qhead;
    if(cidn==NULL)
      return;
    sdl->qhead=sdl->qhead->next;
    free(cidn);
  }
  sdl->qsize=0;
  return;
}

/**********************************************************
 *returns true if a particular cid is a holder of the lock
 **********************************************************/
int cid_in_dlock(server_dlock *sdl, uint32 lookup){
  int i;
  
  i=get_cid_index(lookup, sdl);
  if(i==-1){
    return 0;
  }
  else
    return 1;
}

int cid_in_q(server_dlock *sdl, uint32 lookup){
  int i;
  
  i=get_q_index(lookup, sdl);
  if(i==-1){
    return 0;
  }
  else
    return 1;
}

/**********************************************************
 *removes a holder from the holder array, decrements the length
 **********************************************************/
int rem_dlock_cid(server_dlock *sdl, uint32 rem){
  cid_entry *cid1, *cid2;
  int i=1;

  if(sdl->head==NULL){
    return(-1);
  }   /*do nothing if list is empty*/
  else if(sdl->head->cid==rem){/*remove first element*/
    cid1=sdl->head;
    sdl->head=sdl->head->next;
    free(cid1);
    sdl->num_holders--;
    return 1;
 }  /*free the memory*/
  else{
    cid1=sdl->head->next;
    cid2=sdl->head;
    while(cid1!=NULL){
      if(cid1->cid==rem){ 
	cid2->next=cid1->next;
	free(cid1);
	sdl->num_holders--;
	return 1;}
      cid2=cid1;    /*move through the list cid2 follows cid1*/
      cid1=cid1->next;
      i++;
    }
    return(-1);
  }
}

int rem_q_cid(server_dlock *sdl, uint32 rem){
  cid_entry *cid1, *cid2;
  int i=1;
  
  if(sdl->qhead==NULL){
    return(-1);
  }   /*do nothing if list is empty*/
  else if(sdl->qhead->cid==rem){/*remove first element*/
    cid1=sdl->qhead;
    sdl->qhead=sdl->qhead->next;
    free(cid1);
    sdl->qsize--;
    /*dbprintf("able to free q");*/
    return 1;
 }  /*free the memory*/
  else{
    cid1=sdl->qhead->next;
    cid2=sdl->qhead;
    while(cid1!=NULL){
      if(cid1->cid==rem){ 
	cid2->next=cid1->next;
	free(cid1);
	sdl->qsize--;
	/*dbprintf("able to free q");*/
	return 1;}
      cid2=cid1;    /*move through the list cid2 follows cid1*/
      cid1=cid1->next;
      i++;
    }
    /*dbprintf("unable to free q!!!!!!");*/
    return(-1);
  }
}


/**********************************************************
 *adds a holder into the array of holders. increments length
 **********************************************************/
int add_dlock_cid(server_dlock *sdl, uint32 add, send_req_t *sr){
  cid_entry *cide;

  cide=new_cid_entry_();
  cide->cid=add;
  cide->next=sdl->head;
  cide->soc=sr;
  assert(cide->soc != NULL );/**/
  assert( sr != NULL );/**/
  sdl->head=cide;
  sdl->num_holders++;
  return(1);
}

int add_q_cid(server_dlock *sdl, uint32 add, uint8 state, send_req_t *sr){
  cid_entry *cide;

  if(!(cid_in_q(sdl, add))){
    cide=new_cid_entry_();
    cide->cid=add;  
    cide->next=sdl->qhead;
    cide->soc=sr;
    sdl->qhead=cide;
    cide->state=state;
    sdl->qsize++;
    return(1);
  }
  else return (0);
}

void dnrv(drop_rdf *drp, uint32 ip){}


partition_locks *new_partition_locks(int numlocks){
  partition_locks *pal;
  int                     i;

  pal=(partition_locks *)malloc(sizeof(partition_locks));
  pal->dlocks_per_partition=numlocks;
  pal->lock_timeout_interval=EXPIRE_TIME;
  pal->max_lock_holds=MAX_LOCK_HOLDS;


  pal->sdl=(server_dlock **)malloc(numlocks*sizeof(server_dlock *));  
  for(i=0; i<numlocks; i++){
    pal->sdl[i]=new_server_dlock();
    pal->sdl[i]->lock_number=(uint32)i;
  }
  return(pal);
}

void delete_partition_locks(partition_locks *pal){
  int i;
  
  for(i=0; i<pal->dlocks_per_partition; i++){
    delete_server_dlock(pal->sdl[i]);}
  free(pal->sdl);
  free(pal);
}


/**************************************************************
 ******************  cid_entry functions   ********************
 **************************************************************/


cid_entry *new_cid_entry_(){
  cid_entry *cidn;
  /*changed for debug*/
  cidn=(cid_entry *)malloc(sizeof(cid_entry));
  cidn->next=NULL;
  cidn->state=STATE_UNLOCKED;
  return(cidn);}

cid_entry *new_cid_entry(uint32 entry){
  cid_entry *cidn;
  /*changed for debug*/
  cidn=(cid_entry *)malloc(sizeof(cid_entry));
  cidn->next=NULL;
  cidn->state=STATE_UNLOCKED;
  cidn->cid=entry;
  return(cidn);}
 





#endif /* __PARTITION_C__ */
