/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Andrew Barry <barry@lcse.umn.edu>
** 
**    Module name: partition.h
**
*******************************************************************************
******************************************************************************/


#ifndef __PARTITION_H__
#define __PARTITION_H__

#include "global.h"
#include "mode_page.h"
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>

struct server_dlock; typedef struct server_dlock server_dlock;
struct partition_locks; typedef struct partition_locks partition_locks; 
struct cid_ntry; typedef struct cid_ntry cid_ntry; typedef cid_ntry cid_entry;

partition_locks *new_partition_locks(int numlocks);
server_dlock    *new_server_dlock();
cid_entry       *new_cid_entry_();
cid_entry       *new_cid_entry(uint32 entry);

/*********************************************************
 *********************************************************
 **get_cid_list()
 **
 **Description: return the ith cid_entry in the server
 **dlock. 
 **Returns: cid_entry pointer. 
 *********************************************************
 *********************************************************/
cid_entry *get_cid_list(int i, server_dlock *sdl);
cid_entry *get_q_list  (int i, server_dlock *sdl);

/*********************************************************
 *********************************************************
 **get_cid_index()
 **
 **Description: returns the index of the input cid value
 **within the input server dlock. 
 *********************************************************
 *********************************************************/
int  get_cid_index(uint32 lookup, server_dlock *sdl);
int  get_q_index  (uint32 lookup, server_dlock *sdl);

/*********************************************************
 *********************************************************
 **sdl_reset()
 **
 **Description: updates the lock's expire time. 
 *********************************************************
 *********************************************************/
void sdl_reset       (server_dlock *sdl);

/*********************************************************
 *********************************************************
 **sdl_clear_cid()
 **
 **Description: removes all client id's from a dlock.
 *********************************************************
 *********************************************************/
void sdl_clear_cid   (server_dlock *sdl);
void sdl_clear_q     (server_dlock *sdl);

/*********************************************************
 *********************************************************
 **add_dlock_cid()
 **
 **Description: adds a client id to a dlock rdf.
 **Returns:1 always
 *********************************************************
 *********************************************************/
int  add_dlock_cid   (server_dlock *sdl, uint32 add, send_req_t *sr);
int  add_q_cid       (server_dlock *sdl, uint32 add, uint8 state, send_req_t *sr);


/*********************************************************
 *********************************************************
 **rem_dlock_cid()
 **
 **Description: removes the input client id from a dlock.
 **
 **Input: uint32 of client id value.
 **Returns: -1 on error, 1 on success.
 *********************************************************
 *********************************************************/
int  rem_dlock_cid   (server_dlock *sdl, uint32 rem);
int  rem_q_cid       (server_dlock *sdl, uint32 rem);


/*********************************************************
 *********************************************************
 **cid_in_dlock()
 **
 **Description: returns true if the cid is in the lock, 
 ** false if not.
 **
 **Input: uint32 of cid value.
 *********************************************************
 *********************************************************/
int  cid_in_dlock    (server_dlock *sdl, uint32 lookup);
int  cid_in_q        (server_dlock *sdl, uint32 lookup);


/*********************************************************
 *********************************************************
 **delete_partition_locks()
 **
 **Description: like a deconstructor function.
 *********************************************************
 *********************************************************/
void delete_partition_locks(partition_locks *pal);

/*********************************************************
 *********************************************************
 **delete_server_dlock()
 **
 **Description: deconstructor function.
 *********************************************************
 *********************************************************/
void delete_server_dlock(server_dlock *sdl);


/**********************************************************
 **********************************************************
 **********************************************************
 ***DATA TYPE 
 *** struct cid_ntry
 ***
 *** data structure used to hold a single client ID
 ***
 **********************************************************
 **********************************************************/
struct cid_ntry {
  struct cid_ntry *next;
  u8     state; /*for queue*/
  u32    cid;
  send_req_t *soc;
};



/***********************************************************************
 ***********************************************************************
 *struct server_dlock  DATA TYPE   
 *
 *each server dlock keeps track of all state information for a single dlock. 
 *client ids are stored in a linked list. Since the average number of
 *clients holding each lock is much less than one, this implementation offers
 *no performance penalty.
 *
 *qsize and qhead are part of the server callbacks. Currently the queue
 *is only one deap, and is not enforced. It simply informs the machine 
 *in the queue when the lock is dropped. This is likely to be radically 
 *changed soon. 
 *
 ***********************************************************************
 ***********************************************************************/
struct server_dlock{
  
  pthread_mutex_t        pmut;
  uint32                 lock_number;
  uint32                 version_number;
  uint8                  state;
  uint8                  act;
  uint8                  num_holders;
  uint8                  expired;
  uint8                  exclusive_pending;
  double                 expire_time;
  cid_entry              *head;
  uint8                  qsize;
  cid_entry              *qhead;
};


/**********************************************************
 **********************************************************
 **********************************************************
 ***DATA TYPE 
 *** struct partition_locks
 ***
 *** data structure used to hold all relevant information about
 *** a server's dlocks. A partition_locks data
 *** structure can hold any number of locks.
 ***
 **********************************************************
 **********************************************************/
struct partition_locks{
  struct server_dlock   **sdl;
  uint32                dlocks_per_partition;
  uint32                lock_timeout_interval;
  uint8                 max_lock_holds;

};


/******
 *does nothing, returns void
 ******/
void dnrv(drop_rdf *drp, uint32 ip);






#endif /* __PARTITION_C__ */
