/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: gfs_tool.c
**
**    Description:
**      Tool for changing debug flags in GFS, finding the subpool status,
** 	finding the file status and moving a file to a given destination
** 	subpool.
**
*******************************************************************************
******************************************************************************/

/******************************************************************************
*******************************************************************************
**
** $Header: /home/brule/research/CVS/GFS2/src/tools/gfs_tool/gfs_tool.c,v 1.5 1999/05/19 17:53:51 agarwal Exp $
**
** $Log: gfs_tool.c,v $
** Revision 1.5  1999/05/19 17:53:51  agarwal
** General Utility function used by gfs_tool and gfs_balancer.
**
** Revision 1.4  1999/04/13 20:09:44  kpreslan
** gfs_tool doesn't need to know which version of the locking code we are using.
**
 * Revision 1.3  1999/01/22  00:59:38  kpreslan
 * Copyright changes (It's 1999, whee!!)
 *
 * Revision 1.2  1998/09/22  03:06:45  kpreslan
 * Snapshot
 *
 * Revision 1.1  1998/09/11  21:49:36  teigland
 * Initial import.  GFS parts commented out.
 *
** Revision 1.8  1998/07/01  19:39:39  kpreslan
** Fixed the weasel words to reflect the fact that GFS is under the GPL.
**
** Revision 1.7  1998/06/26  09:35:40  kpreslan
** Fixed Copyright weasel words.
**
** Revision 1.6  1998/02/06  21:47:11  kpreslan
** gfs_tool now works even when GFS_DEBUG_ON isn't the same as POOL_DEBUG_ON!!!
**
** Revision 1.5  1997/07/11  19:42:07  grant
** Added the ability to control the debugging flags for both GFS and the
** pool driver separately.
**
** Revision 1.4  1997/07/02  13:22:22  grant
** Added a RCS log header.
**
*******************************************************************************
******************************************************************************/

#ifdef IRIX
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/attributes.h>
#endif

#ifdef linux
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <sys/stat.h>
#include <sys/dir.h>
#endif

#include <ctype.h>
#include "global.h"

#define WANT_DFLAG_T
#include "pool_debug.h"

#include "pool.h"

#include "gfs_debug.h"
#include "gfs_ioctl.h"


/* Preprocessor Defines */


/* Function Macros */


/* Type Definitions */


/* Global Variables */

const char *usage[]=
{
  "Add debug flags:\n",
  "  gfs_tool on <Pool_Device or GFS_Directory> <debug_flags>\n",
  "\n",
  "Remove debug flags:\n",
  "  gfs_tool off <Pool_Device or GFS_Directory> <debug_flags>\n",
  "\n",
  "Report subpool status:\n",
  "  gfs_tool show_subpools <GFS File or Directory>\n",
  "\n",
  "Report file block statistics:\n",
  "  gfs_tool which_subpool <GFS File or Directory>\n",
  "\n",
  "Move a file to a given Resource Group:\n",
  "  gfs_tool move_to_subpool <Parameter File>\n",
  "\n",
#ifdef IRIX
  "Change mount resource group:\n",
  "  gfs_tool mrg <root_directory> [number]\n",
  "\n",
  "Change a file's resource group:\n",
  "  gfs_tool frg <filename> [number]\n",
#endif
  ""
};



/* Function Prototypes */

void set_debug(int argc, char **argv);
#ifdef IRIX
void set_mrg(int argc, char **argv);
void set_frg(int argc, char **argv);
#endif
void gfs_stat(int argc, char **argv);
void gfs_fstat(int argc,char **argv);
void gfs_mv_to_sp(int argc, char **argv);

extern void fill_params(gfs_lnkname_t * lnk, char * istr);
extern int gfs_show_subpools_i(gfs_spstatlist_t *spstatlistp, int fd);
extern void gfs_mv_file(gfs_lnkinfo_t *filep, int verbose);




int main(int argc,char *argv[])
{
  int x;
  
  if (argc < 2)
    {
      for (x = 0; usage[x][0]; x++) {
	fputs(usage[x],stderr);
      }

      exit(EXIT_FAILURE);
    }

  if (strcmp(argv[1], "on") == 0 || strcmp(argv[1], "off") == 0)
    {
      set_debug(argc, argv);		/*  Change GFS debug level  */
    }
#ifdef IRIX
  else if (strcmp(argv[1], "mrg") == 0)
    {
      set_mrg(argc, argv);
    }
  else if (strcmp(argv[1], "frg") == 0)
    {
      set_frg(argc, argv);
    }
#endif
  else if(strcmp(argv[1], "show_subpools") == 0)
    {
      gfs_stat(argc,argv);
    }
  else if(strcmp(argv[1], "which_subpool") == 0)
    {
      gfs_fstat(argc,argv);
    }
  else if(strcmp(argv[1], "move_to_subpool") == 0)
    {
      gfs_mv_to_sp(argc,argv);
    }
  else
    {
      fprintf(stderr, "%s: Bad option - %s\n", argv[0], argv[1]);

      for (x = 0; usage[x][0]; x++) {
	fputs(usage[x], stderr);
      }
      
      exit(EXIT_FAILURE);
    }
  return 0;
}


/******************************************************************************
*******************************************************************************
**
** void set_debug()
**
** Description:
**   This routine uses an ioctl command to change the GFS or pool driver
**   debugging level.
**
** Input(s):
**    argc - 
**  **argv - 
**
*******************************************************************************
******************************************************************************/

void set_debug(int argc, char **argv)
{
  char buff[256];
  struct stat statBuf;
  int x, fd, arg;
  char done;
  uint64 flags = 0;
  int pool;



  if (argc < 4)
    {
      for (x = 0; usage[x][0]; x++)
	fputs(usage[x], stderr);
      
      exit(EXIT_FAILURE);
    }



  if (lstat(argv[2], &statBuf) < 0)
    {
      perror("lstat");
      exit(EXIT_FAILURE);
    }

  pool=(S_ISCHR(statBuf.st_mode) || S_ISBLK(statBuf.st_mode)) ? TRUE : FALSE;

  if ((fd = open(argv[2], O_RDONLY)) < 0)
    {
      sprintf(buff, "%s: Can't open %s", argv[0], argv[2]);
      perror(buff);
      
      exit(EXIT_FAILURE);
    }



  if (pool)
  {
    for (arg = 3; arg < argc; arg++)
    {
      done = FALSE;
      
      /* Check the pool driver debugging flags */
      
      for (x = 0; POOL_DFLAGS[x].name[0]; x++)
      {
	if (strcmp(POOL_DFLAGS[x].name, argv[arg]) == 0)
	{
	  flags |= POOL_DFLAGS[x].flag;
	  
	  done = TRUE;
	  break;
	}
      }
      
      if (!done)
      {
	fprintf(stderr, "%s:  Bad flag - %s\nValid flags are:\n", 
		argv[0], argv[arg]);
	
	fprintf(stderr, "  Pool debugging flags:\n");
	
	for (x = 0; POOL_DFLAGS[x].name[0]; x++)
	  fprintf(stderr, "    %s\n", POOL_DFLAGS[x].name);
	
	exit(EXIT_FAILURE);
      }
    }
    
#ifdef DEBUG
  printf("Flags: %.16"FMT_64"X\n", flags);
#endif
  
    if (ioctl(fd, ((strcmp(argv[1], "on") == 0) ? POOL_DEBUG_ON : POOL_DEBUG_OFF), &flags) != 0)
    {
      strcpy(buff, "gfs_tool: Error doing ioctl");
      perror(buff);

      exit(EXIT_FAILURE);
    }
  }
  else
  {
    for (arg = 3; arg < argc; arg++)
    {
      done = FALSE;
      
      /* Check the GFS debugging flags */
      
      for (x = 0; GFS_DFLAGS[x].name[0]; x++)
      {
	if (strcmp(GFS_DFLAGS[x].name, argv[arg]) == 0)
	{
	  flags |= GFS_DFLAGS[x].flag;
	  
	  done = TRUE;
	  break;
	}
      }
      
      if (!done)
      {
	fprintf(stderr, "%s:  Bad flag - %s\nValid flags are:\n", 
		argv[0], argv[arg]);
	
	fprintf(stderr, "  GFS debugging flags:\n");
	
	for (x = 0; GFS_DFLAGS[x].name[0]; x++)
	  fprintf(stderr, "    %s\n", GFS_DFLAGS[x].name);
	
	exit(EXIT_FAILURE);
      }
    }
    
#ifdef DEBUG
  printf("Flags: %.16"FMT_64"X\n", flags);
#endif

    
    if (ioctl(fd, ((strcmp(argv[1], "on") == 0) ? GFS_DEBUG_ON : GFS_DEBUG_OFF), &flags) != 0)
    {
      strcpy(buff, "gfs_tool: Error doing ioctl");
      perror(buff);
      
      exit(EXIT_FAILURE);
    }
  }



  close(fd);
}




/******************************************************************************
*******************************************************************************
**
** gfs_lnkinfo_t * GetParams()
**
** Description:
**   This routine read the input parameter file and returns the corresponding
**   gfs_lnkinfo_t structure.
**
** Input(s):
**   ParamsFile - file containing the parameters.
**
** Returns:
**   The corresponding gfs_linkinfo_t structure pointer.
**
*******************************************************************************
******************************************************************************/

gfs_lnkinfo_t * GetParams(const char *ParamsFile)
{
  gfs_lnkinfo_t	*lnkinfo;
  char	      	linebuf[GFS_MAXPATHLEN], *ctmp;
  char		arg1[GFS_MAXPATHLEN];
  int 		arg2;
  int 		cnt=0;
  FILE		*fp = NULL;
  struct stat 	statBuf;


  if ((fp = fopen(ParamsFile, "r")) == NULL) 
    {
      fprintf(stderr, "Error opening the parameter file \"%s\"\n", ParamsFile);
      perror(":");
      exit(1);
    }


  lnkinfo = (gfs_lnkinfo_t *)malloc(sizeof(gfs_lnkinfo_t));
  lnkinfo->l_name = NULL;


  while(fgets(linebuf, GFS_MAXPATHLEN, fp) != NULL)
    {
      ctmp = linebuf;

      while((*ctmp != '\n') && (isspace(*ctmp))) ctmp++;

      if(*ctmp == '#')
	continue;

      if(*ctmp == '\n')
	continue;
  
      sscanf(linebuf, "%s %d", arg1, &arg2);

      if(!strcmp(arg1,"drg")) 
	{
	  lnkinfo->d_sp = arg2;
	} 
      else if(!strcmp(arg1,"lnk_count")) 
	{
	  lnkinfo->lnk_cnt = arg2;
	  lnkinfo->l_name = (gfs_lnkname_t *) malloc(sizeof(gfs_lnkname_t)*arg2);
	}
      else 
	{
	  if(lnkinfo->l_name == NULL) 
	    {
	      fprintf(stderr, "Error:: lnk_count option should come before the link path \n");
	      fclose(fp);
	      free(lnkinfo);
	      exit(1);
	    }
	  
	  if(cnt > lnkinfo->lnk_cnt) 
	    {
	      fprintf(stderr, "Error:: lnk_count is less than the number of links \n");
	      free(lnkinfo->l_name);
	      free(lnkinfo);
	      fclose(fp);
	      exit(1);
	    }

	  if (lstat(arg1, &statBuf) < 0)
	    {
	      perror("lstat");
	      free(lnkinfo->l_name);
	      free(lnkinfo);
	      fclose(fp);
	      exit(EXIT_FAILURE);
	    }

	  if(S_ISDIR(statBuf.st_mode))
	    {
	      fprintf(stderr, "%s:: Cannot move a directory\n", arg1);
	      free(lnkinfo->l_name);
	      free(lnkinfo);
	      fclose(fp);
	      exit(EXIT_FAILURE);
	    }

	  if(statBuf.st_nlink != lnkinfo->lnk_cnt)
	    {
	      fprintf(stderr, "Error:: Invalid Link %s \n",arg1);
	      free(lnkinfo->l_name);
	      free(lnkinfo);
	      fclose(fp);
	      exit(1);
	    }

	  fill_params(&lnkinfo->l_name[cnt], arg1);
	  cnt++;	
	}
    }
  

  if(cnt != lnkinfo->lnk_cnt)
    {
      fprintf(stderr, "Error:: Number of links do not match  \n");
      free(lnkinfo->l_name);
      free(lnkinfo);
      fclose(fp);
      exit(1);
    }

  fclose(fp);
  return(lnkinfo);
}


/******************************************************************************
*******************************************************************************
**
** void gfs_stat()
**
** Description:
**   This routine uses an ioctl command to gather stats about 
**   the subpools, and prints them out.
**
** Input(s):
**    argc - argc from main()
**  **argv - argv from main()
**
*******************************************************************************
******************************************************************************/

void gfs_stat(int argc, char **argv)
{
  int 			x,fd;
  struct stat 		statBuf;
  char 			buff[256];
  gfs_spstatlist_t 	spstatlist;
  uint32 		j;

  if (argc < 3)
    {
      for (x = 0; usage[x][0]; x++)
	fputs(usage[x], stderr);
      
      exit(EXIT_FAILURE);
    }

  if (lstat(argv[2], &statBuf) < 0)
    {
      perror("lstat");
      exit(EXIT_FAILURE);
    }

  if ((fd = open(argv[2], O_RDONLY)) < 0)
    {
      sprintf(buff, "%s: Can't open %s", argv[0], argv[2]);
      perror(buff);
      
      exit(EXIT_FAILURE);
    }

  if(!gfs_show_subpools_i(&spstatlist,fd))
  {
    /* print the results onto the screen */
    printf("          \t\tFree Blocks     Total Blocks    Dinodes\n");
    printf("---------------------------------------------------------------\n");
    for(j=0; j < spstatlist.sp_cnt ; j++)
      {
	printf("Subpool %"FMT_32"d\t\t%8"FMT_64"d         %8"FMT_64"d  %8"FMT_64"d\n", j, spstatlist.sp_list[j].bfree, spstatlist.sp_list[j].blocks,
	       spstatlist.sp_list[j].dinds);
      }

    free(spstatlist.sp_list);
  }

  close(fd);

}



/******************************************************************************
*******************************************************************************
**
** void gfs_fstat()
**
** Description:
**   This routine uses an ioctl command to gather subpool information about 
**   the given file, and prints it out.
**
** Input(s):
**    argc - argc from main()
**  **argv - argv from main()
**
*******************************************************************************
******************************************************************************/

void gfs_fstat(int argc, char **argv)
{
  int 			x,fd;
  struct stat 		statBuf;
  char 			buff[256];
  gfs_fspstatlist_t 	fspstatlist;
  uint32 		i,j;

  if (argc < 3)
    {
      for (x = 0; usage[x][0]; x++)
	fputs(usage[x], stderr);
      
      exit(EXIT_FAILURE);
    }

  if (lstat(argv[2], &statBuf) < 0)
    {
      perror("lstat");
      exit(EXIT_FAILURE);
    }

  if ((fd = open(argv[2], O_RDONLY)) < 0)
    {
      sprintf(buff, "%s: Can't open %s", argv[0], argv[2]);
      perror(buff);
      
      exit(EXIT_FAILURE);
    }

  /* first try with 128 subpool stat structures */
  fspstatlist.fsp_cnt = i = 16;

  do
    {
      i = fspstatlist.fsp_cnt;

      if((fspstatlist.blks = (uint64 *)malloc(sizeof(uint64) * i)) != NULL)
	memset((char *)fspstatlist.blks, 0, (sizeof(uint64) * i));
      else
	{
	  fprintf(stderr, "%s: unable to allocate memory on line %d.\n", argv[0], __LINE__); 
	  close(fd);
	  exit(EXIT_FAILURE); 
	}

      if (ioctl(fd,GFS_WHICH_SUBPOOL,&fspstatlist ))
	{
	  strcpy(buff, "*gfs_tool: Error doing ioctl");
	  perror(buff);
	  close(fd);
	  free(fspstatlist.blks);
	  exit(EXIT_FAILURE);
	}

      /* print the results onto the screen */
      if (fspstatlist.fsp_cnt <= i)
	{
	  printf("          \t\tTotal Blocks\n");
	  printf("---------------------------------------\n");

	  for(j=0; j < fspstatlist.fsp_cnt ; j++)
	      printf("Subpool %"FMT_32"d\t\t%8"FMT_64"d\n",j,fspstatlist.blks[j]);
	  free(fspstatlist.blks);
	  break;
	}

      free(fspstatlist.blks);
    }  while (i > fspstatlist.fsp_cnt); /* try until it succeeds */
  
  close(fd);
}



/******************************************************************************
*******************************************************************************
**
** void gfs_mv_to_sp()
**
** Description:
**   This routine uses an ioctl command to move a given file from its 
**   current location to the given resource group.
**
** Input(s):
**    argc - argc from main()
**  **argv - argv from main()
**
** Output:
** 	None.
**
*******************************************************************************
******************************************************************************/

void gfs_mv_to_sp(int argc, char **argv)
{
  int 			x;
  gfs_lnkinfo_t		*lnkinfo;


  if (argc < 3)
    {
      for (x = 0; usage[x][0]; x++)
	fputs(usage[x], stderr);
      
      exit(EXIT_FAILURE);
    }

  lnkinfo = GetParams(argv[2]);

  gfs_mv_file(lnkinfo, 1);
 
  free(lnkinfo->l_name);
  free(lnkinfo);
  return;
}




#ifdef IRIX
/********************************************************
***  Change mount resource group.
********************************************************/
void set_mrg(int argc, char **argv)
{
  char buff[256];
  char data[65536];
  int x, len=65536;



  if (argc==3)
  {
    if (attr_get(argv[2], GFS_ATTR_MRGRP, data, &len, 0)!=0)
    {
      sprintf(buff, "gfs_tool:  error getting mount group number");
      perror(buff);

      exit(EXIT_FAILURE);
    }

    printf("%d\n", atoi(data));
  }
  else if (argc==4)
  {
    if (attr_set(argv[2], GFS_ATTR_MRGRP, argv[3], strlen(argv[3]), 0)!=0)
    {
      sprintf(buff, "gfs_tool:  error changing mount group number");
      perror(buff);

      exit(EXIT_FAILURE);
    }
  }
  else
  {
    for (x=0; usage[x][0]; x++)
      fputs(usage[x], stderr);
    
    exit(EXIT_FAILURE);
  }
}


/********************************************************
***  Change file resource group.
********************************************************/
void set_frg(int argc, char **argv)
{
  char buff[256];
  char data[65536];
  int x, len=65536;



  if (argc==3)
  {
    if (attr_get(argv[2], GFS_ATTR_FRGRP, data, &len, 0)!=0)
    {
      sprintf(buff, "gfs_tool:  error getting file group number");
      perror(buff);

      exit(EXIT_FAILURE);
    }

    printf("%d\n", atoi(data));
  }
  else if (argc==4)
  {
    if (attr_set(argv[2], GFS_ATTR_FRGRP, argv[3], strlen(argv[3]), 0)!=0)
    {
      sprintf(buff, "gfs_tool:  error changing file group number");
      perror(buff);

      exit(EXIT_FAILURE);
    }
  }
  else
  {
    for (x=0; usage[x][0]; x++)
      fputs(usage[x], stderr);
    
    exit(EXIT_FAILURE);
  }
}
#endif










