/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: repair.c
**
**    Description:
**	Filesystem repair.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include "global.h"
#include "gfsck.h"
#include "gfs_ondisk.h"

extern char *prog_name;
extern node_t nil;
extern int allyes;


char di_errs[][24] = { "magic number", "resource group number", "file type",
		       "nopen count", "block count", "depth", "same depth",
		       "no dot entry", "missing . or ..", "entry count",
		       "link count", "not in any dir", "not connected to root",
		       "dinode number", "not defined", "unknown"};

char de_errs[][24] = { "hash value", "record length", "dinode not found" };

int fixnopen = 0, fixbm = 0;


/******************************************************************************
*******************************************************************************
** NAME
**	repair_all
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

void repair_all(filesystem_t *fs) 
{
  node_t *x;
  gfs_sb_t sb;
  ent_info_t *ent;
  di_info_t *info;
  char buf[sizeof(gfs_sb_t)];


  /* fix mount count in sb */

  do_lseek(fs->fd, GFS_SB_ADDR * GFS_BASIC_BLOCK);  
  do_read(fs->fd, buf, sizeof(gfs_sb_t));
  gfs_sb_in(&sb, buf);
  
  if (sb.sb_mounts && query_yn("Set mount count in superblock to zero? [yN]: ")) {
    sb.sb_mounts = 0;
    
    gfs_sb_out(&sb, buf);
    do_lseek(fs->fd, GFS_SB_ADDR * GFS_BASIC_BLOCK);  
    do_write(fs->fd, buf, sizeof(gfs_sb_t));
  }


  /* fix bad_ents */

  x = Tree_Minimum(fs->bad_ent);

  if (x != &nil)
    fprintf(stderr, "\n** Repairing bad directory entries. **\n");

  while (x != &nil) {

    ent = (ent_info_t *)(x->data);
    
    fprintf(stderr, "Dir dinode: %"FMT_64"u, entry: %"FMT_64"u, err: %s \n",
	    ent->dir->addr, ent->ino, de_errs[fake_log(ent->code)]);


    if (in_badblk(fs, ent->ino)) {
      fprintf(stderr, "Entry also has bad blocks\n");
      goto erase;
    }

    if (ent->code & DE_NODI) {
      if (recover_bmdi(fs, ent->ino) < 0) {
	goto erase;
      }
    }

    if (ent->code & DE_HASH) {
      fprintf(stderr, "No recovery for bad hash\n");
      goto erase;
    }

    if (ent->code & DE_RECLEN) {
      fprintf(stderr, "No recovery for bad record length\n");
      goto erase;
    }
    
    x = Tree_Successor(x);
    continue;

  erase:    
    erase_entry(fs, ent->addr, ent->dir, ent->ino, ent->islf);     

    x = Tree_Successor(x);
  }


  /* fix bad_di */

  x = Tree_Minimum(fs->bad_di);

  if (x != &nil)
    fprintf(stderr, "\n** Repairing bad dinodes. **\n");

  while (x != &nil) {
    
    info = (di_info_t *)(x->data);

    fprintf(stderr, "Dinode %"FMT_64"u, err: %s\n", info->addr, 
	    di_errs[fake_log(info->code)]);


    if (info->code & (DI_ADDR | DI_MAGIC | DI_RGRP | DI_TYPE | DI_OPEN | \
		      DI_ENTCNT | DI_BLKS)) {
      if (repair_difield(fs, info->addr, info->code, info->other) < 0) {
	erase_dinode(fs, info->addr);
	goto next;
      }
    }

    if (info->code & (DI_DEPTH | DI_SAMED | DI_NODOT | DI_BADDOT | \
		      DI_LINK | DI_NOHOME)) {
      fprintf(stderr, "No recovery available\n");
      erase_dinode(fs, info->addr);
    }

  next:
    x = Tree_Successor(x);
  }

}


/******************************************************************************
*******************************************************************************
** NAME
**	repair_difield
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

int repair_difield(filesystem_t *fs, uint64 addr, uint32 code, uint64 other)
{
  int fd = fs->fd;
  gfs_sb_t *sb = fs->sb;
  gfs_dinode_t dinode;
  char data[sb->sb_bsize];


  do_lseek(fd, addr * sb->sb_bsize);
  do_read(fd, data, sb->sb_bsize);

  gfs_dinode_in(&dinode, data);


  if (code & DI_BLKS) {

    if (in_badblk(fs, addr)) {
      fprintf(stderr, "Dinode also in badblks list\n");
      return -1;
    }

    fprintf(stderr, "Current blocks count: %"FMT_64"u, correction: %"FMT_64"u\n",
	    dinode.di_blocks, other);
    dinode.di_blocks = other;
  }

  if (code & DI_ADDR) {
    fprintf(stderr, "Current dinode number: %"FMT_64"u, correction: "
	    "%"FMT_64"u\n", addr, other);
    dinode.di_addr = other;
  }

  if (code & DI_MAGIC) {
    fprintf(stderr, "Current magic number: %x, correction: %x\n", 
	    dinode.di_magic, GFS_MAGIC_DI);
      dinode.di_magic = GFS_MAGIC_DI;
  }  

  if (code & DI_RGRP) {
    fprintf(stderr, "Current rgrp: %u, correction: %u\n", dinode.di_rgrp, 
	    (uint32)other);
    dinode.di_rgrp = other;
  }

  if (code & DI_TYPE) {
    fprintf(stderr, "Current type: 0x%x, correction: 0x%x\n", dinode.di_type,
	    FILE_REG);
    dinode.di_type = FILE_REG;
  }

  if (code & DI_OPEN) {
    fprintf(stderr, "Current nopen: %u, correction: 0\n", dinode.di_nopen);
    dinode.di_nopen = 0;
  }

  if (code & DI_ENTCNT) {
    fprintf(stderr, "Current entry count: %u, correction: %u\n", dinode.di_entries,
	    (uint32)other);
    dinode.di_entries = other;
  }

  if ((fixnopen && !(code & ~DI_OPEN)) ||
      query_yn("Write corrections to disk? [yN]: ")) {

    if (!fixnopen && query_yn("Make all further nopen corrections automatic? [yN]: "))
      fixnopen = 1;

    gfs_dinode_out(&dinode, data);
    do_lseek(fd, addr * sb->sb_bsize);
    do_write(fd, &data, sizeof(gfs_dinode_t));
    return 0;
  }

  return -1;
}


/******************************************************************************
*******************************************************************************
** NAME
**	erase_dinode
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

void erase_dinode(filesystem_t *fs, uint64 di)
{
  int fd = fs->fd;
  gfs_rindex_t *ri = fs->ri, *rip;
  char empty[sizeof(gfs_dinode_t)];
  uint32 rg;
  uint64 offset, off, addr;
  char buf, mask;

  addr = di;

  if (query_yn("Erase this dinode on disk? [yN]: ")) {
    do_lseek(fd, addr * fs->sb->sb_bsize);
    memset(empty, 0, sizeof(gfs_dinode_t));
    do_write(fd, &empty, sizeof(gfs_dinode_t));
  }
  else
    return;

  if (map_bm(fs, &addr, &rg, &offset) < 0)
    return;

  rip = &ri[rg];
  off = rip->ri_addr * fs->sb->sb_bsize + sizeof(gfs_rgrp_t) + offset;
  do_lseek(fd, off);
  do_read(fd, &buf, 1);
  
  mask = (MARK_DI << 2*(addr % GFS_NBBY));
  buf &= ~mask;

  if (query_yn("Make bitmap correction for erased dinode? [yN]: ")) {
    do_lseek(fd, off);
    do_write(fd, &buf, 1);
  }
}


/******************************************************************************
*******************************************************************************
** NAME
**	record_badblk
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

void record_badblk(filesystem_t *fs, uint64 blknum, uint64 diaddr)
{
  node_t *x;
  blk_info_t *blk;
  
  type_zalloc(blk, blk_info_t, sizeof(blk_info_t));
  
  blk->addr = blknum;
  blk->dinode = diaddr;
  
  x = New_Treenode(blk->addr, (char *)blk);
  Tree_Insert(&fs->bad_blk, x);
}


/******************************************************************************
*******************************************************************************
** NAME
**	in_badblk
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int in_badblk(filesystem_t *fs, uint64 di)
{
  node_t *x;
  blk_info_t *blk;

  x = Tree_Minimum(fs->bad_blk);

  while (x != &nil) {
  
    blk = (blk_info_t *)(x->data);

    if (blk->dinode == di)
      return 1;

    x = Tree_Successor(x);
  }

  return 0;
}


/******************************************************************************
*******************************************************************************
** NAME
**	record_baddi
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

void record_baddi(filesystem_t *fs, uint64 addr, uint32 code, uint64 other)
{
  node_t *x;
  di_info_t *di;

  type_zalloc(di, di_info_t, sizeof(di_info_t));

  di->addr = addr;
  di->code = code;
  di->other = other;

  x = New_Treenode(di->addr, (char *)di);
  Tree_Insert(&fs->bad_di, x);
}


/******************************************************************************
*******************************************************************************
** NAME
**	record_badent
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

void record_badent(filesystem_t *fs, uint64 ino, uint64 addr, dir_info_t *dir,
		   int islf, uint32 code)
{
  node_t *x;
  ent_info_t *ent;

  type_zalloc(ent, ent_info_t, sizeof(ent_info_t));

  ent->ino = ino;
  ent->addr = addr;
  ent->dir = dir;
  ent->code = code;
  ent->islf = islf;

  x = New_Treenode(ent->addr, (char *)ent);
  Tree_Insert(&fs->bad_ent, x);
}


/******************************************************************************
*******************************************************************************
** NAME
**	record_badlf
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

void record_badlf(filesystem_t *fs, dir_info_t *dir, gfs_leaf_t *lf)
{


}



/******************************************************************************
*******************************************************************************
** NAME
**	repair_bm
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

void repair_bm(filesystem_t *fs, int rg, int byte, int bit, char old, char new)
{
  int fd = fs->fd;
  gfs_sb_t *sb = fs->sb;
  gfs_rindex_t *ri = fs->ri, *rip;
  uint64 off;
  char buf, out;
  
  memset(&buf, 0, 1);
  memset(&out, 0, 1);

  rip = &ri[rg];

  off = rip->ri_addr * sb->sb_bsize + sizeof(gfs_rgrp_t) + byte;

  do_lseek(fd, off);

  do_read(fd, &buf, 1);


  if (new && !old) {				/* case 2, 5 */
    fprintf(stderr, "\tblock used by file, but not marked in "
	    "on-disk allocation bitmaps\n");
    out = buf | (new << (2*bit));
  }

  else if (old && !new) {       		/* case 4 */
    fprintf(stderr, "\tblock not used by file, but marked in "
	    "allocation bitmaps\n");
    out = buf & ~(old << (2*bit));
  }

  else if ((old == 3) && (new == 1)) {
    fprintf(stderr, "\tregular block marked as dinode block\n");
    out = buf & ~(2 << (2*bit));
  }

  else if ((old == 1) && (new == 3)) {		/* shouldn't happen */
    fprintf(stderr, "\tdinode block marked as regular block\n");
    out = buf | (2 << (2*bit));
  }

  else {
    fprintf(stderr, "(shouldn't be here, no error)\n");
    return;
  }
	
  
  fprintf(stderr, "\tcurrent: %x, suggested repair: %x\n", 0xff & buf, 
	  0xff & out);

  
  if (fixbm || query_yn("Write correction to disk? [yN]: ")) {
    
    if (!fixbm && query_yn("Make all future bitmap corrections? [yN]: "))
      fixbm = 1;
    
    do_lseek(fd, off);
    do_write(fd, &out, 1);
  }
}


/******************************************************************************
*******************************************************************************
** NAME
**	recover_bmdi
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int recover_bmdi(filesystem_t *fs, uint64 addr)
{
  int fd = fs->fd;
  gfs_sb_t *sb = fs->sb;
  gfs_rindex_t *ri = fs->ri, *rip;
  gfs_dinode_t di;
  char buf[sb->sb_bsize];
  char bmbyte, out;
  uint32 rgnum;
  uint64 rgend, off;

  memset(&bmbyte, 0, 1);
  memset(&out, 0, 1);

  if (addr >= sb->sb_fssize) {
    /* fprintf(stderr, "\tbad entry dinode number (out of range)%"FMT_64"u\n", 
       addr); */
    return -1;
  }

  /* dinode wasn't marked (11) in bitmap so was never read */

  do_lseek(fs->fd, addr * sb->sb_bsize);
  do_read(fs->fd, &buf, sb->sb_bsize);
  gfs_dinode_in(&di, buf);

  if (di.di_magic != GFS_MAGIC_DI) {
    /* fprintf(stderr, "\tcan't recover dinode %"FMT_64"u\n", addr); */
    return -1;
  }

  /* we'll assume we can recover this dinode; fix the bitmap and restart */

  rgend = sb->sb_length;

  for (rgnum=0; rgnum < sb->sb_rgrps; rgnum++) {
    rip = &ri[rgnum];
    rgend += rip->ri_data + rip->ri_length;

    if (addr < rgend)
      break;    
  }

  rip = &ri[rgnum];

  addr -= rip->ri_data1;

  off = (rip->ri_addr * sb->sb_bsize) + sizeof(gfs_rgrp_t) + (addr / GFS_NBBY);

  do_lseek(fs->fd, off);
  do_read(fs->fd, &bmbyte, 1);

  out = bmbyte | (3 << (2*(addr % GFS_NBBY)));

  fprintf(stderr, "Recovering dinode in bitmap, ");
  fprintf(stderr, "current bitmap: %x, suggested repair: %x\n", 0xff & bmbyte,
	  0xff & out);
  
  if (query_yn("Write correction to disk? [yN]: ")) {
    do_lseek(fd, off);
    do_write(fd, &out, 1);
  }

  return 0;
}


/******************************************************************************
*******************************************************************************
** NAME
**	erase_entry
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

void erase_entry(filesystem_t *fs, uint64 addr, dir_info_t *dir, uint64 ino, 
		 int islf)
{
  int fd = fs->fd;
  gfs_sb_t *sb = fs->sb;
  gfs_leaf_t leaf;
  gfs_dinode_t dinode;
  gfs_dirent_t de, prev;
  char data[sb->sb_bsize];
  char de_out[sizeof(gfs_dirent_t)];
  char lf_out[sizeof(gfs_leaf_t)];
  char di_out[sizeof(gfs_dinode_t)];
  uint32 space, offset, prev_off=0, ent=0;

  do_lseek(fd, addr * sb->sb_bsize);
  do_read(fd, data, sb->sb_bsize);

  if (islf) {
    gfs_leaf_in(&leaf, data);

    if (leaf.lf_magic != GFS_MAGIC_LF) {
      fprintf(stderr, "Can't erase bad entry\n");
      return;
    }
    
    space = sb->sb_bsize - sizeof(gfs_leaf_t);
    offset = sizeof(gfs_leaf_t);
  }

  else {
    gfs_dinode_in(&dinode, data);
    
    if (dinode.di_magic != GFS_MAGIC_DI) {
      fprintf(stderr, "Can't erase bad entry\n");
      return;
    }

    space = sb->sb_bsize - sizeof(gfs_dinode_t);
    offset = sizeof(gfs_dinode_t);
  }

  while (offset < space) {

    gfs_dirent_in(&de, data + offset);

    if (de.de_ino == 0)
      goto next;

    if (ino == de.de_ino) {

      fprintf(stderr, "Clearing bad dir entry.\n");
      if (query_yn("Write correction to disk? [yN]: ")) {

	/* clear the bad entry */

	de.de_ino = 0;
	gfs_dirent_out(&de, de_out);
	do_lseek(fd, addr * sb->sb_bsize + offset);
	do_write(fd, &de_out, de.de_rec_len);


	/* adjust previous record length */

	if (ent++ != 0) {
	  gfs_dirent_in(&prev, data + prev_off);
	  prev.de_rec_len += de.de_rec_len;      
	  gfs_dirent_out(&prev, de_out);
	  do_lseek(fd, addr * sb->sb_bsize + prev_off);
	  do_write(fd, &de_out, prev.de_rec_len);	
	}

	/* correct the entries count in dinode and leaf */

	if (islf) {
	  leaf.lf_entries--;
	  gfs_leaf_out(&leaf, lf_out);
	  do_lseek(fd, addr * sb->sb_bsize);
	  do_write(fd, &lf_out, sizeof(gfs_leaf_t));

	  do_lseek(fd, dir->addr * sb->sb_bsize);
	  do_read(fd, data, sb->sb_bsize);
	  gfs_dinode_in(&dinode, data);
	  dinode.di_entries--;	  
	  gfs_dinode_out(&dinode, di_out);
	  do_lseek(fd, dir->addr * sb->sb_bsize);
	  do_write(fd, &di_out, sizeof(gfs_dinode_t));	  
	}
	else {
	  dinode.di_entries--;
	  gfs_dinode_out(&dinode, di_out);
	  do_lseek(fd, addr * sb->sb_bsize);
	  do_write(fd, &di_out, sizeof(gfs_dinode_t));
	}
      }  
      /* dir->di_ents--; */
      return;
    }

  next:
    prev_off = offset;
    offset += de.de_rec_len;
  }

  fprintf(stderr, "Can't erase bad entry\n");  
}


/******************************************************************************
*******************************************************************************
** NAME
**	query_yn
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int query_yn(char *str)
{
  char response[3];

  printf("%s", str);

  if (allyes) {
    printf("\n");
    return 1;
  }

  scanf("%s", response);

  if (tolower(response[0]) == 'y')
    return 1;

  return 0;
}


/* stupid, fixme pls */
int fake_log(uint32 n)
{  
  switch (n)
    {   
    case 1: return 0;
    case 2: return 1;
    case 4: return 2;
    case 8: return 3;
    case 16: return 4;
    case 32: return 5;
    case 64: return 6;
    case 128: return 7;
    case 256: return 8;
    case 512: return 9;
    case 1024: return 10;
    case 2048: return 11;
    case 4096: return 12;
    case 8192: return 13;
    case 16384: return 14;
    case 32768: return 15;
    case 65536: return 16;
    };

  return 16; /* unknown */
}


