#include <stdio.h>
#include <stdlib.h>
#include <string.h>



#include "dlock_interface.h"
#include "trans_scsi.h"
#include "dl_endian.h"
#include "trans_debug.h"

typedef struct scsi_data 
{
  char device[256];
  scsi_ops_t *ops;
  void *private;
} scsi_data_t;


void *scsi_open(char *table_name)
{
  struct scsi_data *data;
  int error;

  ENTER(SCSI_OPEN);

  data = (struct scsi_data *)malloc(sizeof(struct scsi_data));

  if (!data) {
    EXIT(SCSI_OPEN);
    return(NULL);
  }

  strcpy(data->device, table_name);

  dprintf("device = %s\n", data->device);

#if defined(__linux__)
  data->ops = &scsi_ops_linux;
#endif

  error = data->ops->open(data->device, &data->private);

  if (error) {
    dprintf("scsi_open: error on os specific transport open\n");
    free(data);
    EXIT(SCSI_OPEN);
    return NULL;
  }

  EXIT(SCSI_OPEN);
  return(data);
}


void scsi_close(void *handle)
{
  scsi_data_t * data = handle;

  ENTER(SCSI_CLOSE);
  data->ops->close(data->device, data->private);

  EXIT(SCSI_CLOSE);
  free(handle);
}


int scsi_do_dlock(void *handle, dlock_data_t *request)
{
  struct scsi_data *data;
  unsigned char cdb[16], s_data[16384];
  int error, len, x;

  ENTER(SCSI_DO_DLOCK);

  data = (scsi_data_t *) handle; 

  cdb[0] = 0x83;
  cdb[1] = request->action & 0xFF;

  
  cdb[2] = (request->locknum >> 24) & 0xFF;
  cdb[3] = (request->locknum >> 16) & 0xFF;
  cdb[4] = (request->locknum >> 8) & 0xFF;
  cdb[5] = (request->locknum >> 0) & 0xFF;

  cdb[6] = (request->clientid >> 24) & 0xFF;
  cdb[7] = (request->clientid >> 16) & 0xFF;
  cdb[8] = (request->clientid >> 8) & 0xFF;
  cdb[9] = (request->clientid >> 0) & 0xFF;

  cdb[10] = 0x00;
  cdb[11] = 0x00;
  cdb[12] = 0x40;
  cdb[13] = 0x00;

  cdb[14] = request->inputversion & 0xFF;
  cdb[15] = 0x00;

  dprintf("scsi_do_dlock: action: %d, lock: %d, cid: %d, inputversion: %d\n", 
	  request->action, request->locknum, request->clientid, request->inputversion);

  error = data->ops->write_cdb(cdb, 16, data->private);

  if (error != 16) {
    dprintf("scsi_do_dlock: error on write_cdb, retval: %d\n", error);
    EXIT(SCSI_DO_DLOCK);
    return 1;
  }

  error = data->ops->read_cdb(s_data, 16384, data->private);

  if (error < 16 ) {
    dprintf("scsi_do_dlock: error on read_cdb, retval: %d\n", error);
    EXIT(SCSI_DO_DLOCK);
    return 1;
  }

  

  if (request->action != 9) {
    request->result =   (s_data[4] >> 7) & 0x01;
    request->activity = (s_data[4] >> 6) & 0x01;
    request->expired =  (s_data[4] >> 3) & 0x03;
    request->state   =  (s_data[4] >> 0) & 0x03;
    request->version =  (s_data[0] << 24) |
                        (s_data[1] << 16) |
                        (s_data[2] << 8 ) |
                        (s_data[3] << 0 );

    request->holders = s_data[5];
    
    len = ((s_data[6] << 8) | (s_data[7] << 0)) / 4;

    if (len != request->holders)
      eprintf(DLF_ERROR, "******scsi_do_dlock: len != request->holders\n");

    dprintf("scsi_do_dlock: clientids %d\n", len);

    for (x=0; x < len; x++) {
      request->clientids[x] = (s_data[4 * x + 8] << 24) |
                              (s_data[4 * x + 9] << 16) |
                              (s_data[4 * x + 10] << 8) |
                              (s_data[4 * x + 11] << 0);

    }

    if (request->holders) {
      dprintf("[");

      for (x=0; x < request->holders; x++)
	dprintf(" %u", request->clientids[x]);

      dprintf("]\n");
    }

    dprintf("scsi_do_dlock: result: %d, state: %d, expired: %d, activity: %d, version: %d, holders: %d\n",
	  request->result, request->state, request->expired, request->activity, request->version, request->holders);
  }

  else {
    len = (s_data[2] << 8) | (s_data[3] << 0);

    memcpy(request->bits, s_data + 4, len);
  }
  
  EXIT(SCSI_DO_DLOCK);
  return(0);
}


int scsi_get_info(void *handle, dlock_info_t *info)
{

  ENTER(SCSI_GET_INFO);
  printf("scsi_get_info()\n");

  EXIT(SCSI_GET_INFO);
  return(0);
}


int scsi_set_info(void *handle, dlock_info_t *info)
{

  ENTER(SCSI_SET_INFO);
  printf("scsi_set_info()\n");

  EXIT(SCSI_SET_INFO);
  return(0);
}



dlock_interface_t trans_scsi =
{
  "scsi",
  scsi_open,
  scsi_close,
  scsi_do_dlock,
  scsi_get_info,
  scsi_set_info
};

