#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <scsi/scsi.h>
#include <scsi/sg.h>

#include "global.h"
#include "trans_scsi.h"
#include "trans_debug.h"

#define REPLY_LEN 16384

typedef struct scsi_linux_s {
  int fd;
} scsi_linux_t;

int scsi_open_linux(char *name, void **private) {
  scsi_linux_t *sl;

  ENTER(SCSI_OPEN_LINUX);

  *private = malloc(sizeof(scsi_linux_t));

  if (! *private) {
    dprintf("scsi_open_linux: can't malloc for private\n");
    EXIT(SCSI_OPEN_LINUX);
    return 1;
  }

  sl = (scsi_linux_t *) *private;

  sl->fd = open(name, O_RDWR);

  if (sl->fd == -1) {
    dprintf("scsi_open_linux: bad fd on open\n");
    free(*private);
    EXIT(SCSI_OPEN_LINUX);
    return 1;
  }

  EXIT(SCSI_OPEN_LINUX);
  return 0;
}

int scsi_close_linux(char *name, void *private) {
  scsi_linux_t *sl;

  ENTER(SCSI_CLOSE_LINUX);

  sl = (scsi_linux_t *) private;

  close(sl->fd);
  free(private);
  
  EXIT(SCSI_CLOSE_LINUX);
  return 0;
}

int read_cdb_linux(void *cdb, int len, void *private) {
  struct sg_header *sg_head;
  scsi_linux_t *sl;
  unsigned char *reply;
  int retval, rep_size;

  ENTER(READ_CDB_LINUX);

  sl = (scsi_linux_t *) private;
  rep_size = sizeof(*sg_head) + len;

  if (!(reply = malloc(rep_size))) {
    dprintf("read_cdb_linux: can't malloc for rep_size\n");
    EXIT(READ_CDB_LINUX);
    return -1;
  }

  retval = read(sl->fd, reply, rep_size);

  if (retval < 0) {
    dprintf("read_cdb_linux: error on read\n");
    goto free_fail;
  }

  sg_head = (struct sg_header *) reply;

  if (sg_head->result || sg_head->sense_buffer[0]) {
    dprintf("read_cdb_linux: bad result on read\n");
    goto free_fail;
  }

  memcpy(cdb, reply + sizeof(*sg_head), retval - sizeof(*sg_head));

  free(reply);
  EXIT(READ_CDB_LINUX);
  return retval;

free_fail:
  free(reply);
  EXIT(READ_CDB_LINUX);
  return -1;

}

int write_cdb_linux(void *cdb, int len, void *private) {
  struct sg_header *sg_head;
  scsi_linux_t *sl;
  unsigned char *req;
  int retval, req_size;

  ENTER(WRITE_CDB_LINUX);

  sl = (scsi_linux_t *) private;
  req_size = sizeof(*sg_head) + len;

  if (!(req = malloc(req_size))) {
    dprintf("write_cdb_linux: can't malloc for req_size\n");
    EXIT(WRITE_CDB_LINUX);
    return -1;
  }

  #ifdef SG_BIG_BUFF
  if (sizeof(struct sg_header) + len > SG_BIG_BUFF ||
      sizeof(struct sg_header) + REPLY_LEN > SG_BIG_BUFF)
  {
   
      dprintf("scsi_cmd():  too big a transfer\n");
      EXIT(WRITE_CDB_LINUX);
    return -1;
  }

#else  /*  SG_BIG_BUFF  */
  if (sizeof(struct sg_header) + len > 4096 ||
      sizeof(struct sg_header) + REPLY_LEN > 4096)
  {
  
    dprintf("scsi_cmd():  too big a transfer\n");
    EXIT(WRITE_CDB_LINUX);
    return -1;
  }
#endif  /*  SG_BIG_BUFF  */

  sg_head = (struct sg_header *)req;
  memset(sg_head, 0, sizeof(struct sg_header));
  sg_head->pack_len = sizeof(struct sg_header) + len;
  sg_head->reply_len = sizeof(struct sg_header) + REPLY_LEN;
  /*sg_head->pack_id = tag;*/
  sg_head->twelve_byte = 0; /* 0.9.4 specific */

  memcpy(req + sizeof(*sg_head), cdb, len);

  retval = write(sl->fd, req, sizeof(*sg_head) + len);

  if (retval < 0) {
    dprintf("write_cdb_linux: error on write\n");
    goto free_fail;
  }

  if (retval != sizeof(*sg_head) + len) { 
    dprintf("write_cdb_linux: Didn't write all the bytes\n");
  }

  free(req);
  EXIT(WRITE_CDB_LINUX);
  return retval - sizeof(*sg_head);

 free_fail:
  free(req);
  EXIT(WRITE_CDB_LINUX);
  return -1;

}

scsi_ops_t scsi_ops_linux = {
  scsi_open_linux,
  scsi_close_linux,
  read_cdb_linux,
  write_cdb_linux
};
